/* admind.c
 * (c) SUSE Linux AG 2003
 * Author: Till Franke <franke@suse.com>

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHOR OR SUSE BE LIABLE FOR ANY CLAIM, DAMAGES
OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/


#include <stdio.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <syslog.h>
#include <libgen.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include "identc.h"

char * VERSION1 = "1.0";
int ARG_IDENT = 1;
int ARG_PEER = 1;
int ARG_DEBUG = 0;
int ARG_PORT=0;

void usage(char *cmdname) {
  fprintf(stderr,"Usage: %s [-vIP] [configfile]\n",cmdname);
  fprintf(stderr,"\t-v: verbose logging\n\t-I: no ident lookup\n\t-P: no peer lookup\n");
  /*  fprintf(stderr,"\t-p <port>: tcp socket to open\n");*/
}

typedef struct command_s {
  char key;
  char *command;
  struct command_s * next;
} command_t;

typedef struct server_s {
  in_addr_t server_ip;
  struct server_s *next;
} server_t;
typedef struct user_s {
  char * username;
  struct user_s *next;
} user_t;

int read_config(FILE *configfile, command_t ** commands, server_t ** servers, user_t ** users) {
  char *buf = NULL;
  int bufsize = 0;
  int c;
  char *p,**q;
  int linelen = 0;
  command_t *cmd;
  server_t * srv;
  user_t * usr;
  struct hostent *hostentry;

  *commands = NULL;
  *servers = NULL; 
  *users = NULL;

  cmd = *commands;
  srv = *servers;
  usr = *users;

  do {
    c = fgetc(configfile);
    switch(c) {
    case EOF:
    case '\n' :
      /* read config line */
      if (linelen) {
	buf[linelen] = '\0';
	p = buf;
	switch(*p) {
	case '#': 
	  break;
	case 'S':
	  p += 2;
	  hostentry = gethostbyname(p);
	  if (! hostentry) {
	    syslog(LOG_WARNING,"could not resolve host %s:%s",p,hstrerror(h_errno));
	    break;
	  }
	  if (hostentry->h_addrtype != AF_INET) {
	    syslog(LOG_WARNING,"could not resolve host %s: no AF_INET address",p);
	    break;
	  }
	  syslog(LOG_DEBUG,"adding host %s",hostentry->h_name);
	  q=hostentry->h_addr_list;
	  while (*q) {
	    syslog(LOG_DEBUG,"adding address %hhu.%hhu.%hhu.%hhu",
	    	   (*q)[0],(*q)[1],(*q)[2],(*q)[3]);
	    srv = calloc(1,sizeof(server_t));
	    memcpy(&(srv->server_ip), *q, sizeof(in_addr_t));
	    srv->next = *servers; *servers = srv;
	    q ++;
	  }
	  break;
	case 'U':
	  p += 2;
	  usr = calloc(1,sizeof(user_t));
	  usr->username = calloc(1,strlen(p)+1);
	  strcpy(usr->username, p);
	  usr->next = *users; *users = usr;
	  break;
	case 'X':
	  p += 2;
	  cmd = calloc(1,sizeof(command_t));
	  cmd->key = *p;
	  p += 2;
	  cmd->command = calloc(1,strlen(p)+1);
	  strcpy(cmd->command, p);
	  cmd->next = *commands; *commands = cmd;
	  break;
	default:
	  syslog(LOG_WARNING, "Ignored unknown config line %s", buf);
	}
      }
      linelen = 0;
      break;
    default : 
      if (linelen > bufsize - 2) {
	bufsize += 128;
	buf = realloc(buf, bufsize);
	if (! buf) {
	  syslog(LOG_ERR,"memory reallocation failed:%s",strerror(errno));
	  return -7;
	}
      }
      buf[linelen++] = c;
    }
  }while ( c != EOF) ;
  if (buf) free(buf);
  return 1;
}
      
void drop_privileges() {
  struct passwd * pwd;
  
  pwd = getpwnam("nobody");
  if (pwd) {
    if (initgroups(pwd->pw_name, pwd->pw_gid) != 0) exit(10);
    if (setgid(pwd->pw_gid) != 0) exit(10);
    if (setuid(pwd->pw_uid) != 0) exit(10);
  }
  else {
    exit(11);
  }
}

/*
 * get_command
 * find command in commands list by key
 * return command line on success, NULL if not found
*/
char *get_command( command_t *commands, char c) {
  char *rv = NULL;
  while (commands != NULL) {
    if (c == commands->key) {
      rv = commands->command;
      break;
    }
    commands = commands->next;
  }
  return rv;
}

char *get_server_name_ip(server_t * servers, in_addr_t server_ip) {
  char *rv = NULL;
  char *c = (char*)&server_ip;
  syslog(LOG_DEBUG, "checking peer %hhu.%hhu.%hhu.%hhu", c[0],c[1],c[2],c[3]);

  while (servers != NULL) {
    c = (char*)&(servers->server_ip);
    if (servers->server_ip == server_ip) {
      rv = "OK";
      break;
    }
    servers = servers->next;
  }
  return rv;
}

char *get_username(user_t * users, char *username) {
  char *rv = NULL;
  syslog(LOG_DEBUG,"checking username '%s'",username);
  while (users != NULL) {
    if (! strcmp(username, users->username)) {
      rv = users->username;
      break;
    }
    users = users->next;
  }
  return rv;
}

int run_command(char *command) {
  syslog(LOG_INFO,"executing command '%s'\n",command);
  return(system(command));

}
  
int start_server(command_t *commands, int *serverpipe, int *clientpipe) {
  int srv[2];
  int cli[2];
  pid_t pid;
  
  if ( pipe(srv) < 0 || pipe(cli) < 0) {
    return 0;
  }

  if ((pid = fork()) < 0) return 0;
  if (pid) { /* parent = client, low privilege */
    close(srv[1]); 
    close(cli[0]);
    *serverpipe = srv[0]; /* client reads from this */
    *clientpipe = cli[1]; /* client writes into this */
    drop_privileges();
    
  } else { /* child = server, high privilege */
    close(srv[0]);
    close(cli[1]);
    fclose(stdin);
    fclose(stdout);
    fclose(stderr);

    while ( 1 ) { /* endless loop, exit only on error and client close */
      char c;
      int i,j;
      char *command;
      
      if (read(cli[0], &c,1)) {
	command = get_command(commands, c);
	if (command) {
	  i = run_command(command);
	} else {
	  syslog(LOG_WARNING, "no command found for command key '%c' (0x%x)",
		 (int)c,(int)c);
	  i = -1;
	}
	do {
	  j = write(srv[1], &i,sizeof(int));
	} while (errno == EINTR);
	if (j < (int)sizeof(int)) {
	  syslog(LOG_ERR,"server: cannot write returncode to client");
	}
      } else if (errno != EINTR) {
	syslog(LOG_DEBUG,"server shutting down");
	exit(0);
      }
    }
  }

  return pid;
}

int check_peer(int socket, server_t * servers) {
  struct sockaddr_in sa;
  socklen_t salen;
  in_addr_t *addr;
  char *c;
  salen = sizeof(sa);
  if (getpeername(socket, (struct sockaddr*)&sa, &salen) < 0) {
    syslog(LOG_ERR, "cannot get socket peer name");
    return 0;
  }
  addr = &sa.sin_addr.s_addr;
  if (get_server_name_ip(servers, *addr)) return 1;
  c = (char*)addr;
  syslog(LOG_NOTICE, "peer %hhu.%hhu.%hhu.%hhu rejected", c[0],c[1],c[2],c[3]);
  return 0;
}

int check_ident(int socket, user_t *users) {
  char * username = NULL;
  int i;
  i = identc(socket, &username);
  if (i <= 0) return i;
  if ( get_username(users, username) ) return 1;
  syslog(LOG_NOTICE, "user %s rejected", username);
  return 0;
}



void client_loop(int sock_out, int clientpipe, int serverpipe) {
  int i;
  char returncode[32];
  while ((i = getchar()) != EOF) {
    if ( isalnum(i)) {
      char c = (char) i;
      int rv;
      do {
	i = write(clientpipe, &c, 1);
      } while (errno == EINTR);
      if (i <= 0) {
	syslog(LOG_ERR, "cannot write to server: %s", strerror(errno));
	break;
      }
      do {
	i = read(serverpipe, &rv, sizeof(int));
      } while (errno == EINTR);
      if (i == sizeof(int)) {
	snprintf(returncode,31,"%d\n",rv);
      } else {
	snprintf(returncode,31,"-1\n");
	syslog(LOG_ERR, "cannot read return code from server");
      }
      do {
	i = write(sock_out, returncode, strlen(returncode));
      } while (errno == EINTR);
    }
  }
}

  /* main Main MAIN */

int main(int argc, char **argv) {
  FILE *configfile;
  char *configfilename= SYSCONFDIR"/SLEPOS/admind.conf";
  int i;
  char *arg;
  pid_t pid;
  command_t * commands;
  user_t * users;
  server_t *servers;
  int serverpipe; /* server to client, read end */
  int clientpipe; /* client to server, write end */
  char *myname;
  char message[81];


  myname = basename(argv [0]);
  openlog(myname, LOG_CONS | LOG_PID, LOG_DAEMON);
  int sock_in, sock_out;

  sock_in = fileno(stdin);
  sock_out = fileno(stdout);

  i = 0;
  while (++i < argc  && *argv[i] == '-') {
    arg = argv[i]; ++arg;
    switch( *arg) {
    case 'I': ARG_IDENT = 0; break;
    case 'P': ARG_PEER = 0; break;
    case 'v': ARG_DEBUG =1; break;
    case 'p':  ++arg;
      if (*arg && isdigit(*arg)) {
	ARG_PORT=atoi(arg);
      } else if (++i < argc && isdigit(*argv[i])) {
	ARG_PORT = atoi(argv[i]);
      } else { usage(myname) ; exit(5); }
      break;
    default: usage(myname) ; exit(4);
    }
  }
  if (! ARG_DEBUG) {
    setlogmask(setlogmask(0) ^ LOG_MASK(LOG_DEBUG));
  }
  if (ARG_PORT) {
    fprintf(stderr,"port %d chosen, but currently unimplemented - use inetd\n",ARG_PORT);
    exit(6);
  }
    
  if ( i < argc ) {
    configfilename = argv[i];
  }
  
  configfile = fopen(configfilename, "r");
  if (! configfile) {
    syslog(LOG_ERR,"Cannot read config file %s: %s", configfilename, strerror(errno));
    exit(1);
  }
  
  if (! read_config(configfile, &commands, &servers, &users)) {
        syslog(LOG_ERR,"invalid config file file %s", configfilename);
    exit(2);
  }
  fclose(configfile); 

  if (! (pid = start_server(commands, &serverpipe, &clientpipe))) {
    syslog(LOG_ERR, "cannot start server: %s", strerror(errno));
    exit(3);
  }

  /* main loop for client */
  syslog(LOG_DEBUG,"starting up, server is %d", pid);
  snprintf(message,80, "ADMIND(%s) %s\n",myname, VERSION1);
  do {
    i = write(sock_out, message, strlen(message));
  } while (errno == EINTR);
  
  /* verify socket peer */

  if (ARG_PEER) {
    int peer_ok;
    if ((peer_ok =  check_peer(sock_in, servers)) < 0) {
      syslog(LOG_WARNING, "cannot verify peer");
      exit(7);
    }
    if (! peer_ok)   exit(8);
  }
  if (ARG_IDENT) {
    int ident_ok;
    if ((ident_ok = check_ident(sock_in, users)) < 0) {
      syslog(LOG_WARNING, "cannot verify remote user identity");
      exit(8);
    }
    if (! ident_ok) exit(9);
  }

  client_loop(sock_out, clientpipe, serverpipe);
  
  close(clientpipe);
  waitpid(pid, &i, 0);
  syslog(LOG_DEBUG,"shutting down");
  exit(0); 
}


