/* ident client lookup */

#include <sys/socket.h>
#include <sys/types.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include "identc.h"
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>

/* maximum ident response length in bytes 
 * (RFC1413 specifies 512 bytes) */
#define MAXRESPONSELEN 8192

int
identc(int sock, char **username) {
  struct in_addr *addr;
  in_port_t localport, remoteport;
  struct sockaddr_in sar, sal, sai;
  socklen_t salen;
  struct servent * serviceentry;
  char request[64];
  char *response = NULL;
  int resplen = 0;
  int bufsize = 0;
  int i;
  char c;
  char *tok, *cp;
  int idsock;
  
  *username = NULL;
  salen = sizeof(sar);
  if (getpeername(sock, (struct sockaddr*)&sar, &salen) < 0) {
    syslog(LOG_ERR, "cannot get socket peer name");
    return -1;
  }
  addr = &sar.sin_addr;
  remoteport = ntohs(sar.sin_port);
  salen = sizeof(sal);
  if (getsockname(sock, (struct sockaddr*)&sal, &salen) < 0) {
    syslog(LOG_ERR, "cannot get socket local name");
    return -2;
  }
  localport = ntohs(sal.sin_port);

  if ((idsock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    syslog(LOG_ERR,"cannot create ident socket: %s",strerror(errno));
    return -3;
  }
  sai.sin_family = AF_INET;
  sai.sin_addr = *addr;

  serviceentry = getservbyname("ident","tcp");
  if (serviceentry) {
    sai.sin_port = serviceentry->s_port;
  } else {
    sai.sin_port = htons(113L);
  }
  salen = sizeof(sai);
  if (connect(idsock, (struct sockaddr*)&sai, salen) < 0 ) {
    syslog(LOG_ERR,"cannot create ident socket: %s",strerror(errno));
    return -4;
  }
  snprintf(request,63,"%d,%d\n", remoteport, localport);
  syslog(LOG_DEBUG,"sending ident request '%s'", request);
  do {
    i = write(idsock, request, strlen(request));
  } while (errno == EINTR);
  if (i < 0) {
    syslog(LOG_ERR,"cannot send ident request");
    return -5;
  }
  do {
    do {
      i = read(idsock, &c, 1);
    } while (errno == EINTR);
    if (c == '\n') break;
    if (resplen > bufsize - 2) {
      if (bufsize > MAXRESPONSELEN) {
	syslog(LOG_ERR,"ident response too large");
	return -6;
      }

      bufsize += 128;
      response = realloc(response, bufsize);
      if (! response) {
	syslog(LOG_ERR,"memory reallocation failed:%s",strerror(errno));
	return -7;
      }
    }
    response[resplen++] = c;
  } while (i > 0);
  if (! resplen) {
    syslog(LOG_INFO, "server denied ident request");
    return 0;
  }
  if (response[resplen-1] != '\r') resplen++; /* overwrite CR */
  response[resplen-1] = '\0'; /* terminate string */
  syslog(LOG_DEBUG,"ident server returned '%s",response);

  /* response should be 'request ":" condition ":" type [":" username]' */
  tok = strtok(response, ":"); /* this is the request again */
  if (! tok) {    
    syslog(LOG_INFO,"RFC1413 protocol violated : no ':' tokens found");
    return 0;
  }
  tok = strtok(NULL,":"); while (isspace(*tok)) tok++;
  if (! tok) {
    syslog(LOG_INFO,"RFC1413 protocol violated :  no condition ");
    return 0;
  }

  if (! strncmp(tok, "ERROR",5)) {
    syslog(LOG_INFO, "server returned error on ident request");
    tok = strtok(NULL,":");
    syslog(LOG_DEBUG,"server identd token :'%s'",tok);
    return 0;
  }
  if (strncmp(tok,"USERID",6)) {
    syslog(LOG_INFO,"RFC1413 protocol violated :  no USERID or ERROR ");
    syslog(LOG_DEBUG,"server identd token :'%s'",tok);
    return 0;
  }

  tok = strtok(NULL,":");  /* the OS type - ignored */
  
  tok = strtok(NULL,":"); while (isspace(*tok)) tok++; /* the user ID  */
  if (! tok) {
    syslog(LOG_INFO,"RFC1413 protocol violated :  no userid");
    return 0;
  }
  cp = tok + strlen(tok) ;
  while(--cp >= tok && isspace(*cp)) { /*kill whitespace at end of string */
    *cp = '\0';
  }
  if (! *tok) {
    syslog(LOG_INFO,"RFC1413 protocol violated :  no userid, only white space");
    return 0;
  }
  resplen = strlen(tok);
  *username = malloc(resplen + 1);
  strcpy(*username, tok);
  

  free(response);
  return resplen;
}

  
