package POS::Admin;

use warnings;
use strict;

use POS::LDAP;
use POS::LDAP::Cache;
use POS::Config;
use POS::Syslog;
use POS::Admin::LdapStruct qw(%LDAP_STRUCT);
use POS::Admin::Constant
  qw(POS_NO_ERROR POS_UNKNOWN_OBJECTCLASS POS_VALIDATION_FAILED POS_LDAP_ERROR POS_NO_LDAP POS_FILE_NOT_FOUND POS_NO_DATA POS_ALREADY_EXISTS POS_GENERAL_ERROR POS_NO_ENTRY POS_NO_RECURSIVE);

#use vars qw( $VERSION @ISA @EXPORT );

sub new
{
    my ($self, %args) = @_;
    my $type = ref($self) || $self;
    $self = bless {%args}, $type;

    # by default set to batch processing when using as object
    $self->{'ERROR'}      = POS_NO_ERROR;
    $self->{'BATCH_MODE'} = 1 unless defined $self->{'BATCH_MODE'};
    $self->{'CONFIG'}     = POS::Config->new() unless $self->{'CONFIG'};
    $self->{'LDAP'}       = POS::LDAP->new('CONFIG' => $self->{'CONFIG'}) unless $self->{'LDAP'};
    # error messages produced by POS::LDAP, error handling left on upper level
    return if (not $self->{'LDAP'});
    $self->{'CACHE'} = POS::LDAP::Cache->new(
        'CONFIG'       => $self->{'CONFIG'},
        'LDAP'         => $self->{'LDAP'},
        'LDAP_DATA'    => [],
        'CHANGES_ONLY' => 1
    ) if $self->{'BATCH_MODE'};
    # all ok
    return $self;
} ## end sub new

## @fn _getRunningMode ($self)
# determine if procedure is called as method (oop) or as function
# @param $self either POS::Admin ref or POS::LDAP ref or something else
# @return ref for POS::Admin or POS_NO_LDAP
sub _getRunningMode
{
    my ($self) = @_;
    if (ref $self eq 'POS::Admin') {
        return $self;
    }
    elsif (ref $self eq 'POS::LDAP') {
        $self = POS::Admin->new('BATCH_MODE' => 0, 'CONFIG' => $self->{'CONFIG'}, 'LDAP' => $self);
    }
    else {
        $self = POS::Admin->new('BATCH_MODE' => 0);
    }
    unless ($self) {
        return POS_NO_LDAP;
    }
    return $self;
} ## end sub _getRunningMode

## @method clearError void
# reset error status of POS::Admin class
# @return void
sub clearError
{
    my ($self) = @_;
    $self->{'ERROR'} = POS_NO_ERROR;
} ## end sub clearError

###############################################################################
# INTERFACE METHODS

## @method add ($self_or_ldap, $dn, $objectClass, $data)
# adds defaults, validates data and add entry into LDAP
# @param $self_or_ldap Either POS::Admin or POS::LDAP object, depending on calling (as function or as method)
# @param $dn distinguished name of adding entry
# @param $objectClass objectClass of adding entry, must match entry in %POS::Admin::LdapStruct
# @param $data hash_ref containg attributes and values to be added into LDAP entry
# @return POS_NO_ERROR on success or any constant from POS::Admin::Constant on error
sub add
{
    my ($self, $dn, $objectClass, $data, $entry);
    $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    ($dn, $objectClass, $data) = @_;
    $entry = Net::LDAP::Entry->new($dn);
    my $nosm = exists $data->{'nosm'} ? 1 : 0;
    delete $data->{'nosm'};
    # find object class
    if (not defined $LDAP_STRUCT{$objectClass}) {
        log_error_printf('Unknown object class: %s', $objectClass);
        $self->{'ERROR'} = POS_UNKNOWN_OBJECTCLASS if ($self);
        return POS_UNKNOWN_OBJECTCLASS;
    }
    # add proper objectClasses, remove if they are in data
    delete $data->{'objectClass'};
    foreach my $objClass ($LDAP_STRUCT{$objectClass}{'objectClass'}) {
        $entry->add('objectClass' => $objClass);
    }
    # add defaults if not set
    foreach my $defAttr (keys %{ $LDAP_STRUCT{$objectClass}{'default'} }) {
        unless (defined $data->{$defAttr} and $data->{$defAttr}) {
            log_notice_printf('adding default %22s => %s', $defAttr, $LDAP_STRUCT{$objectClass}{'default'}{$defAttr});
            $data->{$defAttr} = $LDAP_STRUCT{$objectClass}{'default'}{$defAttr};
        }
    }
    # fill attributes into entry
    eval { $entry->add(%$data); };
    if ($@) {
        log_error_printf('Entry attribute list is malformed: %s', $@);
        return POS_VALIDATION_FAILED;
    }
    # pre-write operations
    $self->_preLdapWrite($objectClass, $entry);
    # validate entry
    return POS_VALIDATION_FAILED if ($self->_validateEntry($objectClass, $entry) != POS_NO_ERROR);
    # add to internal cache or directly update LDAP
    if ($self->{'BATCH_MODE'}) {
        $self->{'CACHE'}->mergeEntryToCache($entry);
    }
    else {
        if ($self->_validateParent($objectClass, $dn) != POS_NO_ERROR) {
            return $self->{'ERROR'};
        }
        my $res = $entry->update($self->{'LDAP'});
        if ($res->is_error()) {
            log_error_printf('LDAP error %n when adding entry (%s)', $res->code, $res->error);
            return POS_LDAP_ERROR;
        }
        # do postLdapWrite
        $self->_postLdapWrite($objectClass, $entry, $nosm);
    }
    $self->clearError();
    return POS_NO_ERROR;
} ## end sub add

## @method modify ($self_or_ldap, $dn, $objectClass, $data)
# modifies entry, validates data and update it back into LDAP
# @param $self_or_ldap Either POS::Admin or POS::LDAP object, depending on calling (as function or as method)
# @param $dn distinguished name of adding entry
# @param $objectClass objectClass of adding entry, must match entry in %POS::Admin::LdapStruct
# @param $data hash_ref containg attributes and values to be modified in LDAP entry
# @return POS_NO_ERROR on success or any constant from POS::Admin::Constant on error
sub modify
{
    my ($self, $dn, $objectClass, $data, $entry);
    $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    ($dn, $objectClass, $data) = @_;
    # find object class
    use POS::Admin::LdapStruct(qw(%LDAP_STRUCT));
    if (not defined $LDAP_STRUCT{$objectClass}) {
        log_error_printf('Unknown object class "%s"', $objectClass);
        $self->{'ERROR'} = POS_UNKNOWN_OBJECTCLASS if ($self);
        return POS_UNKNOWN_OBJECTCLASS;
    }
    if ($self->{'BATCH_MODE'}) {
        # create fake entry for delete - existence is checked during sync
        use Net::LDAP::Entry;
        my $entry = Net::LDAP::Entry->new($dn);
        $entry->add('objectClass' => $objectClass);
        $entry->changetype('modify');
        foreach (keys(%$data)) {
            # replace defined data, delete where val is empty
            if (defined $data->{$_}) {
                $entry->replace($_ => $data->{$_});
            }
            else {
                $entry->delete($_);
            }
        }
        $self->{'CACHE'}->mergeEntryToCache($entry);
    }
    else {
        use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);
        my $search = $self->{'LDAP'}->search(scope => 'base', base => $dn, filter => "(objectClass=$objectClass)");
        if ($search->code == LDAP_NO_SUCH_OBJECT) {
            log_error_printf('Specified entry (%s) does not exists, check DN', $dn);
            return POS_NO_ENTRY;
        }
        elsif ($search->is_error()) {
            log_error_printf('LDAP error %n when searching for entry (%s)', $search->code, $search->error);
            return POS_LDAP_ERROR;
        }
        elsif ($search->count() == 0) {
            log_error_printf('Entry with DN "%s" exists but is of different objectClass then requested (%s)',
                $dn, $objectClass);
            return POS_NO_ENTRY;
        }
        else {
            # because we requested base scope, only one entry can be returned
            $entry = $search->pop_entry();
        }
        my ($moddn, $modrdn);
        # detect modrdn operation - dn part attribute is changed -> simple rename
        my ($dnPart, $dnPartValue) = $dn =~ /^([^=]+)=([^,]+)/;
        if ($data->{$dnPart}) {
            log_debug_print('modrdn operation detected');
            $modrdn = $dnPart . '=' . $data->{$dnPart};
            delete $data->{$dnPart};
        }
        # detect moddn operation - --newBase attribute is present -> move entry to another subtree
        if ($data->{'newBase'}) {
            log_debug_print('moddn operation detected');
            $moddn = $data->{'newBase'};
            delete $data->{'newBase'};
            return $self->{'ERROR'}
              if ($self->_validateParent($objectClass, sprintf('%s=%s,%s', $dnPart, $dnPartValue, $moddn)) !=
                POS_NO_ERROR);
        }
        if (%$data) {
            # fill attributes into entry
            eval { $entry->replace(%$data); };
            if ($@) {
                log_error_printf('Entry attribute list is malformed: %s', $@);
                return POS_VALIDATION_FAILED;
            }
            # pre-write operations
            $self->_preLdapWrite($objectClass, $entry);
            # validate entry
            return $self->{'ERROR'} if ($self->_validateEntry($objectClass, $entry) != POS_NO_ERROR);
            my $res = $entry->update($self->{'LDAP'});
            if ($res->is_error()) {
                log_error_printf('LDAP error %n when modifying entry (%s)', $res->code, $res->error);
                return POS_LDAP_ERROR;
            }
        }
        if ($moddn or $modrdn) {
            $entry->changetype('moddn');
            # if we are not renaming, use old cn=... as rdn. rdn is mandatory!
            my $res = $entry->update(
                $self->{'LDAP'},
                newsuperior => $moddn || undef,
                newrdn => $modrdn || sprintf('%s=%s', $dnPart, $dnPartValue),
                deleteoldrdn => '1'
            );
            if ($res->is_error()) {
                log_error_printf('LDAP error %n when moving/renaming entry (%s)', $res->code, $res->error);
                return POS_LDAP_ERROR;
            }
            # LDAP write was success, remove possible old ACLs and SM registrations
            $entry->changetype('delete');
            $self->_postLdapWrite($objectClass, $entry);
            # then write new in usual way, we just need to update entry DN and return original changetype
            my $newDN = $modrdn || sprintf('%s=%s', $dnPart, $dnPartValue);
            if ($moddn) {
                $newDN = $newDN . ',' . $moddn;
            }
            else {
                my ($oldBase) = $entry->dn() =~ /^[^,]+,(.+)$/;
                $newDN = $newDN . ',' . $oldBase;
            }
            $entry->dn($newDN);
            $entry->changetype('moddn');
        }
        # do postLdapWrite
        $self->_postLdapWrite($objectClass, $entry);
    }
    $self->clearError();
    return POS_NO_ERROR;
} ## end sub modify

## @method remove ($self_or_ldap, $dn, $recursive)
# removes entry from LDAP
# @param $self_or_ldap Either POS::Admin or POS::LDAP object, depending on calling (as function or as method)
# @param $dn distinguished name of adding entry
# @param $recursive Set to allow removing of whole subtrees
# @return POS_NO_ERROR on success or any constant from POS::Admin::Constant on error
sub remove
{
    my ($self, $dn, $objectClass, $entry, $recursive, $nosm);
    $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    ($dn, $recursive, $nosm, $objectClass) = @_;
    if ($self->{'BATCH_MODE'}) {
        # create fake entry for delete - existence is checked during sync
        use Net::LDAP::Entry;
        my $entry = Net::LDAP::Entry->new($dn);
        $entry->add('objectClass' => $objectClass);
        $entry->changetype('delete');
        $self->{'CACHE'}->mergeEntryToCache($entry);
    }
    else {
        # find object class
        use POS::Admin::LdapStruct(qw(%LDAP_STRUCT));
        use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);
        my @sortedEntries;
        my $search =
          $self->{'LDAP'}->search(scope => 'sub', base => $dn, filter => "(objectClass=*)", attrs => ['objectClass']);
        if ($search->code == LDAP_NO_SUCH_OBJECT) {
            log_warn_printf('Specified entry (%s) does not exists, nothing to remove', $dn);
        }
        elsif ($search->is_error()) {
            log_error_printf('LDAP error %n when searching for entry (%s)', $search->code, $search->error);
            return POS_NO_ENTRY;
        }
        @sortedEntries = $search->sorted();
        if (scalar @sortedEntries > 1 and not $recursive) {
            log_error_print(
                'Deleting entry has children but recursive mode was not enabled. Check DN or allow recursive mode!');
            return POS_NO_RECURSIVE;
        }
        while ($entry = pop @sortedEntries) {
            #find correct objectClass for _postLdapWrite
            ($objectClass) = grep { $LDAP_STRUCT{$_} } $entry->get_value('objectClass');
            my $res = $self->{'LDAP'}->delete($entry);
            if ($res->is_error()) {
                if ($res->code == LDAP_NO_SUCH_OBJECT) {
                    log_info_printf('Entry %s is not present in LDAP, skipping', $dn);
                }
                else {
                    log_error_printf('LDAP error %n when deleting entry (%s)', $res->code, $res->error);
                    return POS_LDAP_ERROR;
                }
            }
            # do postLdapWrite
            $entry->changetype('delete');
            $self->_postLdapWrite($objectClass, $entry, $nosm);
        }
    }
    $self->clearError();
    return POS_NO_ERROR;
} ## end sub remove

## @method query ($self_or_ldap_or_nothing, $data)
# queries LDAP
# @param $self_or_ldap Either POS::Admin or POS::LDAP object, depending on calling (as function or as method) or nothing
# @param $base search base
# @param $filter search filter
# @param $attrs attributes to be retrieved from LDAP
# @param $scope search scope
# @return POS::LDAP::Cache ref on success else error code
sub query
{
    my $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    my ($data) = @_;
    my %data;
    eval { %data = %$data };
    if ($@) {
        log_crit_print('Wrong parameter format for query! Cannot construct parameter structure');
        $self->{'ERROR'} = POS_GENERAL_ERROR;
        return POS_GENERAL_ERROR;
    }
# adapt $data for ldapsearch
# if defined data should specify search constraints as filter
#   'attr_name' => 'attr_value'
# if attr_value is undef, * is assumed
# to specify which attributes to print use 'attr_name' eq 'attributes' and 'attr_value' as comma separated list of attribute names
#   'attributes' => 'objectClass, *, +' ...
# if attr_value is undef,
# search
    my ($filter, $scope, $base, @attributes) = ('', '', '');
    foreach (keys(%$data)) {
        my ($key) = /-?-?(.*)/;
        if ($key eq 'attributes') {
            @attributes = split(/,/, $data->{$_});
            unless (@attributes) {
                @attributes = ('*');
            }
        }
        elsif ($key eq 'scope') {
            $scope = $data->{$_};
        }
        elsif ($key eq 'base') {
            $base = $data->{$_};
        }
        else {
            if ($data->{$_}) {
                $filter = $filter . '(' . $key . '=' . $data->{$_} . ')';
            }
            else {
                $filter = $filter . '(objectClass=' . $key . ')';
            }
        }
    }
    if ($filter) {
        $filter = '(&' . $filter . '(objectClass=*))';
    }
    else {
        $filter = '(objectClass=*)';
    }
    unless ($base) {
        $base = $self->{'CONFIG'}{'ALL_LDAPBASE'};
    }
    unless ($scope) {
        $scope = 'sub';
    }
    unless (@attributes) {
        @attributes = ('*');
    }
    log_debug_printf('using filter: %s', $filter);
    log_debug_printf('using attrs : %s', join(',', @attributes));
    log_debug_printf('using base  : %s', $base);
    log_debug_printf('using scope : %s', $scope);
    my $search = $self->{'LDAP'}->search(filter => $filter, attrs => \@attributes, scope => $scope, base => $base);
    if ($search->is_error()) {
        if ($search->code != LDAP_NO_SUCH_OBJECT) {
            log_error_printf('LDAP query error %n (%s)', $search->code, $search->error);
        }
        else {
            log_debug_print('LDAP reports no such object');
        }
        $self->{'ERROR'} = $search->code;
        return $search->code();
    }
    my $cache = POS::LDAP::Cache->new(
        'LDAP_DATA' => $search,
        'BASE'      => $base,
        'CONFIG'    => $self->{'CONFIG'},
        'LDAP'      => $self->{'LDAP'}
    );
    unless (defined $cache) {
        $self->{'ERROR'} = POS_LDAP_ERROR;
        return POS_LDAP_ERROR;
    }
    $self->clearError();
    return $cache;
} ## end sub query

## @method validate ($self_or_ldap_or_nothing, $base)
# validates LDAP or LDAP subtree if $base is specified
# @param $self_or_ldap Either POS::Admin or POS::LDAP object, depending on calling (as function or as method) or nothing
# @param $base specifies subtree to validate
# @return POS_NO_ERROR on success
sub validate
{
    my ($self, $base);
    $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    ($base) = @_;
    $self->{'ERROR'} = POS_VALIDATION_FAILED;
    log_debug_print('starting LDAP validation');
    $base = $self->{'CONFIG'}{'ALL_LDAPBASE'} unless defined $base;
    use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);
    my $search = $self->{'LDAP'}->search(scope => 'sub', base => $base, filter => '(objectClass=*)');

    if ($search->code == LDAP_NO_SUCH_OBJECT) {
        log_info_printf('No entry under base: "%s". May happed during initial import.', $base);
    }
    elsif ($search->is_error()) {
        log_error_printf('LDAP error: %s (%n)', $search->error, $search->code);
        return POS_LDAP_ERROR;
    }
    # baseDN, rootClass, relaxed, ldapData, mergeData, mergeMode
    unless ($self->_validate($base, undef, 0, $search)) {
        log_error_printf('LDAP validation failed!');
        return POS_VALIDATION_FAILED;
    }
    log_debug_print('validation successful');
    $self->clearError();
    return POS_NO_ERROR;
} ## end sub validate

## @method sync ($self)
# validates changes in internal cache against LDAP and apply them if validation is successful
# @param $self POS::Admin reference
# @return POS_NO_ERROR on success
sub sync
{
    my ($self) = @_;
    log_debug_print('starting sync');
    if (ref $self ne 'POS::Admin') {
        log_error_print('calling sync out of POS::Admin context');
        return POS_GENERAL_ERROR;
    }
    $self->{'CACHE'}->adjustBase();
    if (not $self->validate($self->{'CACHE'}->getRoot())) {
        # update - go from root to leafes and replace LDAP entries with CACHE entries. At the end parse deletes
        use Net::LDAP::Constant qw(LDAP_ALREADY_EXISTS LDAP_NO_SUCH_OBJECT);
        my @DNs      = ($self->{'CACHE'}->getRoot());
        my @toDelete = ();
        while (@DNs) {
            my $dn = pop @DNs;
            my $res;
            push(@DNs, @{ $self->{'CACHE'}->getChildren($dn) });
            my $entry = $self->{'CACHE'}->getLDAPentry($dn);
            # skip glue entries
            next if not $entry;
            my $changeType = $entry->changetype();
            if ($changeType =~ /add|modify/) {
                $res = $entry->update($self->{'LDAP'});
            }
            elsif ($changeType eq 'delete') {
                push @toDelete, $entry;
                next;
            }
            else {
                log_error_printf('Unknown changetype %s', $changeType);
                next;
            }
            if ($res->is_error()) {
                if ($res->code != LDAP_ALREADY_EXISTS) {
                    log_error_printf('Error updating entry %s', $dn);
                    return POS_LDAP_ERROR;
                }
            }
            else {
                $self->_postLdapWrite($self->{'CACHE'}->getClass($dn), $entry);
                log_info_printf('Entry %s updated in LDAP', $dn);
            }
            $self->{'CACHE'}->setFlag($dn, 'updated');
        }
        while (@toDelete) {
            my $entry = pop @toDelete;
            my $dn    = $entry->dn();
            my $res   = $self->{'LDAP'}->delete($dn);
            if ($res->is_error()) {
                if ($res->code == LDAP_NO_SUCH_OBJECT) {
                    log_info_printf('Entry %s is not present in LDAP, skipping', $dn);
                }
                else {
                    log_error_printf('Error deleting entry %s, delete phase aborted!', $dn);
                    return;
                }
            }
            else {
                $self->_postLdapWrite($self->{'CACHE'}->getClass($dn), $entry);
                log_info_printf('Entry %s deleted from LDAP', $dn);
            }
        }
    }
    else {
        $self->{'ERROR'} = POS_VALIDATION_FAILED;
        return POS_VALIDATION_FAILED;
    }
    log_debug_print('sync completed');
    return POS_NO_ERROR;
} ## end sub sync

sub refreshLdap
{
    use POS::Manager;
    use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);
    my $self = _getRunningMode(shift);
    return if ($self == POS_NO_LDAP);
    my $searchBase;
    my $searchFilter;
    # determine AS or BS and adjust base accordingly
    my $config = $self->{'CONFIG'};
    if ($self->{'CONFIG'}{'POS_ADMIN_DEPLOYED'}) {
        $searchBase   = $self->{'CONFIG'}{'ALL_LDAPBASE'};
        $searchFilter = '(objectClass=*)';
    }
    elsif ($self->{'CONFIG'}{'POS_BRANCH_DEPLOYED'}) {
        $searchBase = $self->{'CONFIG'}{'BRANCH_LDAPBASE'};
        # ignore scLocation and OUs
        $searchFilter = '(&(!(objectClass=scLocation))(!(objectClass=organizationalUnit)))';
    }
    else {
        log_alert_print('Unrecognized SLEPOS deployment, neither AS nor BS detected!');
        return POS_GENERAL_ERROR;
    }
    # precreate SUSE Manager object to avoid recreating it every time
    $self->{'SUSEMANAGER'} = POS::Manager->new(CONFIG => $self->{'CONFIG'});
    log_info_printf('Using seach base "%s" and search filter "%s"', $searchBase, $searchFilter);
    my $ldap = $self->{'LDAP'};
    my $entries = $ldap->search(base => $searchBase, scope => 'sub', filter => $searchFilter);
    if (not($entries) or ($entries->is_error() and $entries->code != LDAP_NO_SUCH_OBJECT)) {
        log_error_printf(
            'LDAP access error during LDAP data refresh: (%u) %s',
            $entries->code()  || 0,
            $entries->error() || 'Unknown error'
        );
        return POS_LDAP_ERROR;
    }
    my $entry;
    while ($entry = $entries->shift_entry()) {
        my $objectClass = $entry->get_value('objectClass');
        # call preLdapWrite for each entry
        $self->_preLdapWrite($objectClass, $entry);
        # if changed, update LDAP
        if ($entry->changes()) {
            my $res = $entry->update($ldap);
            if (not($res) or $res->is_error()) {
                log_error_printf(
                    'Failed to update entry %s: (%u) %s',
                    $entry->dn(),
                    $res->code()  || 0,
                    $res->error() || 'Unknown error'
                );
            }
        }
        # call postLdapWrite for each entry
        $self->_postLdapWrite($objectClass, $entry);
    }
    return POS_NO_ERROR;
} ## end sub refreshLdap

###############################################################################
# VALIDATION
sub _validateEntry
{
    my ($self, $objectClass, $entry) = @_;
    # presume validation error
    $self->{'ERROR'} = POS_VALIDATION_FAILED;
    # validate against ACFL
    use POS::Admin::Validator::ACFL;
    # always do generic check - validates also MUST attributes
    unless (POS::Admin::Validator::ACFL::generic_check($entry)) {
        return POS_VALIDATION_FAILED;
    }
    # if specialized check exists, do it
    if ($ACFL_CHECKLIST{$objectClass}) {
        unless ($ACFL_CHECKLIST{$objectClass}->($entry, $entry->dn(), $self->{'LDAP'})) {
            return POS_VALIDATION_FAILED;
        }
    }
    # all's good, clear error
    $self->clearError();
    return POS_NO_ERROR;
} ## end sub _validateEntry

sub _validateParent
{
    my ($self, $objectClass, $dn) = @_;
    log_debug_printf('Validating parent structure for entry %s', $dn);
    $self->{'ERROR'} = POS_VALIDATION_FAILED;
    # first check if entry does not already exists
    use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);
    my $res = $self->{'LDAP'}->search(base => $dn, filter => '(objectClass=*)', scope => 'base', attrs => ['1.1']);
    if (not $res->code == LDAP_NO_SUCH_OBJECT) {
        log_error_printf('Entry with DN "%s" already exists!', $dn);
        return POS_ALREADY_EXISTS;
    }
    #get parent DN
    my ($parentDN) = $dn =~ /[^,]+,(.*)/;
    # and find it in LDAP
    $res =
      $self->{'LDAP'}
      ->search(base => $parentDN, filter => '(objectClass=*)', scope => 'base', attrs => ['objectClass']);
    if ($res->is_error()) {
        if ($res->code == LDAP_NO_SUCH_OBJECT) {
            log_error_printf('Entry parent (DN: %s) does not exist!', $parentDN);
            return POS_VALIDATION_FAILED;
        }
        else {
            log_error_printf('Error (%s) when searching for parent with DN "%s"', $res->error, $parentDN);
            return POS_VALIDATION_FAILED;
        }
    }
    # get parent`s objectClass
    my @parentObjectClasses = $res->pop_entry()->get_value('objectClass');
    my ($parentObjectClass) = grep { $LDAP_STRUCT{$_} } @parentObjectClasses;
    if (not $parentObjectClass) {
        log_error_printf('Parent (%s) is not SLEPOS object class, check DN!', $parentDN || '');
        return POS_VALIDATION_FAILED;
    }
    log_debug_printf('Trying to add %s under %s', $objectClass, $parentObjectClass);
    # and validate if entry can be placed under parent
    my $subentries = $LDAP_STRUCT{$parentObjectClass}{'subentries'};
    # catch undefined or empty subentries
    if (not defined $subentries or not @$subentries) {
        log_error_printf('No entry can be placed under objectClass "%s"!', $parentObjectClass);
    }
    else {
        foreach (@$subentries) {
            my ($objClass, $quantifier) = @$_;
            if ($objClass eq $objectClass) {
                #found it, ignoring siblings check, leaving that to whole LDAP validate
                $self->clearError();
            }
        }
    }
    if ($self->{'ERROR'} != POS_NO_ERROR) {
        log_error_printf('Object class "%s" cannot be placed under parent "%s". Check entry DN or objectClass!',
            $objectClass, $parentObjectClass);
    }
    # returns POS_NO_ERROR only if passed subentries check, else POS_VALIDATION_FAILED
    log_debug_printf('Entry %s validated with result %u (0 is good)', $dn, $self->{'ERROR'});
    return $self->{'ERROR'};
} ## end sub _validateParent

#validate ldapdata, ldapdata may be Net::LDAP::Search or ref to array of Net::LDAP::Entry objects
sub _validate
{
    my ($self, $base, $rootClass, $relaxed, $ldapData) = @_;
    use POS::Admin::Validator;
    my $validator = POS::Admin::Validator->new(
        CONFIG           => $self->{'CONFIG'},
        ENTRY_CHECK_BASE => $base,
        LDAP_DATA        => $ldapData,
        MERGE_DATA       => $self->{'CACHE'},
        RELAXED          => $relaxed,
        ROOTCLASS        => $rootClass
    );
    if ($validator) {
        log_debug_print('validator object created, validating');
        if ($validator->validate()) {
            if (not defined $self->{'CACHE'}) {
                $validator->postValidation();
                #$validator->printStats();
            }
            return 1;
        }
    }
    return;
} ## end sub _validate
###############################################################################

###############################################################################
# ENTRY HELPERS
## @method _preLdapWrite ($objectClass, $entry)
# executes actions associated with $objectClass before validation and writing to LDAP
# @param $objectClass object class of entry matching POS::Admin::LdapStruct
# @param $entry ref to Net::LDAP::Entry, actual entry data
# @return any result code from POS::Admin::Constant
sub _preLdapWrite
{
    my ($self, $objectClass, $entry) = @_;
    $_ = $objectClass;
    if (/scConfigFileTemplate/) {
        return $self->_copyFileIntoEntry($entry, 'scConfigFileData');
    }
    elsif (/scConfigFileSyncTemplate/) {
        return $self->_linkFileWithEntry($entry, 'scConfigFileLocalPath');
    }
} ## end sub _preLdapWrite

## @method _preLdapWrite ($objectClass, $entry)
# executes actions associated with $objectClass after successful writing to LDAP
# WARNING: when updating this function consider modify behavior in case of moddn operation!
# @param $objectClass object class of entry matching POS::Admin::LdapStruct
# @param $entry ref to Net::LDAP::Entry, actual entry data
# @param $delete specifies whether entry was added/modified or deleted
# @return any result code from POS::Admin::Constant
sub _postLdapWrite
{
    my ($self, $objectClass, $entry, $nosm) = @_;
    my $delete = 0;
    $delete = 1 if $entry->changetype() eq 'delete';
    $_ = $objectClass;
    # ACLs
    if (/scLocation|scRefObjectContainer/) {
        if ($delete) {
            $self->_removeACL($entry->dn(), $objectClass);
        }
        else {
            $self->_addACL($entry->dn(), $objectClass);
        }
    }
    # SUSE Manager
    if (/scLocation|scImage|scHardware|scWorkstation/ && !$nosm) {
        if ($delete) {
            $self->_removeFromManager($entry, $objectClass);
        }
        else {
            $self->_addToManager($entry, $objectClass);
        }
    }
} ## end sub _postLdapWrite

## @method
#
# @param
# @return
sub _copyFileIntoEntry
{
    my ($self, $entry, $attrName) = @_;
    my $filename = $entry->get_value($attrName);
    my $md5sum;
    if (defined $filename) {
        # count new lines to avoid stat error
        if ($filename =~ y/\n// < 2 && -f $filename) {
            use IO::File;
            my $fh = IO::File->new($filename, O_RDONLY);
            if (not defined $fh) {
                log_error_printf('unable to open file "%s" specified by "%s" attribute', $filename, $attrName);
                return POS_FILE_NOT_FOUND;
            }
            # replace the file name by the content of the file
            $entry->replace($attrName => join('', $fh->getlines()));
            $fh->close();
            $md5sum = $self->_createMd5FromEntry($entry, $attrName);
        }
        else {
            log_debug_printf('%s attribute does not point any file, leaving content intact', $attrName);
            $md5sum = $self->_createMd5FromEntry($entry, $attrName);
        }
        if ($entry->exists('scConfigMd5')) {
            $entry->replace('scConfigMd5' => $md5sum);
        }
        else {
            $entry->add('scConfigMd5' => $md5sum);
        }
    }
    else {
        log_error_print('Trying to copy file without defined filename!');
        return POS_FILE_NOT_FOUND;
    }
    return POS_NO_ERROR;
} ## end sub _copyFileIntoEntry

## @method
#
# @param
# @return
sub _linkFileWithEntry
{
    my ($self, $entry, $attrName) = @_;
    my $filename = $entry->get_value($attrName);
    my $md5sum;
    if (defined $filename) {
        if (-f $filename) {
            $md5sum = $self->_createMd5FromFile($filename);
        }
        else {
            log_error_printf('File "%s" does not exist!', $filename);
            return POS_FILE_NOT_FOUND;
        }
        if ($entry->exists('scConfigMd5')) {
            $entry->replace('scConfigMd5' => $md5sum);
        }
        else {
            $entry->add('scConfigMd5' => $md5sum);
        }
    }
    else {
        log_error_printf('Attribute %s does not specify any filename!', $attrName);
        return POS_FILE_NOT_FOUND;
    }
    return POS_NO_ERROR;
} ## end sub _linkFileWithEntry

## @method
#
# @param
# @return
sub _createMd5FromFile
{
    my ($self, $filename) = @_;
    use Digest::MD5;
    use IO::File;
    my $fh = IO::File->new($filename, O_RDONLY);
    if (not defined $fh) {
        log_error_printf('unable to open file "%s" for MD5 sum computation', $filename);
        return POS_FILE_NOT_FOUND;
    }
    $fh->binmode();
    my $md5sum = Digest::MD5->new->addfile($fh)->b64digest;
    $fh->close();
    return $md5sum;
} ## end sub _createMd5FromFile

## @method
#
# @param
# @return
sub _createMd5FromEntry
{
    my ($self, $entry, $attribute) = @_;
    use Digest::MD5;
    my $data = $entry->get_value($attribute);
    if (not defined $data) {
        log_error_printf('supplied entry does not contain any data for attribute "%s", $attribute');
        return POS_NO_DATA;
    }
    my $md5sum = Digest::MD5->new()->add($data)->b64digest();
    return $md5sum;
} ## end sub _createMd5FromEntry
###############################################################################

###############################################################################
# ACL MANAGEMENT
sub _addACL
{
    my ($self, $targetDN, $objectClass) = @_;
    use POS::LdapControl;
    # 0) prepare ACL for given objectClass
    my $newrule       = '';
    my $defaultAccess = '';
    # scLocation is default objectClass if not defined
    $objectClass = 'scLocation' unless (defined $objectClass);
    if ($objectClass eq 'scLocation') {
        $newrule       = "\tby dn.base=\"$targetDN\" write\n";
        $defaultAccess = "\tby users none\n\tby anonymous auth";
    }
    elsif ($objectClass eq 'scRefObjectContainer') {
        # need to get parent DN
        my ($parentDN) = $targetDN =~ /^[^,]*,(.*)/;
        # now check, that parent is organizationalUnit => DN begins with ou
        return unless ($parentDN =~ /^ou=.*/);
        $newrule       = "\tby dn.children=\"$parentDN\" read\n";
        $defaultAccess = "\tby anonymous none";
    }
    # 1) add ACL dynamicaly via olcAccess
    my $ldapcontrol = POS::LdapControl->new('CONFIG' => $self->{'CONFIG'});
    if (defined $ldapcontrol) {
        my $ret = $ldapcontrol->addAcl($targetDN, $objectClass);
        if ($ret != 0) {
            log_error_print 'Applying of ACL failed, check LDAP configuration. See log file for detail.';
            return;
        }
    }
    else {
        log_error_print 'Error opening LDAP connection, check LDAP configuration';
        exit 1;
    }
} ## end sub _addACL

sub _removeACL
{
    my ($self, $targetDN, $objectClass) = @_;
    use POS::LdapControl;
    # 1) add ACL dynamicaly via olcAccess
    my $ldapcontrol = POS::LdapControl->new('CONFIG' => $self->{'CONFIG'});
    if (defined $ldapcontrol) {
        my $ret = $ldapcontrol->delAcl($targetDN, $objectClass);
        if ($ret != 0) {
            log_error_print 'Deleting of ACL failed, check LDAP configuration. See log file for detail.';
            return;
        }
    }
    else {
        log_error_print 'Error opening LDAP connection, check LDAP configuration';
        exit 1;
    }
} ## end sub _removeACL
###############################################################################

###############################################################################
# SUSE Manager
# register and unregister routines for SM integration
sub _addToManager
{
    my ($self, $entry, $objectClass, $sm) = @_;
    use POS::Manager;
    my $dn   = $entry->dn();
    my $name = $dn;
    $sm = $self->{'SUSEMANAGER'} unless $sm;
    $sm = POS::Manager->new(CONFIG => $self->{'CONFIG'}) unless $sm;
    return unless ($sm and $sm->{'SM'});
    $name =~ s/o=|c=|ou=|cn=//;
    $name =~ s/,o=|,c=|,ou=|,cn=/./g;

    if ($objectClass eq 'scLocation') {
        #check if group with this $name already exists:
        my $res = $sm->call('systemgroup.getDetails', $name);
        if ($res and not $res->{'faultCode'}) {
            log_debug_printf('systemgroup %s already registered, skipping systemgroup creation', $name);
        }
        else {
            #add new group to SM
            $res = $sm->call('systemgroup.create', $name, "SLEPOS group $dn");
            if (not $res) {
                log_warn_printf('Creation of systemgroup %s failed, skipping', $name);
                return;
            }
            elsif ($res->{'faultCode'}) {
                log_warn_printf('Creation of systemgroup %s failed: %s:%s, skipping',
                    $name, $res->{'faultCode'}, $res->{'faultString'});
                return;
            }
        }

        #check if user with this $name already exists:
        $res = $sm->call('user.getDetails', $name);
        if ($res and not $res->{'faultCode'}) {
            log_debug_printf('user %s already registered, skipping user account creation', $name);
        }
        else {
            #add new branch user to SM
            my $res = $sm->call('user.create', $name, $entry->get_value('userPassword'),
                'SLEPOS', 'SLEPOS', 'root@' . $sm->{'SMHOST'});
            if (not defined $res) {
                log_error_printf('Creation of user account for %s failed: Unknown error', $dn);
            }
            elsif (ref($res) eq "HASH") {
                log_error_printf('Creation of user account for %s failed: (%s) %s',
                    $dn, $res->{'faultCode'}, $res->{'faultString'});
            }
            else {
                log_debug_printf('User %s created', $name);
            }
        }
        # always try to add user's privileges. User may exists without them
        $sm->call('user.addAssignedSystemGroup', $name, $name, $sm->bool(1));
        $sm->call('user.addRole',                $name, 'system_group_admin');
        $sm->call('user.addRole',                $name, 'activation_key_admin');
    }
    elsif ($objectClass eq 'scImage') {
        #add new software channel to SM
    }
    elsif ($objectClass eq 'scHardware') {
        if ($entry->get_value('scSUSEManager')) {
            # check SLEPOS channels
            foreach my $channel ('POS_SM_CHANNEL_32', 'POS_SM_CHANNEL_64') {
                my $res = $sm->call('channel.software.getDetails', lc $self->{'CONFIG'}{$channel});
                if ($res and not $res->{'faultCode'}) {
                    log_debug_printf('software channel %s already exists', $self->{'CONFIG'}{$channel});
                }
                else {
                    $res = $sm->call(
                        'channel.software.create',
                        lc $self->{'CONFIG'}{$channel},
                        $self->{'CONFIG'}{$channel},
                        'SUSE Linux Enterprise Point Of Service generic software channel',
                        $channel eq 'POS_SM_CHANNEL_32' ? 'channel-ia32' : 'channel-x86_64',
                        '',
                    );
                    if (ref($res) eq 'HASH' and $res->{'faultCode'}) {
                        log_warn_printf('Failed to register generic SLEPOS software channel in SUSE Manager');
                    }
                }
            }
            # check scLocation and scImage registrations
            my $res = $self->{'LDAP'}->search(
                base   => $self->{'LDAP'}{'BASE'},
                filter => '(|(objectClass=scLocation)(objectClass=scImage))',
                scope  => 'sub',
                attrs  => ['objectClass','userPassword']
            );
            if ($res->is_error() and $res->code() != 32) {
                log_error_print('Error while getting list of BranchServers: %s', $res->error());
                return;
            }
            while (my $entry = $res->pop_entry()) {
                $self->_addToManager($entry, grep(/scLocation|scImage/, $entry->get_value('objectClass')), $sm);
            }
        }
    }
} ## end sub _addToManager

sub _removeFromManager
{
    my ($self, $entry, $objectClass) = @_;
    use POS::Manager;
    my $dn   = $entry->dn();
    my $name = $dn;
    my $sm   = POS::Manager->new(CONFIG => $self->{'CONFIG'});
    return unless (defined $sm);
    $name =~ s/o=|c=|ou=|cn=//;
    $name =~ s/,o=|,c=|,ou=|,cn=/./g;

    if ($objectClass eq 'scLocation') {
        #delete group, user from SM
        $sm->call('systemgroup.delete', $name);
        $sm->call('user.delete',        $name);
    }
    elsif ($objectClass eq 'scImage') {
        #delete software channel from SM
    }
    elsif ($objectClass eq 'scWorkstation') {
        #delete workstation from SM
        my $sysIds = $sm->call('system.getId', $name);
        #create id's array:
        my @ida;
        foreach my $sys_hash (@$sysIds) {
            @ida = (@ida, $sys_hash->{'id'});
        }
        $sm->call('system.deleteSystems', @ida);
    }

} ## end sub _removeFromManager
##############################################################################
1;
