package POS::Admin::GUI::Data;

use strict;
use warnings;
# GUI common, includes yui - needed for visual error reporting
use base 'POS::Admin::GUI';
our @ISA = qw(POS::Admin::GUI);

use POS::Admin;
use POS::Admin::Ldap2XML 0.2;
use POS::Syslog;

## STRINGS
my $errorXMLmalformed = 'XML file "%s" is malformed and will be ignored.%sParser output:%s%s';
my $errorBadSleposXML = 'Imported XML is not valid SLEPOS XML file';
my $imageNotFound     = "Cannot find image \"%s\" for CR \"%s\".\n\nCheck exported XML or correct CR image!";
my $roleNotFound      = "Cannot find role \"%s\" for CR \"%s\".\n\n Will use global instead!";
my $saveFileFound =
  "Temporary XML file was found from last application unexpected termination.\n Attempt to restore data?";

## @var const XMLVERSION
# SLEPOS XML version of exported XML.
our constant $XMLVERSION = '0.1';

# SLEPOS XML namespace URI
our constant $posnsURI = 'http://www.suse.com/SLEPOS';
## @var const posnsPX
# SLEPOS XML namespace prefix
our constant $posnsPX = 'slepos';
## @var const xincnsURI
# XInclude XML namespace URI
our constant $xincnsURI = 'http://www.w3.org/2001/XInclude';
## @var const xincnsPX
# XInclude XML namespace prefix
our constant $xincnsPX = 'xi';

sub new
{
    #create class
    my $self = shift;
    my $type = ref($self) || $self;
    $self = bless {}, $type;
    $self->{'FACTORY'} = shift;
    $self->{'ENV'}     = shift;
    $self->{'TMPFILE'} = '.tmpSLEPOS.xml';
    $self->{'CACHE'}   = undef;
    return $self;
} ## end sub new

sub initData
{
    my ($self, @args) = @_;
    # POS::Config load
    my $config;
    eval {
        # try to load POS::Config if present
        require POS::Config;
        POS::Config->import();
        $config = POS::Config->new();
    };
    if ($@) {
        # POS packages are either not installed installed, or neither AS nor BS are configured
        $self->{'ONADMIN'} = 0;
        $config = { 'ALL_LDAPBASE' => 'o=myorg,c=us' };
    }
    else {
        $self->{'ONADMIN'} = 1
          if (defined $config->{'POS_ADMIN_DEPLOYED'} and $config->{'POS_ADMIN_DEPLOYED'} eq 'yes');
        # store admin info
        $self->{'ADMIN'}   = $config->{'POS_ADMIN_DN'};
        $self->{'ADMINPW'} = $config->{'POS_ADMIN_PASSWORD'};
    }
    #basic doc
    my $indoc;
    # 1 )check tmp file
    if (-f $self->{'TMPFILE'}) {
        if ($self->yesnoPopup($saveFileFound)) {
            $args[0] = $self->{'TMPFILE'};
        }
    }
    # 2) then import
    if (scalar @args > 0) {
        # here goes XML import and defaults loading
        my $parser = POS::Admin::Ldap2XML->new(CONFIG => $config);
        eval { ($indoc, undef) = $parser->parseXML($args[0], 1) };
        unless ($indoc) {
            $self->okPopup(sprintf($errorXMLmalformed, $args[0], "\n", "\n", $@));
            undef $indoc;
        }
    }
    my $outAdmin = POS::Admin->new(CONFIG => $config, LDAP => 'noneNeeded');
    my $outCache = $outAdmin->{'CACHE'};
    # disable changes_only mode on cache
    $outCache->{'CHANGES_ONLY'} = 0;
    if (defined $indoc) {
        my $inCache = $indoc->{'CACHE'};
        # set correct $base
        $inCache->adjustBase();
        # update ALL_LDAPBASE first in case it differs
        ($config->{'ALL_LDAPBASE'}) = $inCache->getRoot() =~ /(o=[^,]+,c=.+)$/;
        if ($inCache->getRoot() ne $config->{'ALL_LDAPBASE'}) {
            # input doc does not start on org level, add our organization to create proper starting point for merge
            my ($o) = $config->{'ALL_LDAPBASE'} =~ /o=([^,]+)/;
            $outAdmin->add($config->{'ALL_LDAPBASE'}, 'organization', { o => $o });
        }
        # merge input to output
        my @children = ($inCache->getRoot());
        while (@children) {
            my $child = shift @children;
            push @children, @{ $inCache->getChildren($child) };
            $outCache->mergeEntryToCache($inCache->getLDAPentry($child));
        }
    }
    else {
        # create empty OUTDOC with defaults from POS::Admin
        my ($o) = $config->{'ALL_LDAPBASE'} =~ /o=([^,]+)/;
        $outAdmin->add($config->{'ALL_LDAPBASE'}, 'organization', { o => $o });
    }
    if (not $outCache->{ 'cn=global,' . $config->{'ALL_LDAPBASE'} }) {
        $outAdmin->add('cn=global,' . $config->{'ALL_LDAPBASE'}, 'scRefObjectContainer', { cn => 'global' });
    }
    $self->{'OUTDOC'} = $outAdmin;
} ## end sub initData

################################################
### update Helpers
#########################
# update is quite tricky and can be time consuming
# mainly renames of org, ous, locations, roles need almost full tree traversal
# branchserver container is recreated from scratch
sub updateBranchServer
{
    my ($self, $bsinfo, $originalDN) = @_;
    my $doc   = $self->{'OUTDOC'};
    my $admin = $self->{'OUTDOC'};
    my $cache = $admin->{'CACHE'};
    # announce multiple changes, so that LDS tree is not recalculated after each operation
    $cache->startMultipleChanges();
    ## refresh org&country
    my ($ruleO, $ruleOU, $ruleLoc, $ruleRole);
    my $dn = "o=$bsinfo->{'organization'},c=$bsinfo->{'country'}";
    if ($cache->getRoot() ne $dn) {
        # organization and/or country changed
        $cache->getLDAPentry($cache->getRoot())->dn($dn);
    }
    ## refresh ous
    # following procedure descent from top OU to lowest OU for given branchserver.
    # Note this only work in current version of front UI as we assume only one branchserver
    # can be edited.
    # procedure creates new OU chain and move location under it, then deletes not used DNs
    my @oldOUs = $originalDN =~ /ou=([^,]+)+/g;
    @oldOUs = reverse @oldOUs;
    my @curOUs = @{ $bsinfo->{'organizationalUnit'} };
    # init DNs variables to root dn
    my $curDN         = $dn;
    my $oldDN         = $dn;
    my $deleteSinceDN = '';
    for (my $i = 0; $i < scalar @curOUs; ++$i) {
        $curDN = "ou=$curOUs[$i],$curDN";
        if ($cache->getLDAPentry($curDN)) {
            # if we match this, OU already existed and on same position
            $oldDN = $curDN;
        }
        else {
            $admin->add($curDN, 'organizationalUnit', { 'ou' => $curOUs[$i] });
            if (not $deleteSinceDN) {
                $deleteSinceDN = "ou=$oldOUs[$i],$oldDN";
            }
        }
    }
    # TODO: this works only in current UI state. When multiple BSs are enabled in UI, this needs to be removed
    if ($deleteSinceDN) {
        $admin->remove($deleteSinceDN, 1);
    }
    ## refresh scLocation
    $dn = 'cn=' . $bsinfo->{'scLocation'} . ',' . $curDN;

    if ($originalDN ne $dn) {
        $cache->moveEntry($originalDN, $curDN, 'cn=' . $bsinfo->{'scLocation'});
    }
    my $localDN = $dn;
    #update scLocation attributes
    $admin->modify($dn, 'scLocation', $bsinfo->{'LOCATION'});
    #update scServerContainer
    my $container = shift @{ $cache->getChildrenByClass($dn, 'scServerContainer') };
    if ($container) {
        # update existing
        my ($dnPart) = $container =~ /^cn=([^,]+),/;
        if ($dnPart ne $bsinfo->{'scServerContainer'}) {
            $cache->getLDAPentry($container)->dn("cn=$dnPart,$dn");
        }
        $dn = $container;
    }
    else {
        # create new
        $dn = "cn=$bsinfo->{'scServerContainer'},$dn";
        $admin->add($dn, 'scServerContainer', { 'cn' => $bsinfo->{'scServerContainer'} });
    }

    #update scBranchServer
    my $server = shift @{ $cache->getChildrenByClass($dn, 'scBranchServer') };
    if ($server) {
        # update existing
        my ($dnPart) = $server =~ /^cn=([^,]+),/;
        if ($dnPart ne $bsinfo->{'scBranchServer'}) {
            $cache->getLDAPentry($server)->dn("cn=$dnPart,$dn");
        }
        $dn = $server;
    }
    else {
        # create new
        $dn = "cn=$bsinfo->{'scBranchServer'},$dn";
        $admin->add($dn, 'scBranchServer', { 'cn' => $bsinfo->{'scBranchServer'} });
    }

    #update scServices
    # some services may be deleted, some added, some renamed. Delete all and recreate them
    foreach ($cache->getChildren($dn)) {
        $admin->remove($_, 1);
    }
    foreach my $servName (keys %{ $bsinfo->{'SERVICES'} }) {
        $bsinfo->{'SERVICES'}{$servName}{'cn'} = $servName;
        $admin->add("cn=$servName,$dn", 'scService', $bsinfo->{'SERVICES'}{$servName});
    }
    #update scNetworkcards
    # already deleted during service delete
    foreach my $nicName (keys %{ $bsinfo->{'NICS'} }) {
        $bsinfo->{'NICS'}{$nicName}{'scDevice'} = $nicName;
        $admin->add("scDevice=$nicName,$dn", 'scNetworkcard', $bsinfo->{'NICS'}{$nicName});
    }

    ## refresh roles
    # rename and update, if deleted, move cash registers under global/local
    $dn = "cn=global," . $cache->getRoot();
    my @globalRoles = $cache->getChildrenByClass($dn, 'scRole');

    my ($localContainer) = $cache->getChildrenByClass($localDN, 'scRefObjectContainer');
    if (not $localContainer) {
        $localDN = "cn=local,$localDN";
        $admin->add($localDN, 'scRefObjectContainer', { 'cn' => 'local' });
    }
    else {
        $localDN = $localContainer;
    }
    my @localRoles = $cache->getChildrenByClass($localDN, 'scRole');
    my @roleList = (@globalRoles, @localRoles);
    # add roles from $bsinfo
    foreach my $roleName (keys %{ $bsinfo->{'ROLES'} }) {
        # get global or location element
        # TODO: when/if regional suport is added, this has to be updated
        my $parent;
        my $role = $bsinfo->{'ROLES'}{$roleName};
        if ($role->{'ROLE'} eq 'local') {
            $parent = $localDN;
        }
        elsif ($role->{'ROLE'} eq 'global') {
            $parent = $dn;
        }
        else {
            die 'Assert error, only local/global supported';
        }
        my $expectedDN = "cn=$roleName,$parent";
        # role already exist in Cache if already has DN
        if ($role->{'DN'}) {
            # at first update all
            $self->updatePOSattributes($role->{'DN'}, $role, qw(ROLE|TEMPLATES|DN));
            $self->updateFileTemplates($role->{'DN'}, $role);
            # than we need to check if role wasn't moved/renamed thats if its $name and $parent match its dn
            if ($role->{'DN'} ne $expectedDN) {
                my ($oldName)   = $role->{'DN'} =~ /^cn=([^,]+)/;
                my ($oldParent) = $role->{'DN'} =~ /^cn=[^,]+,(.+)$/;
                if ($parent ne $oldParent) {
                    $cache->moveEntry($role->{'DN'}, $parent);
                }
                if ($oldName ne $roleName) {
                    #rename
                    $cache->renameEntry($role->{'DN'}, $expectedDN);
                }
            }
            # remove from $roleList so it won't be deleted from cache on cleanup
            # use old $role->{'DN'}, $expectedDN is not in the list when moving/renaming
            my $i = 0;
            ++$i until $roleList[$i] eq $role->{'DN'};
            splice(@roleList, $i, 1);
        }
        else {
            # this is new or recreated role, lets find out which one
            if (not $cache->getLDAPentry($expectedDN)) {
                # create new
                my %attrs = %{ $bsinfo->{'ROLES'}{$roleName} };
                delete $attrs{'ROLE'};
                delete $attrs{'TEMPLATES'};
                $roleName =~ s/\s+/_/g;
                $attrs{'cn'} = $roleName;
                $admin->add($expectedDN, 'scRole', { 'cn' => $roleName, 'scRoleName' => $roleName });
            }
            $self->updatePOSattributes($expectedDN, $role, qw(ROLE|TEMPLATES|DN));
            $self->updateFileTemplates($expectedDN, $role);
            # role is not in role list, no need to splice
        }
    }
    #delete remaining
    foreach (@roleList) {
        # do we have cashregister?
        my @crs = $cache->getChildrenByClass($_, 'scCashRegister');
        foreach (@crs) {
            # move them under global/local
            if ($_ =~ /$localDN/) {
                $cache->moveEntry($_, $localDN);
            }
            elsif ($_ =~ /$dn/) {
                $cache->moveEntry($_, $dn);
            }
        }
        $admin->remove($_, 1);
    }
    ## update structual changes and sanitize
    $cache->endMultipleChanges();
    ## export to _tmpfile
    $self->exportXML();
} ## end sub updateBranchServer

sub updateImages
{
    my ($self, $iminfo, $branchDN) = @_;
    # no defaults
    my $admin = $self->{'OUTDOC'};
    my $cache = $admin->{'CACHE'};
    $cache->startMultipleChanges();
    # get list of images
    my @registeredImages = $self->getImageList($branchDN);
    # get local DN
    my $localDN = 'cn=default,cn=local,' . $branchDN;
    if (not $cache->getLDAPentry($localDN)) {
        $admin->add($localDN, 'scDistributionContainer', { 'cn' => 'default' });
    }
    # get global DN
    my $globalDN = 'cn=default,cn=global,' . $cache->getRoot();
    if (not $cache->getLDAPentry($globalDN)) {
        $admin->add($globalDN, 'scDistributionContainer', { 'cn' => 'default' });
    }
    # TODO: if support for multiple kernels is added, container selection must be reworked
    foreach (keys %$iminfo) {
        # get global or location element
        # TODO: when/if regional suport is added, this has to be updated
        my $parent;
        my $image     = $iminfo->{$_};
        my $imageName = $image->{'scImageName'};
        if ($image->{'ROLE'} eq 'local') {
            $parent = $localDN;
        }
        elsif ($image->{'ROLE'} eq 'global') {
            $parent = $globalDN;
        }
        else {
            die 'Assert error, only local/global supported';
        }
        # convert 0/1 to TRUE/FALSE
        foreach (keys %{ $image->{'VERSIONS'} }) {
            if ($image->{'VERSIONS'}{$_}{'scDisabled'} eq '1') {
                $image->{'VERSIONS'}{$_}{'scDisabled'} = 'TRUE';
            }
            elsif ($image->{'VERSIONS'}{$_}{'scDisabled'} eq '0') {
                $image->{'VERSIONS'}{$_}{'scDisabled'} = 'FALSE';
            }
        }
        my $expectedDN = "cn=$imageName,$parent";
        # image already exist in Cache if already has DN
        if ($image->{'DN'}) {
            # at first update all
            $self->updatePOSattributes($image->{'DN'}, $image, qw(ROLE|TEMPLATES|DN|VERSIONS));
            $self->updateFileTemplates($image->{'DN'}, $image);
            # than we need to check if image wasn't moved/renamed thats if its $name and $parent match its dn
            if ($image->{'DN'} ne $expectedDN) {
                my ($oldName)   = $image->{'DN'} =~ /^cn=([^,]+)/;
                my ($oldParent) = $image->{'DN'} =~ /^cn=[^,]+,(.+)$/;
                if ($parent ne $oldParent) {
                    $cache->moveEntry($image->{'DN'}, $parent);
                }
                if ($oldName ne $imageName) {
                    #rename
                    $cache->renameEntry($image->{'DN'}, $expectedDN);
                }
            }
            # get registered versions list
            my @registeredVersions = $cache->getChildrenByClass($expectedDN, 'scImageVersion');
            # store version
            foreach my $version (keys %{ $image->{'VERSIONS'} }) {
                my $versionDN = 'scVersion=' . $version . ',' . $expectedDN;
                if (not $cache->getLDAPentry($versionDN)) {
                    $admin->add($versionDN, 'scImageVersion', { 'scVersion' => $version });
                }
                $self->updatePOSattributes($versionDN, $image->{'VERSIONS'}{$version});
                my $i = 0;
                ++$i until $registeredVersions[$i] eq $versionDN;
                splice(@registeredVersions, $i, 1);
            }
            foreach (@registeredVersions) {
                $admin->remove($_);
            }
            # remove from $imageList so it won't be deleted from cache on cleanup
            # use old $image->{'DN'}, $expectedDN is not in the list when moving/renaming
            my $i = 0;
            ++$i until $registeredImages[$i] eq $image->{'DN'};
            splice(@registeredImages, $i, 1);
        }
        else {
            # this is new or recreated image, lets find out which one
            if (not $cache->getLDAPentry($expectedDN)) {
                # create new
                my %attrs = %{$image};
                delete $attrs{'ROLE'};
                delete $attrs{'TEMPLATES'};
                delete $attrs{'VERSIONS'};
                delete $attrs{'DN'};
                $imageName =~ s/\s+/_/g;
                $attrs{'cn'} = $imageName;
                $admin->add($expectedDN, 'scPosImage', \%attrs);
            }
            $self->updatePOSattributes($expectedDN, $image, qw(ROLE|TEMPLATES|DN|VERSIONS));
            $self->updateFileTemplates($expectedDN, $image);
            # store version
            foreach my $version (keys %{ $image->{'VERSIONS'} }) {
                my $versionDN = 'scVersion=' . $version . ',' . $expectedDN;
                $admin->add($versionDN, 'scImageVersion', { 'scVersion' => $version });
                $self->updatePOSattributes($versionDN, $image->{'VERSIONS'}{$version});
            }
        }
    }
    #delete remaining
    foreach (@registeredImages) {
        $admin->remove($_, 1);
    }
    #check if dist containers contain any image. if not, delete them to comply with XML schema
    if (not @{ $cache->getChildren($globalDN) }) {
        $admin->remove($globalDN);
    }
    if (not @{ $cache->getChildren($localDN) }) {
        $admin->remove($localDN);
    }
    $cache->endMultipleChanges();
    $self->exportXML();
} ## end sub updateImages

sub updateCashRegisters
{
    my ($self, $crinfo, $branchDN) = @_;
    # no defaults
    my $admin = $self->{'OUTDOC'};
    my $cache = $admin->{'CACHE'};
    $cache->startMultipleChanges();
    # get list of CRs
    my @registeredCRs = $self->getCashRegistersList($branchDN);
    # get global DN
    my $globalDN = 'cn=global,' . $cache->getRoot();
    # get local DN
    my ($localDN) = $cache->getChildrenByClass($branchDN, 'scRefObjectContainer');
    # TODO: if regional support is added, container selection must be reworked
    # each CR from crinfo compare with registered CRs
    foreach (keys %$crinfo) {
        ## determine base element -> local/global/role
        my $parent;
        my $cr     = $crinfo->{$_};
        my $crName = $cr->{'scCashRegisterName'};
        if ($cr->{'ROLE'} eq 'local') {
            $parent = $localDN;
        }
        elsif ($cr->{'ROLE'} eq 'global') {
            $parent = $globalDN;
        }
        else {
            $parent = $cr->{'ROLE'};
        }
        my $expectedDN = sprintf('cn=%s,%s', $crName, $parent);
        if ($cr->{'DN'}) {
            # cr already exist in cache
            # at first update all
            $self->updatePOSattributes($cr->{'DN'}, $cr, qw(ROLE|TEMPLATES|DN|DISKS));
            $self->updateFileTemplates($cr->{'DN'}, $cr);
            # than we need to check if cr wasn't moved/renamed thats if its $name and $parent match its dn
            if ($cr->{'DN'} ne $expectedDN) {
                my ($oldName)   = $cr->{'DN'} =~ /^cn=([^,]+)/;
                my ($oldParent) = $cr->{'DN'} =~ /^cn=[^,]+,(.+)$/;
                if ($parent ne $oldParent) {
                    $cache->moveEntry($cr->{'DN'}, $parent);
                }
                if ($oldName ne $crName) {
                    $cache->renameEntry($cr->{'DN'}, $expectedDN);
                }
            }
            $self->updateDrives($expectedDN, $cr);
            # remove from crList so it won't be deleted from cache on cleanup
            # use old $cr->{'DN'}, $expectedDN is not in the list when moving/renaming
            my $i = 0;
            ++$i until $registeredCRs[$i] eq $cr->{'DN'};
            splice(@registeredCRs, $i, 1);
        }
        else {
            # this is new or recreated cr, lets find out which one
            if ($cache->getLDAPentry($expectedDN)) {
                $self->updatePOSattributes($expectedDN, $cr, qw(ROLE|TEMPLATES|DN|DISKS));
            }
            else {
                # create new
                my %attrs = %{$cr};
                delete $attrs{'ROLE'};
                delete $attrs{'TEMPLATES'};
                delete $attrs{'DISKS'};
                delete $attrs{'DN'};
                $crName =~ s/\s+/_/g;
                $attrs{'cn'} = $crName;
                $admin->add($expectedDN, 'scCashRegister', \%attrs);
            }
            $self->updateFileTemplates($expectedDN, $cr);
            $self->updateDrives($expectedDN, $cr);
        }
    }
    #delete remaining
    foreach (@registeredCRs) {
        $admin->remove($_, 1);
    }
    $cache->endMultipleChanges();
    $self->exportXML();
} ## end sub updateCashRegisters

sub updateDrives
{
    my ($self, $parent, $cr) = @_;
    my $admin = $self->{'OUTDOC'};
    my $cache = $admin->{'CACHE'};
    # get HDD/RDD versions list
    my @registeredDrives =
      ($cache->getChildrenByClass($parent, 'scHarddisk'), $cache->getChildrenByClass($parent, 'scRamDisk'));
    foreach my $driveCN (keys %{ $cr->{'DISKS'} }) {
        my %drive = %{ $cr->{'DISKS'}{$driveCN} };
        my $type  = delete $drive{'TYPE'};
        my $parts = delete $drive{'PARTS'};
        delete $drive{'DN'};
        my $driveDN = sprintf('cn=%s,%s', $driveCN, $parent);
        $drive{'cn'} = $driveCN;
        # remove empty fields to avoid clashing with xml schema
        foreach (keys %drive) {
            unless ($drive{$_}) {
                delete $drive{$_};
            }
        }
        if ($cache->getLDAPentry($driveDN)) {
            $self->updatePOSattributes($driveDN, \%drive);
            my $i = 0;
            ++$i until $registeredDrives[$i] eq $driveDN;
            splice(@registeredDrives, $i, 1);
        }
        else {
            # create new disk
            $admin->add($driveDN, $type, \%drive);
        }
        if ($type eq 'scHarddisk') {
            # parse partitions
            my @registeredPartitions = $cache->getChildrenByClass($driveDN, 'scPartition');
            foreach my $partition (values %$parts) {
                my $partitionDN = sprintf('scPartNum=%s,%s', $partition->{'scPartNum'}, $driveDN);
                delete $partition->{'DN'};
                if ($cache->getLDAPentry($partitionDN)) {
                    $self->updatePOSattributes($partitionDN, $partition);
                    my $i = 0;
                    ++$i until $registeredPartitions[$i] eq $partitionDN;
                    splice(@registeredPartitions, $i, 1);
                }
                else {
                    $admin->add($partitionDN, 'scPartition', $partition);

                }
            }
            foreach (@registeredPartitions) {
                $admin->remove($_);
            }
        }
    }
    foreach (@registeredDrives) {
        $admin->remove($_, 1);
    }
} ## end sub updateDrives

sub updatePOSattributes
{
    my ($self, $parent, $dataHash, $limit) = @_;
    my %dataHash = %$dataHash;
    if (defined $limit) {
        foreach (keys %dataHash) {
            delete $dataHash{$_} if (not defined $dataHash{$_});
            delete $dataHash{$_} if ($_ =~ /$limit/);
        }
    }
    $self->{'OUTDOC'}->modify($parent, $self->{'OUTDOC'}{'CACHE'}->getClass($parent), \%dataHash);
} ## end sub updatePOSattributes

sub updateFileTemplates
{
    my ($self, $parent, $data) = @_;
    my $admin = $self->{'OUTDOC'};
    my $cache = $admin->{'CACHE'};
    # delete old
    my @oldTemplates = (
        $cache->getChildrenByClass($parent, 'scConfigFileTemplate'),
        $cache->getChildrenByClass($parent, 'scConfigFileSyncTemplate'),
        $cache->getChildrenByClass($parent, 'scPxeFileTemplate')
    );
    # add new
    if (defined $data->{'TEMPLATES'}) {
        foreach my $template (keys %{ $data->{'TEMPLATES'} }) {
            my $dn   = 'cn=' . $template . ',' . $parent;
            my $role = $data->{'TEMPLATES'}{$template}{'ROLE'};
            delete $data->{'TEMPLATES'}{$template}{'ROLE'};
            $data->{'TEMPLATES'}{$template}{'cn'} = $template;
            ## create scPxeFileTemplate
            if ($role eq 'PT') {
                $admin->add($dn, 'scPxeFileTemplate', $data->{'TEMPLATES'}{$template});
            }
            ## create scConfigFileTemplate
            elsif ($role eq 'FT') {
                $admin->add($dn, 'scConfigFileTemplate', $data->{'TEMPLATES'}{$template});
            }
            ## create scConfigFileSyncTemplate
            elsif ($role eq 'FST') {
                $admin->add($dn, 'scConfigFileSyncTemplate', $data->{'TEMPLATES'}{$template});
            }
        }
    }
} ## end sub updateFileTemplates
#########################
### EOF update Helpers
################################################

################################################
### get/import Helpers
#########################
## @method [ARRAY or ARRAYref] getBranchServerList (void)
# get list of DNs of all known branchservers
# parse OUTDOC and return all known branchservers
# @return in scalar context ARRAYref else ARRAY
sub getBranchServerList
{
    my ($self) = @_;
    my $admin  = $self->{'OUTDOC'};
    my $cache  = $admin->{'CACHE'};
    # get OUs
    my @locations = ();
    my @todo      = ($cache->getRoot());
    while (@todo) {
        my $dn    = shift @todo;
        my $class = $cache->getClass($dn);
        if ($class eq 'organization') {
            # get all 1. level OUs
            push(@todo, $cache->getChildrenByClass($dn, 'organizationalUnit'));
        }
        elsif ($class eq 'organizationalUnit') {
            push(@todo,      $cache->getChildrenByClass($dn, 'organizationalUnit'));
            push(@locations, $cache->getChildrenByClass($dn, 'scLocation'));
        }
    }
    if (wantarray) {
        return @locations;
    }
    return \@locations;
} ## end sub getBranchServerList

## @method HASHref getBranchServer ($dn)
# get details of branchserver with DN $dn. If $dn is not provided or does not points to scLocation, new BS is created
# @return HASHref with branchserver details
sub getBranchServer
{
    my ($self, $dn) = @_;
    my $bsinfo = {};
    my $doc    = $self->{'OUTDOC'};
    my $cache  = $self->{'OUTDOC'}{'CACHE'};
    # find all scLocations - if there is only one - select it and get the ous path from dn
    $dn = $cache->getRoot() unless ($dn);

    # org and country is in every DN
    ($bsinfo->{'organization'}, $bsinfo->{'country'}) = $dn =~ /o=([^,]+),c=(.+)$/;
    if ($cache->getLDAPentry($dn) && $cache->getClass($dn) ne 'scLocation') {
        log_notice_print('Creating default BranchServer object');
        # no scLocation DN was provided, create new BS with defaults
        $dn = 'ou=myorgunit,' . $dn;
        $doc->add($dn, 'organizationalUnit', { ou => 'myorgunit' });
        $dn = 'cn=mybranch,' . $dn;
        $doc->add($dn, 'scLocation', { userPassword => 'branchpass', cn => 'mybranch' });
    }
    # get ous
    my @ous = $dn =~ /ou=([^,]+)/g;
    # need to reverse so ous are displayed in proper order
    @ous                            = reverse @ous;
    $bsinfo->{'organizationalUnit'} = \@ous;
    $bsinfo->{'LOCATION'}           = $self->getPOSattributes($dn);
    ($bsinfo->{'LOCATION'}{'scLocation'}) = $dn =~ /^cn=([^,]+)/;
    my @tmp = $cache->getChildrenByClass($dn, 'scServerContainer');
    my $tmp;
    if (scalar @tmp == 1) {
        $tmp = shift @tmp;
        $tmp =~ /^cn=([^,]+)/;
        $bsinfo->{'scServerContainer'} = $1;
    }
    @tmp = $cache->getChildrenByClass($tmp, 'scBranchServer');
    if (scalar @tmp == 1) {
        $tmp = shift @tmp;
        ## get BS hostname
        $tmp =~ /^cn=([^,]+)/;
        $bsinfo->{'LOCATION'}{'scBranchServer'} = $1;
        ## get services
        @tmp = $cache->getChildrenByClass($tmp, 'scService');
        foreach my $service (@tmp) {
            $service =~ /^cn=([^,]+)/;
            $bsinfo->{'SERVICES'}{$1} = $self->getPOSattributes($service);
        }
        ## get network
        @tmp = $cache->getChildrenByClass($tmp, 'scNetworkcard');
        foreach my $nic (@tmp) {
            $nic =~ /scDevice=([^,]+)/;
            $bsinfo->{'NICS'}{$1} = $self->getPOSattributes($nic);
        }
    }

    #fill missing by defaults
    ## check data, fill defaults not provided by POS::Admin
    if (not defined $bsinfo->{'LOCATION'}{'scBranchServer'}) {
        $bsinfo->{'LOCATION'}{'scBranchServer'} = 'bs';
    }
    if (not defined $bsinfo->{'scServerContainer'}) {
        $bsinfo->{'scServerContainer'} = 'container';
    }
    ### services ###
    # services defs are not provided by POS::Admin so lets fill it here
    if (not defined $bsinfo->{'SERVICES'}) {
        # services missing, load them
        ## dhcp
        $bsinfo->{'SERVICES'}{dhcp} = {
            scServiceName        => 'dhcp',
            scDnsName            => 'dhcp',
            scServiceStartScript => 'dhcpd',
            scServiceStatus      => 'TRUE',
            scDhcpDynLeaseTime   => 300,
            scDhcpFixedLeaseTime => 14400
        };
        ## dns
        $bsinfo->{'SERVICES'}{dns} =
          { scServiceName => 'dns', scDnsName => 'dns', scServiceStartScript => 'named', scServiceStatus => 'TRUE' };
        ## tftp
        $bsinfo->{'SERVICES'}{tftp} =
          { scServiceName => 'tftp', scDnsName => 'tftp', scServiceStartScript => 'atftpd', scServiceStatus => 'TRUE' };
        ## posLeases
        $bsinfo->{'SERVICES'}{posleases} = {
            scServiceName        => 'posleases',
            scDnsName            => 'posleases',
            scServiceStartScript => 'posleases2ldap',
            scServiceStatus      => 'TRUE',
            scPosleasesTimeout   => 0,
            scPosleasesChecktime => 10,
            scPosleasesMaxNotify => 6
        };
        ## posASWatch
        $bsinfo->{'SERVICES'}{posASWatch} = {
            scServiceName        => 'posASWatch',
            scDnsName            => 'posASWatch',
            scServiceStartScript => 'posASWatch',
            scServiceStatus      => 'TRUE'
        };
    }
    ### nics ###
    if (not defined $bsinfo->{'NICS'}) {
        # NICs missing, add one
        $bsinfo->{'NICS'}{eth1} = { ipHostNumber => '0.0.0.0' };
    }
    # parse roles
    $bsinfo->{'ROLES'} = $self->getAllRoles($dn);
    return $bsinfo;
} ## end sub getBranchServer

## @method [ARRAY or ARRAYref] getRoleList ($dn)
# get list of DNs of all known roles
# parse OUTDOC and return all known roles
# @return in scalar context ARRAYref else ARRAY
sub getRoleList
{
    my ($self, $localBase) = @_;
    my $cache = $self->{'OUTDOC'}{'CACHE'};
    my @roles = ();
    my @todo  = ('cn=global,' . $cache->getRoot());
    if ($localBase) {
        # expand $localBase to location and individual OUs
        while ($localBase ne $cache->getRoot()) {
            push(@todo, $localBase);
            ($localBase) = $localBase =~ /^[^,]+,(.+)$/;
        }
    }
    while (@todo) {
        my $dn    = shift @todo;
        my $class = $cache->getClass($dn);
        if ($class eq 'scRefObjectContainer') {
            push(@roles, $cache->getChildrenByClass($dn, 'scRole'));
        }
        elsif ($class eq 'organizationalUnit') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRefObjectContainer'));
        }
        elsif ($class eq 'scLocation') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRefObjectContainer'));
        }
    }
    if (wantarray) {
        return @roles;
    }
    return \@roles;
} ## end sub getRoleList

sub getOneRole
{
    my ($self, $roleDN) = @_;
    my $role     = $self->getPOSattributes($roleDN);
    my $locality = 'global';
    if ($roleDN =~ /^cn=[^,]+,cn=[^,]+,cn=[^,]+,/) {
        # only scRoles under scLocation begins with tripple cn=
        $locality = 'local';
    }
    $role->{'ROLE'} = $locality;
    return $role;
} ## end sub getOneRole

sub getAllRoles
{
    my ($self, $localBase) = @_;
    my $roles    = {};
    my @roleList = $self->getRoleList($localBase);
    foreach my $roleDN (@roleList) {
        $roles->{$roleDN} = $self->getOneRole($roleDN);
    }
    return $roles;
} ## end sub getAllRoles

sub getImageList
{
    my ($self, $branchBase) = @_;
    my @images = ();
    my $cache  = $self->{'OUTDOC'}{'CACHE'};
    my $dn     = 'cn=global,' . $cache->getRoot();
    ## get global images
    # get dist containers
    my @containers = $cache->getChildrenByClass($dn, 'scDistributionContainer');
    # get images from all dist containers
    foreach (@containers) {
        foreach ($cache->getChildrenByClass($_, 'scPosImage')) {
            push @images, $_;
        }
    }
    ## get local images
    # get scRefContainer
    @containers = $cache->getChildrenByClass($branchBase, 'scRefObjectContainer');
    if (@containers) {
        $branchBase = $containers[0];
    }
    # get dist containers
    @containers = $cache->getChildrenByClass($branchBase, 'scDistributionContainer');
    # get images from all dist containers
    foreach (@containers) {
        foreach ($cache->getChildrenByClass($_, 'scPosImage')) {
            push @images, $_;
        }
    }
    if (wantarray) {
        return @images;
    }
    return \@images;
} ## end sub getImageList

sub getImages
{
    my ($self, $localBase) = @_;
    my $iminfo = {};
    my $cache  = $self->{'OUTDOC'}{'CACHE'};
    ## if localBase, get local images
    my @refContainers = $cache->getChildrenByClass($localBase, 'scRefObjectContainer');
    foreach (@refContainers) {
        # get dist containers
        my @distContainers = $cache->getChildrenByClass($_, 'scDistributionContainer');
        foreach (@distContainers) {
            # get images
            my @images = $cache->getChildrenByClass($_, 'scPosImage');
            foreach (@images) {
                my $image = $self->getOneImage($_, 'local');
                $iminfo->{"$image->{'scImageName'};local"} = $image;
            }
        }
    }
    ## get global images
    # get dist containers
    my @distContainers = $cache->getChildrenByClass('cn=global,' . $cache->getRoot(), 'scDistributionContainer');
    foreach (@distContainers) {
        # get images
        my @images = $cache->getChildrenByClass($_, 'scPosImage');
        foreach (@images) {
            my $image = $self->getOneImage($_, 'global');
            $iminfo->{"$image->{'scImageName'};global"} = $image;
        }
    }
    return $iminfo;
} ## end sub getImages

sub getOneImage
{
    my ($self, $imageDN, $globality) = @_;
    # get generic attributes
    my $result = $self->getPOSattributes($imageDN);
    $result->{'ROLE'} = $globality;
    ($result->{'cn'}) = $imageDN =~ /^cn=([^,]+)/;
    # get filetemplates
    $self->getFileTemplates($imageDN, $result);
    # get versions
    delete $result->{'scPosImageVersion'};
    ## parse old version string
    foreach my $vs ($self->{'OUTDOC'}{'CACHE'}->getLDAPentry($imageDN)->get_value('scPosImageVersion')) {
        my ($ver, $status) = split(/;/, $vs);
        $result->{'VERSIONS'}{$ver}{'scDisabled'} = $status eq 'active' ? 'FALSE' : 'TRUE';
    }
    ## parse new version subobjects
    foreach my $vs ($self->{'OUTDOC'}{'CACHE'}->getChildrenByClass($imageDN, 'scImageVersion')) {
        my $version = $self->getPOSattributes($vs);
        $result->{'VERSIONS'}{ $version->{'scVersion'} } =
          { 'scDisabled' => $version->{'scDisabled'}, 'scPassword' => $version->{'scPassword'} };
    }
    return $result;
} ## end sub getOneImage

sub getCashRegistersList
{
    my ($self, $localBase) = @_;
    my $admin  = $self->{'OUTDOC'};
    my $cache  = $admin->{'CACHE'};
    my @result = ();
    my @todo   = ('cn=global,' . $cache->getRoot());
    if ($localBase) {
        # expand $localBase to location and individual OUs
        while ($localBase ne $cache->getRoot()) {
            push(@todo, $localBase);
            ($localBase) = $localBase =~ /^[^,]+,(.+)$/;
        }
    }
    while (@todo) {
        my $dn    = shift @todo;
        my $class = $cache->getClass($dn);
        if ($class eq 'scRefObjectContainer') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRole'));
            my $locality = 'global';
            if ($dn =~ /^cn=[^,]+,cn=[^,]+,/) {
                # only scRefObjectContainers under scLocation begins with double cn=
                $locality = 'local';
            }
            foreach my $cr ($cache->getChildrenByClass($dn, 'scCashRegister')) {
                push(@result, $cr);
            }
        }
        elsif ($class eq 'scRole') {
            foreach my $cr ($cache->getChildrenByClass($dn, 'scCashRegister')) {
                push(@result, $cr);
            }
        }
        elsif ($class eq 'scLocation') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRefObjectContainer'));
        }
    }
    return @result;
} ## end sub getCashRegistersList

sub getCashRegisters
{
    my ($self, $localBase) = @_;
    my $admin  = $self->{'OUTDOC'};
    my $cache  = $admin->{'CACHE'};
    my $crinfo = {};
    my @todo   = ('cn=global,' . $cache->getRoot());
    if ($localBase) {
        # expand $localBase to location and individual OUs
        while ($localBase ne $cache->getRoot()) {
            push(@todo, $localBase);
            ($localBase) = $localBase =~ /^[^,]+,(.+)$/;
        }
    }
    while (@todo) {
        my $dn    = shift @todo;
        my $class = $cache->getClass($dn);
        if ($class eq 'scRefObjectContainer') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRole'));
            my $locality = 'global';
            if ($dn =~ /^cn=[^,]+,cn=[^,]+,/) {
                # only scRefObjectContainers under scLocation begins with double cn=
                $locality = 'local';
            }
            foreach my $cr ($cache->getChildrenByClass($dn, 'scCashRegister')) {
                $self->getOneRegister($crinfo, $locality, $cr);
            }
        }
        elsif ($class eq 'scRole') {
            my $rolename = $cache->getLDAPentry($dn)->get_value('scRoleName');
            foreach my $cr ($cache->getChildrenByClass($dn, 'scCashRegister')) {
                $self->getOneRegister($crinfo, $rolename, $cr);
            }
        }
        elsif ($class eq 'scLocation') {
            push(@todo, $cache->getChildrenByClass($dn, 'scRefObjectContainer'));
        }
    }
    return $crinfo;
} ## end sub getCashRegisters

sub getOneRegister
{
    my ($self, $crinfo, $role, $cr) = @_;
    my ($id) = $cr =~ /^cn=([^,]+)/;
    my $crdata = {};
    $crdata = $self->getPOSattributes($cr);
    ## check optional CR attribute existence
    $crdata->{'scRaidScheme'}  = '' unless defined $crdata->{'scRaidScheme'};
    $crdata->{'scDiskJournal'} = '' unless defined $crdata->{'scDiskJournal'};
    $crdata->{'ROLE'}          = $role;
    # get FTs
    $self->getFileTemplates($cr, $crdata);
    # get HDD|RDD
    my @hdds = $self->{'OUTDOC'}->{'CACHE'}->getChildrenByClass($cr, 'scHarddisk');
    foreach (@hdds) {
        $_ =~ /^cn=([^,]+)/;
        $crdata->{'DISKS'}{$1} = $self->getPOSattributes($_);
        $crdata->{'DISKS'}{$1}{'TYPE'} = 'scHarddisk';
        if (defined $crdata->{'DISKS'}{$1}{'scPartitionsTable'}) {
            #decode scParitionsTable to PARTS
            my $i = 0;
            foreach (split(/;/, $crdata->{'DISKS'}{$1}{'scPartitionsTable'})) {
                my ($size, $type, $mount) = split(/\s+/, $_);
                # check attributes existence
                $size  = '' unless defined $size;
                $type  = '' unless defined $type;
                $mount = '' unless defined $mount;
                $crdata->{'DISKS'}{$1}{'PARTS'}{"$i - $mount"} =
                  { 'scPartType' => $type, 'scPartNum' => "$i", 'scPartMount' => $mount, 'scPartSize' => $size };
                $i++;
            }
            delete $crdata->{'DISKS'}{$1}{'scPartitionsTable'};
        }
        else {
            #decode scPartitions to PARTS
            my @parts = $self->{'OUTDOC'}->{'CACHE'}->getChildrenByClass($_, 'scPartition');
            foreach (@parts) {
                my $tmp = $self->getPOSattributes($_);
                # check attributes existence
                $tmp->{'scPassword'}  = '' unless defined $tmp->{'scPassword'};
                $tmp->{'scPartMount'} = '' unless defined $tmp->{'scPartMount'};
                $tmp->{'scPartSize'}  = '' unless defined $tmp->{'scPartSize'};
                $tmp->{'scPartNum'}   = '' unless defined $tmp->{'scPartNum'};
                $tmp->{'scPartType'}  = '' unless defined $tmp->{'scPartType'};
                $crdata->{'DISKS'}{$1}{'PARTS'}{"$tmp->{'scPartNum'} - $tmp->{'scPartMount'}"} = $tmp;
            }
        }
    }
    my @rdds = $self->{'OUTDOC'}->{'CACHE'}->getChildrenByClass($cr, 'scRamDisk');
    foreach (@rdds) {
        $_ =~ /^cn=([^,]+)/;
        $crdata->{'HDD'}{$1} = $self->getPOSattributes($_);
        $crdata->{'DISKS'}{$1}{'TYPE'} = 'scRamDisk';
    }
    $crinfo->{"${id};${role}"} = $crdata;
} ## end sub getOneRegister

sub getFileTemplates
{
    my ($self, $element, $data) = @_;
    my @tmp = $self->{'OUTDOC'}{'CACHE'}->getChildrenByClass($element, 'scConfigFileTemplate');
    foreach (@tmp) {
        $_ =~ /^cn=([^,]+)/;
        $data->{'TEMPLATES'}{$1} = $self->getPOSattributes($_);
        $data->{'TEMPLATES'}{$1}{'ROLE'} = 'FT';
    }
    @tmp = $self->{'OUTDOC'}{'CACHE'}->getChildrenByClass($element, 'scConfigFileSyncTemplate');
    foreach (@tmp) {
        $_ =~ /^cn=([^,]+)/;
        $data->{'TEMPLATES'}{$1} = $self->getPOSattributes($_);
        $data->{'TEMPLATES'}{$1}{'ROLE'} = 'FST';
    }
    @tmp = $self->{'OUTDOC'}{'CACHE'}->getChildrenByClass($element, 'scPxeFileTemplate');
    foreach (@tmp) {
        $_ =~ /^cn=([^,]+)/;
        $data->{'TEMPLATES'}{$1} = $self->getPOSattributes($_);
        $data->{'TEMPLATES'}{$1}{'ROLE'} = 'PT';
    }
} ## end sub getFileTemplates

sub getPOSattributes
{
    my ($self, $element) = @_;
    my $result = {};
    my $entry  = $self->{'OUTDOC'}->{'CACHE'}->getLDAPentry($element);
    my @attrs  = $entry->attributes();
    foreach my $attr (@attrs) {
        next if ($attr eq 'objectClass');
        $result->{$attr} = $entry->get_value($attr);
        # check TRUE|FALSE and convert to 1|0 for libyui
        if ($result->{$attr} eq 'TRUE') {
            $result->{$attr} = 1;
        }
        elsif ($result->{$attr} eq 'FALSE') {
            $result->{$attr} = 0;
        }
    }
    my ($attr) = $entry->dn() =~ /^([^=]+)/;
    delete $result->{$attr} if ($attr eq 'cn');
    # always store entry DN, is needed when saving so movement can be tracked
    $result->{'DN'} = $entry->dn();
    return $result;
} ## end sub getPOSattributes

sub getPOSattribute
{
    my ($self, $dn, $attribute) = @_;
    my $cache = $self->{'OUTDOC'}{'CACHE'};
    return $cache->getLDAPentry($dn)->get_value($attribute);
} ## end sub getPOSattribute
#########################
### EOF get Helpers
###############################################

################################################
### import&export
#########################
sub exportXML
{
    my ($self, $file) = @_;
    unless (defined($file)) {
        $file = $self->{'TMPFILE'};
    }
    my $cache = $self->{'OUTDOC'}{'CACHE'};
    #reset checked flags so XML is regenerated
    foreach (keys %{ $cache->{'LDS'} }) {
        $cache->setFlag($_, 'checked', 0);
    }
    #store doc
    my $xml = POS::Admin::Ldap2XML->new(CONFIG => 'none_needed');
    $xml->getXML($file, $self->{'OUTDOC'}, 'allow_overwrite' => 1);
} ## end sub exportXML
#########################
### EOF import&export
###############################################

sub DESTROY
{
    my ($self) = @_;
    unlink $self->{'TMPFILE'};
} ## end sub DESTROY

1;
