## @file
# SLEPOS LDAP validation and checking module
# - main class

package POS::Admin::Validator;

use strict;
use warnings;

use POS::Config;
use POS::Syslog;

use POS::Admin::Validator::SCS;
use POS::LDAP::Cache;
use POS::Admin::Validator::ACFL qw(%ACFL_CHECKLIST);

sub new
{
    my $type = shift;
    my $self = ref($type) || $type;
    my $this = bless {}, $self;

    my %opt = @_;
    $opt{'RELAXED'}         ||= 0;
    $opt{'CONFIG'}          ||= POS::Config->new();
    $opt{'ATTR_CHECK_BASE'} ||= $opt{'CONFIG'}{'ALL_LDAPBASE'};

    $this->{'LDS'} = POS::LDAP::Cache->new(
        CONFIG     => $opt{'CONFIG'},
        BASE       => $opt{'ENTRY_CHECK_BASE'},
        LDAP_DATA  => $opt{'LDAP_DATA'},
        MERGE_DATA => $opt{'MERGE_DATA'},
        MERGE_MODE => $opt{'MERGE_MODE'}
    );
    return if (not $this->{'LDS'});
    $this->{'SCS'} = POS::Admin::Validator::SCS->new(
        RELAXED   => $opt{'RELAXED'},
        ROOTCLASS => $opt{'ROOTCLASS'},
        LDS       => $this->{'LDS'}
    );
    $this->{'ACFL'} = POS::Admin::Validator::ACFL->new(LDS => $this->{'LDS'}, BASE => $opt{'ATTR_CHECK_BASE'});

    foreach my $objClass (keys %ACFL_CHECKLIST) {
        $this->{'SCS'}->registerHook($objClass, $ACFL_CHECKLIST{$objClass});
    }
    return $this;
} ## end sub new

sub validate
{
    my $this = shift;

    my $LDS  = $this->{'LDS'};
    my $SCS  = $this->{'SCS'};
    my $ACFL = $this->{'ACFL'};

    my $dentry = $LDS->getRoot();
    $SCS->setRoot();
    log_debug_print("Validation initialized");
# check if there is anything to check (may happed during repeated deletion of whole subtree. We just don't want to display undefined errors)
    if ($LDS->getFlag($dentry, 'checked')) {
        log_debug_print('LDS is marked as checked, skipping validation');
        return 1;
    }
    while (1) {
        #        log_debug_printf("Validating %s", $dentry);
        if (not $SCS->transitionDown($dentry)) {
            log_crit_printf("Error transitioning down to %s", $dentry);
            return;
        }
        #skip checking entries - may happen during transitioning up
        if (not $LDS->getFlag($dentry, 'checking')) {
            $LDS->setFlag($dentry, 'checking');
            return unless $ACFL->checkEntry($dentry);
        }
        my $_next = $LDS->getUnprocessedEntry();
        while (not defined $_next) {
            $LDS->setFlag($dentry, 'checked');
            if (not $SCS->transitionUp($dentry)) {
                return 1 if ($SCS->isAtRoot());
                log_crit_printf("Error transitioning up to %s", $dentry);
                return;
            }
            $dentry = $LDS->getParentDN($dentry);
            $_next  = $LDS->getUnprocessedEntry();
        }
        $dentry = $_next;
    }
} ## end sub validate

# get sta
sub postValidation
{
    my $self = shift;
    my $SCS  = $self->{'SCS'};
    #check some SLEPOS-specific variables
    #one organization
    if ($SCS->getFSMcounter('organization') > 1) {
        log_warn_print("More than one organization detected. This is unsupported scenario");
    }
    #at least one scLocation
    if ($SCS->getFSMcounter('scLocation') == 0) {
        log_warn_print("At least one scLocation must be present for working SLEPOS solution");
    }
    #at least one scPosImage
    if ($SCS->getFSMcounter('scPosImage') == 0) {
        log_warn_print("At least one scPosImage must be present for working SLEPOS solution");
    }
    #at least one scCashRegister
    if ($SCS->getFSMcounter('scCashRegister') == 0) {
        log_warn_print("At least one scCashRegister must be present for working SLEPOS solution");
    }
} ## end sub postValidation

sub printStats
{
    my $self = shift;
    print "-----------------------\n";
    my $count = 0;
    foreach (keys %{ $self->{'SCS'}{'FSM'} }) {
        next if $_ eq 'root';
        my $c = $self->{'SCS'}->getFSMcounter($_);
        print $_ . ": " . $c . "\n";
        $count += $c;
    }
    print "-----------------------\n";
    print "Number of entries: $count\n";
} ## end sub printStats

1;
