## @file
# SLEPOS LDAP validation and checking module
# - Search Control Structure

#takes POS::Admin::LdapStruct::%LDAP_STRUCT and creates FSM

package POS::Admin::Validator::SCS;

use strict;
use warnings;

use POS::Admin::LdapStruct qw(%LDAP_STRUCT);
use POS::Syslog;

our $VERSION    = 0.1;
our $ldapStruct = \%POS::Admin::LdapStruct::LDAP_STRUCT;

use constant { NEXT => 0, COUNTER => 1, PATHS => 2, HOOK => 3 };

sub new
{
    my $self = shift;
    my $type = ref($self) || $self;
    my $this = bless {}, $type;
    my %opt  = @_;
    # hash of array references
    $this->{'FSM'} = undef;
    # set root node to state pointer
    $this->{'state'} = 'root';
    # strict structure checking? 0 if not
    $this->{'relaxedStruct'} = $opt{'RELAXED'} || 0;
    # link to LDAP Data Struct
    $this->{'LDS'} = $opt{'LDS'} || die "Missing LDS in SCS init";
    # set what is under root element
    $this->{'root'} = $opt{'ROOTCLASS'} || $this->{'LDS'}->getClass($this->{'LDS'}->getRoot());
    # create FSM structure
    foreach my $obj (keys %$ldapStruct) {
        $this->{'FSM'}{$obj} = _createFSMelement($obj);
    }
    $this->{'FSM'}{'root'} = _createFSMroot($this->{'root'});

    return $this;
} ## end sub new

###############################################################################
# Private methods
sub _createFSMelement
{
    my $obj     = shift;
    my @element = ();
    # set allowed subentries, reset counter, paths and hook
    $element[NEXT]    = $ldapStruct->{$obj}{'subentries'};
    $element[COUNTER] = 0;
    $element[PATHS]   = {};
    $element[HOOK]    = undef;
    return \@element;
} ## end sub _createFSMelement

sub _createFSMroot
{
    my $root    = shift;
    my @element = ();
    # set allowed subentries, reset counter, paths and hook
    $element[NEXT]    = [[$root, 1]];
    $element[COUNTER] = 0;
    $element[PATHS]   = {};
    $element[HOOK]    = undef;
    return \@element;
} ## end sub _createFSMroot

###############################################################################
# Public methods
sub setRoot
{
    my $self = shift;
    $self->{'state'} = 'root';
} ## end sub setRoot

sub isAtRoot
{
    my $self = shift;
    return 1 if ($self->{'state'} eq 'root');
    return;
} ## end sub isAtRoot

sub registerHook
{
    my ($self, $stateName, $funcPtr) = @_;
    if (not defined $self->{'FSM'}{$stateName}) {
        log_error_printf("Requested state name does not exist: %s", $stateName);
        return;
    }
    log_debug_printf("Registering hook function to state %s", $stateName);
    $self->{'FSM'}{$stateName}[HOOK] = $funcPtr;
} ## end sub registerHook

#transition from parent to child
#@fn transitionDown($entry)
sub transitionDown
{
    my ($self, $dn) = @_;
    # shortcut to LDS
    my $LDS = $self->{'LDS'};
    # entry's parent dn = our DN
    my $path = $LDS->getParentDN($dn) || 'root';
    # entry's class = next state name
    my $class = $LDS->getClass($dn) || return;
    # ref to current state object
    #log_debug_printf("trying to transition from %s to %s as class %s", $path, $dn, $class);
    my $curState = $self->{'FSM'}{ $self->{'state'} };
    # search if next state is child of current state
    my @next = ();
    foreach my $_next (@{ $curState->[NEXT] }) {
        if ($$_next[0] eq $class) {
            @next = @$_next;
            last;
        }
    }
    if ($#next == 0) {
        log_error_printf("Structural error: %s is not allowed to be child of %s", $class, $self->{'state'});
        return;
    }
    # do we have path data - we were in this state at least once before
    if (defined $curState->[PATHS]{$path}) {
        #yes, find if transition is inside path limits
        my $found = 0;
        foreach my $_next (@{ $curState->[PATHS]{$path} }) {
            if ($_next eq $class) {
                # found it, check if there is max one visit limit
                #todo for now ignoring choice
                if ($next[1] eq "1" or $next[1] eq "?") {
                    # same class was already visited from same path, but only one visit is allowed
                    log_error_printf("Accessing %s from %s is not allowed more than once", $class, $path);
                    return;
                }
                #we are going to already visited state and are inside path limits
                $found = 1;
                last;
            }
        }
        if ($found == 0) {
            # we are going to visit previously unvisited state
            push @{ $curState->[PATHS]{$path} }, $class;
        }
    }
    else {
        #no, add path data
        $curState->[PATHS]{$path} = [$class];
    }
    #transition
    #log_debug_printf("Transitioning to %s", $class);
    $self->{'state'} = $class;
    return 1;
} ## end sub transitionDown

# transition from child to parent
#@fn bool transitionUp(*entry)
sub transitionUp
{
    my ($self, $dn) = @_;
    # shortcut to LDS
    my $LDS = $self->{'LDS'};
    # $dn stores how exactly we get here
    # $parent is where we go
    my $parent = $LDS->getParentDN($dn);
    my $class;
    if (not $parent) {
        #check if not going to root
        $class = 'root' if ($LDS->getClass($dn) eq $self->{'root'});
    }
    else {
        #check out parent's class
        $class = $LDS->getClass($parent);
    }
    #log_debug_printf("trying to transition from %s to %s as class %s", $dn, $parent || "root", $class);
    # current state reference array ref
    my $curState = $self->{'FSM'}{ $self->{'state'} };
    # is current state a child of parent state?
    my $found = 0;
    foreach my $_next (@{ $self->{'FSM'}{$class}[NEXT] }) {
        if ($$_next[0] eq $self->{'state'}) {
            $found = 1;
            last;
        }
    }
    if ($found == 0) {
        log_error_printf("%s can not be child of %s, invalid up transition in %s", $self->{'state'}, $class, $dn);
        return;
    }
    # compare visited children saved in current state{'next'} if all were checked
    foreach my $_next (@{ $curState->[NEXT] }) {
        #check only for mandatory entries
        #todo choice is omitted for now
        if ($$_next[1] eq "1" or $$_next[1] eq "+") {
            #now find $$_next[0] in $curState->{$path} array
            $found = 0;
            foreach (@{ $curState->[PATHS]{$dn} }) {
                if ($_ eq $$_next[0]) {
                    $found = 1;
                    last;
                }
            }
            if ($found == 0) {
                if ($self->{'relaxedStruct'} == 0) {
                    log_error_printf("%s is missing as child of %s", $$_next[0], $dn);
                }
                else {
                    log_info_printf("%s should be added as child of %s", $$_next[0], $dn);
                }
            }
        }
    }

    #execute hook on current entry if registered
    if (defined $curState->[HOOK]) {
        log_debug_printf("Executing registered hook for %s", $self->{'state'});
        if (not $curState->[HOOK]->($LDS, $dn)) {
            return;
        }
    }
    #log_debug_printf("Transitioning to %s", $class);
    $self->{'FSM'}{ $LDS->getClass($dn) }[COUNTER] += 1;
    $self->{'state'} = $class;
    #trigger invariant check on validator when we are back in root
    return if ($self->{'state'} eq 'root');
    return 1;
} ## end sub transitionUp

sub getFSMcounter
{
    my ($self, $class) = @_;
    return $self->{'FSM'}{$class}[COUNTER];
} ## end sub getFSMcounter
