#!/usr/bin/perl -w
## @file
# Implementation of image file functions

## @class
package POS::BootImage;

use strict;
use Cwd 'abs_path';
use POS::Syslog;
use POS::LDAP;
use POS::LdapUtil;


# version of the package
our $VERSION = 0.1;


## @cmethod object new()
# @return A new object
sub new()
{
    my $package = shift;
    my $data    = {
        'KERNEL_FILE'    => undef,
        'INITRD_FILE'    => undef,
        'KERNEL_SYMLINK' => undef,
        'INITRD_SYMLINK' => undef,
        'SYMLINK_BROKEN' => 0,
        'VERSION'        => '',
        'KERNEL_PATH'    => undef,
        'INITRD_PATH'    => undef,
        'CONTAINER'      => 'default',
    };

    return bless($data);
} ## end sub new()

sub valid
{
    my ($self) = @_;

    return 0 if !defined $self->{'KERNEL_FILE'};
    return 0 if !defined $self->{'INITRD_FILE'};
    return 0 if $self->{'SYMLINK_BROKEN'};
    return 1;
} ## end sub valid

sub get_error
{
    my ($self) = @_;

    return ("Invalid kernel file path",        1) if !defined $self->{'KERNEL_PATH'};
    return ("Invalid initrd file path",        2) if !defined $self->{'INITRD_PATH'};
    return ("Broken kernel or initrd symlink", 3) if $self->{'SYMLINK_BROKEN'};
    return (undef,                             0);
} ## end sub get_error

## @method bool set()
# loads info related to the given file
# @return 1 on success, 0 on failure
sub set
{
    my ($self, $kernel, $initrd, $kernel_symlink, $initrd_symlink) = @_;

    $self->{'KERNEL_FILE'}    = undef;
    $self->{'INITRD_FILE'}    = undef;
    $self->{'KERNEL_SYMLINK'} = undef;
    $self->{'INITRD_SYMLINK'} = undef;
    $self->{'KERNEL_MD5'}     = undef;
    $self->{'INITRD_MD5'}     = undef;
    $self->{'SYMLINK_BROKEN'} = 0;
    $self->{'VERSION'}        = '';
    $self->{'DIR'}            = undef;

    if (defined $kernel && -f $kernel && $kernel =~ /^.*\/(.*)$/) {
        $self->{'KERNEL_FILE'} = $1;
        $self->{'KERNEL_PATH'} = $kernel;
        if ($self->{'KERNEL_FILE'} =~ /^.*\.kernel\.(.*)$/) {
            $self->{'VERSION'} = $1;
        }
        if (-f $kernel . ".md5") {
            $self->{'KERNEL_MD5'} = $kernel . ".md5";
        }
    }

    if (defined $initrd && -f $initrd && $initrd =~ /^.*\/(.*)$/) {
        $self->{'INITRD_FILE'} = $1;
        $self->{'INITRD_PATH'} = $initrd;
        my $initrd_md5 = $initrd;
        $initrd_md5 =~ s/\.gz$//;
        $initrd_md5 .= ".md5";
        if (-f $initrd_md5) {
            $self->{'INITRD_MD5'} = $initrd_md5;
        }
    }

    if (defined $kernel_symlink &&
        $kernel_symlink =~ /^.*\/(.*)$/) {
        $self->{'KERNEL_SYMLINK'} = $1;
        if (abs_path($kernel_symlink) ne abs_path($kernel)) {
            $self->{'SYMLINK_BROKEN'} = 1;
        }
    }

    if (defined $initrd_symlink &&
        $initrd_symlink =~ /^.*\/(.*)$/) {
        $self->{'INITRD_SYMLINK'} = $1;
        if (abs_path($initrd_symlink) ne abs_path($initrd)) {
            $self->{'SYMLINK_BROKEN'} = 1;
        }
    }
    return $self->valid();
} ## end sub set

sub install
{
    my ($self, $dir, $kernel_symlink, $initrd_symlink, $hardlink) = @_;
    return unless $self->valid();

    $hardlink = 1 unless defined $hardlink;

    system('mkdir', '-p', $dir);

    my $dest_kernel = $dir . '/' . $self->{'KERNEL_FILE'};
    if (abs_path($dest_kernel) ne abs_path($self->{'KERNEL_PATH'})) {
        unlink($dest_kernel);
        $hardlink && link($self->{'KERNEL_PATH'}, $dest_kernel) or system('cp', $self->{'KERNEL_PATH'}, $dest_kernel);
    }

    my $dest_initrd = $dir . '/' . $self->{'INITRD_FILE'};
    if (abs_path($dest_initrd) ne abs_path($self->{'INITRD_PATH'})) {
        unlink($dest_initrd);
        $hardlink && link($self->{'INITRD_PATH'}, $dest_initrd) or system('cp', $self->{'INITRD_PATH'}, $dest_initrd);
    }

    if ($self->{'KERNEL_MD5'}) {
        my $dest_kernel_md5 = $dir . '/' . $self->{'KERNEL_FILE'} . ".md5";
        if (abs_path($dest_kernel_md5) ne abs_path($self->{'KERNEL_MD5'})) {
            unlink($dest_kernel_md5);
            system('cp', $self->{'KERNEL_MD5'}, $dest_kernel_md5);
        }
    }

    if ($self->{'INITRD_MD5'}) {
        my $dest_initrd_md5 = $dir . '/' . $self->{'INITRD_FILE'};
        $dest_initrd_md5 =~ s/\.gz$//;
        $dest_initrd_md5 .= ".md5";
        if (abs_path($dest_initrd_md5) ne abs_path($self->{'INITRD_MD5'})) {
            unlink($dest_initrd_md5);
            system('cp', $self->{'INITRD_MD5'}, $dest_initrd_md5);
        }
    }

    my $dest_kernel_symlink;
    if (defined $kernel_symlink) {
        $dest_kernel_symlink = $dir . '/' . $kernel_symlink;
        system('ln', '-sf', $self->{'KERNEL_FILE'}, $dest_kernel_symlink);
        if ($self->{'KERNEL_MD5'}) {
            my $dest_kernel_symlink_md5 = $dir . '/' . $kernel_symlink . ".md5";
            if (abs_path($dest_kernel_symlink_md5) ne abs_path($self->{'KERNEL_MD5'})) {
                unlink($dest_kernel_symlink_md5);
                system('cp', $self->{'KERNEL_MD5'}, $dest_kernel_symlink_md5);
            }
        }
    }

    my $dest_initrd_symlink;
    if (defined $initrd_symlink) {
        $dest_initrd_symlink = $dir . '/' . $initrd_symlink;
        system('ln', '-sf', $self->{'INITRD_FILE'}, $dest_initrd_symlink);
        if ($self->{'INITRD_MD5'}) {
            my $dest_initrd_symlink_md5 = $dir . '/' . $initrd_symlink;
            $dest_initrd_symlink_md5 =~ s/\.gz$//;
            $dest_initrd_symlink_md5 .= ".md5";
            if (abs_path($dest_initrd_symlink_md5) ne abs_path($self->{'INITRD_MD5'})) {
                unlink($dest_initrd_symlink_md5);
                system('cp', $self->{'INITRD_MD5'}, $dest_initrd_symlink_md5);
            }
        }
    }

    my $installed = POS::BootImage->new();
    $installed->set($dest_kernel, $dest_initrd, $dest_kernel_symlink, $dest_initrd_symlink) or return;
    return $installed;
} ## end sub install

sub remove
{
    my ($self) = @_;
    unlink($self->{'KERNEL_PATH'}) if defined $self->{'KERNEL_PATH'};
    unlink($self->{'INITRD_PATH'}) if defined $self->{'INITRD_PATH'};
    unlink($self->{'KERNEL_MD5'})  if defined $self->{'KERNEL_MD5'};
    unlink($self->{'INITRD_MD5'})  if defined $self->{'INITRD_MD5'};
} ## end sub remove

sub ldap_add
{
    my ($self, $POS_GLOBALS, $container_cn) = @_;

    $container_cn = $self->{'CONTAINER'} unless defined $container_cn;

    if (!$self->valid()) {
        return 0;
    }

    my $dn = "cn=" . $container_cn . ",cn=global," . $POS_GLOBALS->{'STATIC'}{'LDAPBASE'};

    my $kernel = $self->{'KERNEL_FILE'};
    my $initrd = $self->{'INITRD_FILE'};

    $kernel = $self->{'KERNEL_SYMLINK'} if defined $self->{'KERNEL_SYMLINK'};
    $initrd = $self->{'INITRD_SYMLINK'} if defined $self->{'INITRD_SYMLINK'};

    my $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->add(
        $dn,
        attr => [
            objectClass     => ['scDistributionContainer', 'top'],
            cn              => $container_cn,
            scKernelName    => $kernel,
            scInitrdName    => $initrd,
            scKernelVersion => $self->{'VERSION'},
            scKernelMatch   => 'MATCH_VERSION',
        ]
    );
    if ($res->code == 68) {
        log_warn_print("scDistributionContainer $dn already exists in LDAP");

        $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->modify(
            $dn,
            replace => {
                objectClass     => ['scDistributionContainer', 'top'],
                cn              => $container_cn,
                scKernelName    => $kernel,
                scInitrdName    => $initrd,
                scKernelVersion => $self->{'VERSION'},
                scKernelMatch   => 'MATCH_VERSION',
            }
        );
        #print $res->code,":",$res->error,"\n";
        if ($res->code) {
            log_error_print $res->code, ":", $res->error;
            return 0;
        }
        log_warn_print("scDistributionContainer was replaced");
    }
    else {
        if ($res->code) {
            log_error_print $res->code, ":", $res->error;
            return 0;
        }
    }
    return 1;
} ## end sub ldap_add


1;
