## @file
# LDAP data structure implementation
# @par Authors
# - Ondrej Holecek
# - Pavel Sladek
# @copy 2011, SuSE Linux

## @class
# POS::LDAP::Cache class.
# @par Provides
# - data storage for LDAP entries addressed by DN
# - per entry flags
# - linear children transitioning
# - mapping to SLEPOS LDAP objectClasses (generic objectClasses are filtered)
# - merging two LDAP datasets
# @todo insert, update
# @par Requires
# - POS::Syslog
# - POS::Config
# - POS::LDAP
# - POS::Admin::LdapStruct
# - Net::LDAP::Constant
# @note
# Previously known as POS::Admin::Validator::LDS;
package POS::LDAP::Cache;

use strict;
use warnings;

use POS::LDAP;
use POS::Admin::LdapStruct qw(%LDAP_STRUCT);
use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT LDAP_OPERATIONS_ERROR);

use POS::Syslog;
use POS::Config;

our $VERSION = 0.1;

## @var const ENTRY
# array element pointing to Net::LDAP::Entry object

## @var const NEXT
# array element pointing to list of children entries

## @var const FLAGS
# array element poinging to list of flags for current entry

use constant {
    # LDSEntry constants
    ENTRY => 0,
    NEXT  => 1,
    FLAGS => 2,
    # other constants
    LDS     => 75,
    LSTRUCT => 76,
    LDAP    => 77
};

## @cmethod POS::LDAP::Cache new($optHashRef)
# Create new POS::LDAP::Cache object.
# @param optHashRef hash reference storing following options:
# - CONFIG POS::Config object ref, if not present new POS::Config object is created
# - BASE search base for ldap search (default POS::Config->{'ALL_LDAPBASE'})
# - LDAP_DATA custom ldap search result (object ref Net::LDAP::Search) or list of Net::LDAP::Entry objects
#(default undef, new search is performed from BASE)
# - MERGE_DATA custom ldap search result (object ref Net::LDAP::Search) or list of Net::LDAP::Entry objects to be merged
#with LDAP_DATA (default undef, no merging)
# - MERGE_MODE merge mode can be one of following values (default update):
#   - import only adds new entries, conflicting entries are ignored
#   - update adds new entries, conflicing entries are updated according to change type (add/modify/delete)
#   - replace adds new entries, conflicting entries (and their subtrees) are deleted
# @note
# Options are not mandatory, if not present, default values are used. Therefor object can be used as simple as my $cache = POS::LDAP::Cache->new();
# @return A new object ref
sub new
{
    my $self = shift;
    my $type = ref($self) || $self;
    my $this = bless {}, $type;
    my %opt  = @_;

    ## set up defaults
    #POS::CONFIG
    my $config = defined $opt{'CONFIG'} ? $opt{'CONFIG'} : POS::Config->new();
    #scope of check
    $this->{'BASE'} = $opt{'BASE'} || $config->{'ALL_LDAPBASE'};
    $this->{'BASE'} = $config->{'ALL_LDAPBASE'} if $this->{'BASE'} eq 'all';
    ## set up our LDAP connection
    $this->{'LDAP'} = $opt{'LDAP'};
    $this->{'LDAP'} = POS::LDAP->new(CONFIG => $config) unless (defined $this->{'LDAP'});
    if (not defined $this->{'LDAP'}) {
        log_crit_print("Unable to connect to LDAP, aborting");
        return;
    }
    # se CHANGES_ONLY flag when POS::Cache is about to store and merge changes and not actual LDAP data
    # mergeEntryToCache will not then change modifyTypes
    $this->{'CHANGES_ONLY'} = defined $opt{'CHANGES_ONLY'} ? 1 : 0;
    #Net::LDAP::Search to check
    my $search = $opt{'LDAP_DATA'};

    #if Net::LDAP::Search is not provided, make our from BASE
    if (not defined $search) {
        my $res = $this->{'LDAP'}->search(
            base   => $this->{'BASE'},
            filter => '(objectClass=scLocation)',
            attrs  => ['objectClass', 'userPassword'],
            scope  => 'base'
        );
        if ($res->is_error() && $res->code ne LDAP_NO_SUCH_OBJECT) {
            log_error_printf("LDAP search failed with message \"%s\"", $res->error());
            return;
        }
        elsif ($res->count() == 0) {
            log_debug_print("Requested scope is not branch server");
        }
        else {
            log_debug_print("Requested scope is branch server");
            my $entry = $res->shift_entry();
            #get user password, set BASE to POS_ADMIN_DN and relogin under scLocation user
            my $pass = $entry->get_value('userPassword');
            my $dn   = $entry->dn();
            $this->{'BASE'} = $config->{'ALL_LDAPBASE'};
            $res = $this->{'LDAP'}->bind($dn, $pass);
            if ($res->is_error()) {
                log_error_printf("Unable to rebind as branch server %s", $dn);
                return;
            }
        }
        $this->{'LDIF'} = '' if !defined $this->{'LDIF'};
        log_debug_printf("BASE is \'%s\', LDIF is \'%s\'", $this->{'BASE'}, $this->{'LDIF'});
        # get ldap data
        $res = $this->{'LDAP'}->search(base => $this->{'BASE'}, filter => '(objectClass=*)', scope => 'sub');
        if ($res->is_error() and $res->code() ne LDAP_NO_SUCH_OBJECT) {
            log_error_printf("LDAP search failed with message \"%s\"", $res->error());
            return;
        }
        $search = $res;
    }

    #create LDSs
    $this->{'LDS'} = _createLDS($this, $search);
    if ($this->{'LDS'}) {
        log_debug_print("LDS created successfuly");
        if ($opt{'MERGE_DATA'}) {
            if (_mergeLDAP($this->{'LDS'}, $opt{'MERGE_DATA'}, $this->{'BASE'}, $opt{'MERGE_MODE'})) {
                log_debug_print("LDS merged successfuly");
            }
            else {
                log_error_print("LDS merge failed");
                return;
            }
        }
    }
    else {
        log_crit_print("failed to create LDAP LDS");
        return;
    }
    $this->adjustBase();

    #lowercase SLEPOS objectClasses for case insensitive search
    my %LSTRUCT = ();
    foreach (keys %LDAP_STRUCT) {
        $LSTRUCT{ lc $_ } = $_;
    }
    $this->{'LSTRUCT'} = \%LSTRUCT;
    return $this;

} ## end sub new
###############################################################################
# private methods

## @method private hash_ref _createLDS($search)
# create LDAP data structure from Net::LDAP::Search or list of Net::LDAP::Entry objects
# @param search Net::LDAP::Search reference or ref to list of Net::LDAP::Entry objects
# @par Algorithm:
# -# create LDS struct for root entry ($self->'BASE')
# -# get $entry from list
# -# if $entry already in LDS (decided by DN), check Net::LDAP::Entry ref and update if necessary, goto 2
# -# create LDS struct for $entry and add it into LDS, goto 2
# -# return LDS ref
# @return LDS reference
sub _createLDS
{
    my ($self, $search) = @_;
    my %lds;
    my $nextEntry;
    if (ref($search) eq "ARRAY") {
        $nextEntry = sub { shift @$search };
    }
    else {
        $nextEntry = sub { $search->shift_entry() };
    }

    $lds{ $self->{'BASE'} } = _createLDSEntry(undef, []);
    while (my $entry = &$nextEntry()) {
        my $dn = $entry->dn();
        if (defined $lds{$dn}) {
            $lds{$dn}[ENTRY] = $entry if (not defined $lds{$dn}[ENTRY]);
            next;
        }
        $lds{$dn} = _createLDSEntry($entry, []);
        return unless _addEntryToLDS(\%lds, $dn);
    }
    log_debug_print("_createLDS finished");
    return \%lds;
} ## end sub _createLDS

## @method private bool _mergeLDAP($LDS, $entries, $base, $mode)
# merge/update LDAP data structure (LDS) with list of Net::LDAP::Entry objects
# @param $LDS LDAP data structure ref
# @param $entries array of Net::LDAP::Entry objects or POS::LDAP::Cache object
# @param $base
# @param $mode merge mode, see constructor for details, deprecated
# @return true if success else undef
sub _mergeLDAP
{
    my ($LDS, $entries, $base, $mode, $changesOnly) = @_;
    log_debug_printf('merge started. base is %s, entries type is %s', $base, ref $entries);
    eval {
        if (ref $entries eq 'POS::LDAP::Cache') {
            my @DNs = ($base);
            while (@DNs) {
                my $dn = pop @DNs;
                push(@DNs, @{ $entries->getChildren($dn) });
                log_debug_printf('merging entry %s', $dn);
                _mergeEntryToCache($LDS, $entries->getLDAPentry($dn), $changesOnly);
            }
        }
        else {
            foreach (@$entries) {
                _mergeEntryToCache($LDS, $_, $changesOnly);
            }
        }
    };
    if ($@) {
        log_error_print($@);
        return;
    }
    log_debug_print('merge finished successfuly');
    return 1;
} ## end sub _mergeLDAP

sub _mergeEntryToCache
{
    my ($LDS, $entry, $changesOnly) = @_;
    # skip glue entries
    return if not $entry;
    my $entryDN    = $entry->dn();
    my $changeType = $entry->changetype();
    if ($changeType eq 'add') {
        log_debug_printf('adding entry %s', $entryDN);
        if ($LDS->{$entryDN} and @{ $LDS->{$entryDN} }) {
            if (not defined $LDS->{$entryDN}[ENTRY]) {
                # this can happen when importing to empty tree, just link the LDAP::Entry object
                $LDS->{$entryDN}[ENTRY] = $entry;
            }
            else {
                my $oldentry = $LDS->{$entryDN}[ENTRY];
                if ($oldentry->changetype() eq 'delete') {
                    log_debug_printf('changing changetype to replace');
                    $LDS->{$entryDN}[ENTRY] = $entry;
                    $entry->changetype('replace');
                    # set to 0 so we unskip readded entries
                    $LDS->{$entryDN}[FLAGS]{'checked'} = 0;
                }
                else {
                    log_warn_printf('skipping already existing entry %s', $entryDN);
                }
            }
        }
        else {
            #create LDSentry and add LDSentry to $LDS
            $LDS->{$entryDN} = _createLDSEntry($entry, []);
            die unless _addEntryToLDS($LDS, $entryDN);
        }
    }
    elsif ($changeType eq 'delete') {
        log_debug_printf('deleting entry %s', $entryDN);
        if ($changesOnly) {
            if (not $LDS->{$entryDN}) {
                $LDS->{$entryDN} = _createLDSEntry($entry);
                die unless _addEntryToLDS($LDS, $entryDN);
            }
            else {
                if (not $LDS->{$entryDN}[ENTRY]) {
                    $LDS->{$entryDN}[ENTRY] = $entry;
                }
                else {
                    $LDS->{$entryDN}[ENTRY]->changetype('delete');
                }
            }
            # validator must skip deleted entries
            $LDS->{$entryDN}[FLAGS]{'checked'} = 1;
        }
        else {
            if ($LDS->{$entryDN}) {
                _removeChild($LDS, $entryDN);
                delete $LDS->{$entryDN};
            }
        }
    }
    elsif ($changeType eq 'modify') {
        log_debug_printf('modify entry %s', $entryDN);
        if (not defined $LDS->{$entryDN}) {
            if (not $changesOnly) {
                log_warn_printf('modifying nonexisting entry %s, trying adding as is', $entryDN);
                $entry->changetype('add');
            }
            $LDS->{$entryDN} = _createLDSEntry($entry);
            die unless _addEntryToLDS($LDS, $entryDN);
        }
        else {
            my $ldsEntry = $LDS->{$entryDN}[ENTRY];
            foreach ($entry->attributes()) {
                $ldsEntry->replace($_ => $entry->get_value($_));
            }
        }
    }
    elsif ($changeType eq 'replace') {
        log_debug_printf('replacing entry %s', $entryDN);
        if (not $changesOnly) {
            if (not defined $LDS->{$entryDN}) {
                log_info_printf('replacing nonexisting entry');
                $entry->changetype('add');
            }
            else {
                $entry->changetype('modify');
          # now we need to use $entry for validation LDS, but also create modify data in original entry for LDAP::update
                my $tmpEntry = $entry->clone();

                foreach ($entry->attributes) {
                    $entry->delete($_);
                }
                foreach ($tmpEntry->attributes()) {
                    next if ($_ eq 'top');
                    $entry->add($_ => [$tmpEntry->get_value($_)]);
                }
            }
            $LDS->{$entryDN}[ENTRY] = $entry;
        }
    }
    else {
        die "unknown change type \"$changeType\" for entry $entryDN";
    }
    return 1;
} ## end sub _mergeEntryToCache

## @fn private array_ref _createLDSEntry($entry, $next)
# creates new LDS entry.
# @param $entry Net::LDAP::Entry object ref, undef is allowed
# @param $next array ref to list of children
# @return ref to new LDS entry
sub _createLDSEntry
{
    my ($entry, $next) = @_;
    my @LDSentry = ($entry, $next || [], {});
    return \@LDSentry;
} ## end sub _createLDSEntry

## @fn private void _addChildren($LDS, $DN, @children)
# adds children to $LDS entry specified by $DN
# @param LDS LDS reference
# @param DN string with LDAP DN
# @param children array of DNs of entry children
# @note
# previous children list is not overwritten (add != replace)
sub _addChildren
{
    my ($lds, $dn, @children) = @_;
    push @{ $lds->{$dn}[NEXT] }, @children;
} ## end sub _addChildren

sub _removeChild
{
    my ($lds, $dn) = @_;
    my $parentDN = (split(/,/, $dn, 2))[1];
    if ($lds->{$parentDN}) {
        $lds->{$parentDN}[NEXT] = [grep (!/$dn/, @{ $lds->{$parentDN}[NEXT] })];
    }
} ## end sub _removeChild

## @method private bool _addEntryToLDS($LDS, $entryDN)
# adds entry to correct place in LDS tree (= updates children list of some LDS entry)
# @param LDS LDS reference
# @param entryDN string with LDAP DN
# @return true if success else undef
# @note
# if trying to add LDS entry which does not have immediate parent in LDS yet, parent is created with undef ENTRY ref.
# Same rule apply this parent, so there is always connected tree. No node is ever disconnected.
sub _addEntryToLDS
{
    my ($LDS, $entryDN) = @_;

    #get DN of parent entry private method
    my $getParentDN = sub { my ($dn) = @_; (split(/,/, $dn, 2))[1]; };

    #get parrent
    my $parentDN = &$getParentDN($entryDN);
    #is parent in LDS?
    if (defined $LDS->{$parentDN}) {
        #add entry as child
        _addChildren($LDS, $parentDN, $entryDN);
    }
    else {
        $LDS->{$parentDN} = _createLDSEntry(undef, [$entryDN]);
        #build up parent tree
        while (defined $parentDN) {
            $entryDN  = $parentDN;
            $parentDN = &$getParentDN($entryDN);
            if (defined $LDS->{$parentDN}) {
                _addChildren($LDS, $parentDN, $entryDN);
                last;
            }
            else {
                $LDS->{$parentDN} = _createLDSEntry(undef, [$entryDN]);
            }
        }
        if (not defined $parentDN) {
            log_crit_print('Something went wrong during building parent tree. Check DNs');
            return;
        }
    }
    return 1;
} ## end sub _addEntryToLDS

## @method void sanitizeCache (void)
# to ease moving and renaming subtrees in cache, only DNs are updated. Then _sanitizeCache will traverse
# through LDS tree and move and reconnect all entries so they match theirs DN and create seamless tree
# @return void
sub _sanitizeCache
{
    my ($lds, $root) = @_;
    my @todo = ($root);
    while (@todo) {
        my $cacheDN = shift @todo;
        my $entryDN = $lds->{$cacheDN}[ENTRY]->dn();
        push(@todo, @{ $lds->{$cacheDN}[NEXT] });
        if ($cacheDN ne $entryDN) {
            # if entry DN is different to cache DN, move it to proper position and update children
            $lds->{$entryDN} = $lds->{$cacheDN};
            delete $lds->{$cacheDN};
        }
        # always check children
        my @newChildren = ();
        foreach (@{ $lds->{$entryDN}[NEXT] }) {
            # update children DN
            # children may not be in LDS yet, skip parsing
            next unless ($lds->{$_} and @{ $lds->{$_} });
            my $entry = $lds->{$_}[ENTRY];
            my $dn    = $entry->dn();
            $dn =~ s/^([^,]+).+$/$1,$entryDN/;
            $entry->dn($dn);
            # update NEXT array. This also takes into account renamed children
            push(@newChildren, $dn);
        }
        $lds->{$entryDN}[NEXT] = \@newChildren;
    }
} ## end sub _sanitizeCache

###############################################################################
# public methods

## @fn string getRoot()
# get root DN of LDS
# @return string with root DN of LDS
sub getRoot
{
    my ($self) = @_;
    return $self->{'BASE'};
} ## end sub getRoot

## @method string getUnprocessedEntry()
# return DN of next child entry with unset flag 'checked' or undef if no such child exists.
# @return string DN of undef
sub getUnprocessedEntry
{
    #remember - DFS
    my ($self) = @_;
    my $curentry = $self->{'currentEntry'};
    unless ($curentry) {
        $self->{'currentEntry'} = $self->{'BASE'};
        $curentry = $self->{'BASE'};
    }
    foreach my $entry (@{ $self->{'LDS'}{$curentry}[NEXT] }) {
        if ($self->getFlag($entry, 'checked')) {
            next;
        }
        $self->{'currentEntry'} = $entry;
        last;
    }
    if ($curentry ne $self->{'currentEntry'}) {
        #we have an unprocessed child
        return $self->{'currentEntry'};
    }
    #all children processed, go up, return undef
    $self->{'currentEntry'} = $self->getParentDN($curentry);
    return;
} ## end sub getUnprocessedEntry

## @fn void setFlag($DN, $flag, $value)
# set $flag to value $value in LDS entry addressed by $DN
# @param DN string DN addressing LDS entry
# @param flag string flag identification
# @param value flag value (anything scalar)
sub setFlag
{
    my ($self, $dn, $flag, $value) = @_;
    $value = 1 unless defined $value;
    $self->{'LDS'}{$dn}[FLAGS]{$flag} = $value;
} ## end sub setFlag

## @fn scalar getFlag($DN, $flag)
# get $flag value from LDS entry addressed by $DN
# @param DN string DN addressing LDS entry
# @param flag string flag identification
# @return $flag value
sub getFlag
{
    my ($self, $dn, $flag) = @_;
    return $self->{'LDS'}{$dn}[FLAGS]{$flag};
} ## end sub getFlag

## @fn string getParentDN($DN)
# get parent DN from DN, if trying to get parent of $self->{'BASE'} returns undef
# @param DN string DN
# @return string DN of parent or undef
sub getParentDN
{
    my ($self, $dn) = @_;
    return if (length $dn <= length $self->{'BASE'});
    return (split(/,/, $dn, 2))[1];
} ## end sub getParentDN

## @fn string getClass($DN)
# get SLEPOS objectClass of entry addressed by $DN
# @param DN string DN addressing LDS entry
# @return string with SLEPOS objectClass name
sub getClass
{
    my ($self, $dn) = @_;
    if (not defined $self->{'LDS'}{$dn}[ENTRY]) {
        log_error_printf('Trying to access unknown DN %s', $dn);
        return;
    }
    my @objectClasses = $self->{'LDS'}{$dn}[ENTRY]->get_value('objectClass');
    #get the proper POS object class (strip default LDAP classes)
    foreach (@objectClasses) {
        my $x = lc $_;
        return $self->{'LSTRUCT'}{$x} if $self->{'LSTRUCT'}{$x};
    }
    log_error_printf('Unable to find SLEPOS LDAP class for %s', $dn);
    return;
} ## end sub getClass

## @fn ref getLDAPentry($DN)
# get reference to Net::LDAP::Entry for LDS entry addressed by $DN. Can be undef.
# @param DN string DN addressing LDS entry
# @return Net::LDAP::Entry ref or undef
sub getLDAPentry
{
    my ($self, $dn) = @_;
    return $self->{'LDS'}{$dn}[ENTRY];
} ## end sub getLDAPentry

## @fn array_ref getChildren($DN)
# get all children of LDS entry addressed by $DN
# @param DN string DN addressing LDS entry
# @return list of children DNs of LDS entry
sub getChildren
{
    my ($self, $dn) = @_;
    return [] unless $dn;
    return $self->{'LDS'}{$dn}[NEXT];
} ## end sub getChildren

## @fn array_ref getChildrenByClass($DN, $classFilter)
# get children of LDS entry only of given class type specified in $classFilter
# @param DN string DN addressing LDS entry
# @param classFilter string name of objectClass to get
# @return list of children DNs
sub getChildrenByClass
{
    my ($self, $dn, $class) = @_;
    my @res;
    my $children = $self->getChildren($dn);
    if (not defined $class) {
        @res = @$children;
    }
    else {
        foreach (@{$children}) {
            if (($self->getClass($_) || '') eq $class) {
                push(@res, $_);
            }
        }
    }
    if (wantarray) {
        return @res;
    }
    return \@res;
} ## end sub getChildrenByClass

sub startMultipleChanges
{
    my ($self) = @_;
    $self->{'MULTIPLE_CHANGES'} = 1;
} ## end sub startMultipleChanges

sub endMultipleChanges
{
    my ($self) = @_;
    $self->{'MULTIPLE_CHANGES'} = 0;
    _sanitizeCache($self->{'LDS'}, $self->{'BASE'});
} ## end sub endMultipleChanges

sub moveEntry
{
    my ($self, $entryDN, $newParentDN, $newName) = @_;
    die 'cannot use this in CHANGES_ONLY mode' if $self->{'CHANGES_ONLY'};
    # determine new DN
    my $newDN = $newName;
    ($newDN) = $entryDN =~ /^([^,]+)/ unless (defined $newName);
    $newDN = $newDN . ',' . $newParentDN;
    if ($self->{'LDS'}{$newDN} and @{ $self->{'LDS'}{$newDN} }) {
        log_error_printf('Target DN %s already exists!', $newDN);
        return;
    }
    # remove itself
    _removeChild($self->{'LDS'}, $entryDN);
    # add to another parend
    _addChildren($self->{'LDS'}, $newParentDN, $newDN);
    # move to new LDS spot
    $self->{'LDS'}{$newDN} = $self->{'LDS'}{$entryDN};
    delete $self->{'LDS'}{$entryDN};
    # set new DN
    $self->getLDAPentry($newDN)->dn($newDN);
    if (not $self->{'MULTIPLE_CHANGES'}) {
        _sanitizeCache($self->{'LDS'}, $self->{'BASE'});
    }
    return 1;
} ## end sub moveEntry

sub renameEntry
{
    my ($self, $entryDN, $newName) = @_;
    my ($parentDN) = $entryDN =~ /^[^,]+,(.+)$/;
    return $self->moveEntry($entryDN, $parentDN, $newName);
} ## end sub renameEntry

## @fn array_ref getChildrenFiltered($DN, $classFilter)
# only alias to getChildrenByClass
# @note
# should not be used, use getChildrenByClass instead
sub getChildrenFiltered { goto &getChildrenByClass; }


#Filter dn entries
# input: ref on list of dn's, ref on hash keyed by dn's, or a single dn
# output: ref on list of dn's compatible with filter
## @fn array_ref FilterEntries($DNref, $generalFilter)
# checks if entry object is consistent with conditions specified in $classFilter
# @param DN string/array reference/hash reference DN addressing LDS entry(s)
# @param generalFilter string LDAP filter (subset of RFC-2254 implemented)
# @return list reference of compatible DNs from DNs provided.
sub FilterEntries
{
    my ($self, $dn_ar, $f) = @_;
    my @fdn_a;
    if (ref(\$dn_ar) eq 'SCALAR') {    #make it ref to array if scalar
        my @dn_a = ($dn_ar);
        $dn_ar = \@dn_a;
    }
    elsif (ref($dn_ar) eq 'HASH') {    #make it ref to array if hash
        my @dn_a = keys(%{$dn_ar});
        $dn_ar = \@dn_a;
    }
    return $dn_ar if !defined $f;
    require Net::LDAP::Filter;
    my $filt = Net::LDAP::Filter->new;
    $filt->parse($f) or return;
    foreach my $dn (@{$dn_ar}) {
        my $res = 1;
        $res = $self->_resolve_filter($self->getLDAPentry($dn), $filt);
        @fdn_a = (@fdn_a, $dn) if $res;
    }
    return \@fdn_a;
} ## end sub FilterEntries


#_resolve_filter
# TODO: multival support may not work
## @fn array_ref _resolve_filter($entry, $filterObject)
# checks if entry object is consistent with filter $filterObject
# @param Net::LDAP::Entry reference
# @param Net::LDAP::Filter reference to parsed filter
# @return 1 if filter passed, 0 if not
sub _resolve_filter
{
    my ($self, $entry, $filt) = @_;
    #make sure $filt object is not damaged
    my @op = keys(%$filt);
    my $op = shift @op;      #only one can be present
    if (@op) {
        log_alert_print('filter parsing error, more than one op');
        return;
    }
    my $finres     = 0;
    my $operands_x = $filt->{$op};
    #eq, subs, <,>,=
    if ($op eq 'equalityMatch') {
        my $assertionValue = $operands_x->{'assertionValue'};
        my $attributeDesc  = $operands_x->{'attributeDesc'};
        my $value          = $entry->get_value($attributeDesc);
        $finres = 1 if (defined $value && lc($value) eq lc($assertionValue));
        return $finres;
    }
    if ($op eq 'substrings') {
        my $type       = $operands_x->{'type'};
        my $substrings = $operands_x->{'substrings'};
        my $testregexp;
        foreach my $subs (@{$substrings}) {
            my ($subtype, $substr) = each(%{$subs});    #only one
            $testregexp .= '^' if ($subtype eq 'initial');
            $testregexp .= $substr;
            $testregexp .= '$' if ($subtype eq 'final');
            $testregexp .= '.*';
        }
        $finres = 1 if ($entry->get_value($type) =~ /$testregexp/);
        return $finres;
    }
    #TODO: ~=  approxMatch, >=  greaterOrEqual,  <=  lessOrEqual,  :=  extensibleMatch

    #or, and, not, present
    if ($op eq 'present') {
        $finres = 1 if defined $entry->get_value($operands_x);
        return $finres;
    }
    #or, and, not (->subfilters)
    my $opr;
    $opr = $operands_x->[0] if ref($operands_x) eq 'ARRAY';
    $opr = $operands_x if (ref($operands_x) eq 'HASH');
    $finres = $self->_resolve_filter($entry, $opr);
    if ($op eq 'not') {
        return !$finres;
    }
    #and, or
    for (my $j = 1; $j < @{$operands_x}; $j++) {
        if ($op eq 'or' && $finres) {
            return 1;
        }
        if ($op eq 'and' && !$finres) {
            return 0;
        }
        my $res = $self->_resolve_filter($entry, $operands_x->[$j]);
        if ($op eq 'or') {
            $finres = $finres || $res;
        }
        if ($op eq 'and') {
            $finres = $finres && $res;
        }
    }
    return $finres;
} ## end sub _resolve_filter

## @fn LDAP_RESULT modify($DN, $options)
# modify LDAP Entry addressed by $DN in LDAP and LDS
# @param DN string DN addressing LDS entry
# @param options which attributes to add/replace/delete, see Net::LDAP for more
# @return LDAP result of operation
sub modify
{
    my $self = shift;
    my $dn   = shift;
    my %opts = @_;
    my $res  = $self->{'LDAP'}->s_modify($dn, %opts);
    if ($res->is_error()) {
        log_debug_printf('Unable to modify entry %s in LDAP. Error: %u %s', $dn, $res->code, $res->error);
        return $res;
    }
    #parse modify to add/delete/replace attributes
    if ($opts{'changes'}) {
        $opts{'add'}     = $opts{'changes'}{'add'}     if (defined $opts{'changes'}{'add'});
        $opts{'delete'}  = $opts{'changes'}{'delete'}  if (defined $opts{'changes'}{'delete'});
        $opts{'replace'} = $opts{'changes'}{'replace'} if (defined $opts{'changes'}{'replace'});
    }
    $self->{'LDS'}{$dn}[ENTRY]->add(@{ $opts{'add'} }) if (defined $opts{'add'} && (ref $opts{'add'} eq 'ARRAY'));
    $self->{'LDS'}{$dn}[ENTRY]->add(%{ $opts{'add'} }) if (defined $opts{'add'} && (ref $opts{'add'} eq 'HASH'));
    $self->{'LDS'}{$dn}[ENTRY]->delete(@{ $opts{'delete'} })
      if (defined $opts{'delete'} && (ref $opts{'delete'} eq 'ARRAY'));
    $self->{'LDS'}{$dn}[ENTRY]->delete(%{ $opts{'delete'} })
      if (defined $opts{'delete'} && (ref $opts{'delete'} eq 'HASH'));
    $self->{'LDS'}{$dn}[ENTRY]->replace(@{ $opts{'replace'} })
      if (defined $opts{'replace'} && (ref $opts{'replace'} eq 'ARRAY'));
    $self->{'LDS'}{$dn}[ENTRY]->replace(%{ $opts{'replace'} })
      if (defined $opts{'replace'} && (ref $opts{'replace'} eq 'HASH'));
    return $res;
} ## end sub modify

## @fn LDAP_RESULT add($DN, $attributes)
# creates and add new LDAP Entry addressed by $DN into LDAP and LDS
# @param DN string DN addressing LDS entry
# @param attributes to add to new entry
# @return LDAP result of operation
sub add
{
    my $self = shift;
    my $dn   = shift;
    my %opts = @_;
    my $res  = $self->{'LDAP'}->s_add($dn, %opts);
    if ($res->is_error()) {
        log_debug_printf('Unable to add entry %s into LDAP. Error: %u %s', $dn, $res->code, $res->error);
        return $res;
    }
    my $tmp = Net::LDAP::Entry->new($dn);
    $tmp->add(@{ $opts{'attr'} }) if (defined $opts{'attr'} && (ref $opts{'attr'} eq 'ARRAY'));
    $tmp->add(%{ $opts{'attr'} }) if (defined $opts{'attr'} && (ref $opts{'attr'} eq 'HASH'));
    $self->{'LDS'}{$dn} = _createLDSEntry($tmp);
    _addEntryToLDS($self->{'LDS'}, $dn);

    return $res;
} ## end sub add

## @fn LDAP_RESULT delete($DN)
# deletes LDAP Entry addressed by $DN from LDAP and LDS
# @param DN string DN addressing LDS entry
# @return LDAP result of operation
sub delete
{
    my ($self, $dn) = @_;
    my $res = $self->{'LDAP'}->s_delete($dn);
    if ($res->is_error()) {
        log_debug_printf('Unable to delete entry %s from LDAP. Error: %u %s', $dn, $res->code, $res->error);
        return $res;
    }
    _removeChild($self->{'LDS'}, $dn);
    delete $self->{'LDS'}{$dn};
    return $res;
} ## end sub delete

## @method bool mergeEntryToCache($entry)
# public wrapper for _mergeLDAP function
# @param $entry initialized Net::LDAP::Entry or array of Net::LDAP::Entry objects
# @return 1 on success, undef on failure
sub mergeEntryToCache
{
    my ($self, $entry) = @_;
    if (ref $entry eq 'Array') {
        return _mergeLDAP($self->{'LDS'}, $entry, $self->{'CHANGES_ONLY'});
    }
    else {
        return _mergeEntryToCache($self->{'LDS'}, $entry, $self->{'CHANGES_ONLY'});
    }
} ## end sub mergeEntryToCache

## @method void adjustBase(void)
# find the lowest base for LDS. ie. if base has only one child and base does not have associated entry, this child becomes new base and old base is discarded.
# recursivelly then find lowest base. Purpose is to minimalize LDAP query operations.
# @return void
sub adjustBase
{
    my ($self) = @_;
    #check if current base has only one child
    my $oldbase = $self->{'BASE'};
    log_debug_printf('adjusting base. old base: %s', $oldbase);
    if ($self->{'LDS'}{$oldbase}[ENTRY] or scalar @{ $self->{'LDS'}{$oldbase}[NEXT] } > 1) {
        # go up LDS tree to find top level DN with defined entry
        my $parentDN = (split(/,/, $self->{'BASE'}, 2))[1];
        if ($self->{'LDS'}{$parentDN}[ENTRY]) {
            $self->{'BASE'} = $parentDN;
            return $self->adjustBase();
        }
    }
    elsif (scalar @{ $self->{'LDS'}{$oldbase}[NEXT] } == 1) {
        # go down LDS tree to find first defined entry
        $self->{'BASE'} = $self->{'LDS'}{$oldbase}[NEXT][0];
        delete $self->{'LDS'}{$oldbase};
        return $self->adjustBase();
    }
    else {
        my $entryDN;
        # find first DN with defined entry and adjust this base
        foreach (keys(%{ $self->{'LDS'} })) {
            if ($self->{'LDS'}{$_}[ENTRY]) {
                $self->{'BASE'} = $_;
                return $self->adjustBase();
            }
        }
        return;
    }
    log_debug_printf('adjusting base. new base: %s', $self->{'BASE'});
} ## end sub adjustBase
