#!/usr/bin/perl -w
## @file
# Implementation of POS::LdapUtil

## @class
# LDAP Utils
package POS::LdapUtil;

use strict;

# TODO rewrite so we have basic object structure similar to Ldapworkstation
#    (have new to init general variables (LDAP...) not needed then for all procedures again)

# version  of the package

our $VERSION = 0.1;


# exportieren der Module
require Exporter;
our @ISA = qw(Exporter);

require Net::LDAP;
use Net::LDAP qw(LDAP_NO_SUCH_OBJECT);
use POS::Syslog;
use Unix::Syslog qw(:macros);
use Net::IP;

# Festlegen der zu exportierenden Module, Methoden, Funktionen, Variablen, Konstanten
our @EXPORT =
  qw(ldap_treesearch ldap_splitdn ldap_shiftbase ldap_check_error ldap_entry_has_value ldap_get_dn_entry ldap_get_ref_entry ldap_get_cn_from_dn ldap_dn2dnsname ldap_get_location_entry ldap_cn2addr ldap_calc_ip_addr ldap_cn2addr_ws_stgt ldap_gen_entrycache ldap_treesearch_cached ldap_get_refpc_entry_cached ldap_get_refimage_entry_cached ldap_gen_entrycache_successive ldap_set_values check_boolean_attribute get_ws_from_ldap get_roles_from_ldap get_ids_from_ldap);


# Festlegen der optional zu exportierenden Module, Methoden, Funktionen, Variablen, Kostanten

#@EXPORT_OK;

## @fn $ ldap_treesearch(%args)
# searchs for an entry matching the FILTER, first in STARTDN and then
# continues in cn=standards up to ROOTDN
# @param args hash with the following items
# @arg @c 'LDAP' Net::LDAP connection,
# @arg @c 'STARTDN' DN to start search in
# @arg @c 'ROOTDN' end search at this DN
# @arg @c 'FILTER' LDAP filter (see man ldap_search, RFC 2254)
# @arg @c 'STANDARDCN' first component of DN for search upwards, default: "cn=standards"
# @return Found Net::LDAP::Entry or undef
sub ldap_treesearch
{
    my %param = (
        LDAP       => undef,
        BASE       => undef,
        UPTOBASE   => undef,
        FILTER     => undef,
        STANDARDCN => 'cn=standards',
        @_,
    );

    # todo: params should be checked here

    my $dn   = $param{'BASE'};
    my $ldap = $param{'LDAP'};
    my ($cn, $actualbase) = ldap_splitdn($dn);
    my $result;
    my $entry;

    while (1) {
        # search LDAP for the actual dn
        $result = $ldap->search(base => $dn, scope => 'base', filter => $param{'FILTER'});
        # when searching for a DN with base => $dn, and nothing was found,
        # LDAP-Servers yield a specific error, which is not really an error
        # in this context. Here, the loop must go on.
        if ($result->code and ($result->code != LDAP_NO_SUCH_OBJECT)) {
            ldap_check_error($result, 0);    # don't die
            return undef;
        }
        last if ($result->count > 0);
        last if ($actualbase eq $param{'UPTOBASE'});
        # the following searches look for the specified dn component
        if ($cn ne $param{'STANDARDCN'}) {
            $cn = $param{'STANDARDCN'};
        }
        else {
            # cut off one component from actualbase each time
            $actualbase = ldap_shiftbase($actualbase);
        }
        $dn = $cn . "," . $actualbase;
    }

    # we should have found an entry now
    if ($result->count > 0) {
        # there should be no or one entry
        # this will ignore possibly existing further entries found
        ($entry) = $result->all_entries;
    }
    return $entry;

} ## end sub ldap_treesearch


#not used
## @fn $ ldap_treesearch_cached(%args)
# searchs in given LDAP entry cache for an entry matching STARTDN and then
# continues searching for cn=standards up to ROOTDN
# @param args hash with the following items
# @arg @c 'ENTRYCACHE' a hash conaining DN => Net::LDAP::Entry, see ldap_gen_entrycache()
# @arg @c 'STARTDN' DN to start search in
# @arg @c 'ROOTDN' end search at this DN
# @arg @c 'STANDARDCN' first component of DN for search upwards, default: "cn=standards"
# @return Found Net::LDAP::Entry or undef
sub ldap_treesearch_cached
{
    my %param = (
        ENTRYCACHE => undef,
        BASE       => undef,
        UPTOBASE   => undef,
        STANDARDCN => 'cn=standards',
        @_,
    );

    # todo: params should be checked here

    my $dn         = $param{'BASE'};
    my $entrycache = $param{'ENTRYCACHE'};
    my ($cn, $actualbase) = ldap_splitdn($dn);
    my $result;
    my $entry;

    while (1) {
        ## search LDAP for the actual dn
        # lookup actual dn in entrycache
        #$result=$ldap->search( base => $dn,
        #       scope => 'base',
        #		filter => $param{'FILTER'});
        # when searching for a DN with base => $dn, and nothing was found,
        # LDAP-Servers yield a specific error, which is not really an error
        # in this context. Here, the loop must go on.
        #if ( $result->code and ($result->code != LDAP_NO_SUCH_OBJECT) ) {
        #	ldap_check_error($result, 0); # don't die
        #	return undef;
        #}
        if (exists $entrycache->{$dn}) {
            $entry = $entrycache->{$dn};
            last;
        }
        last if ($actualbase eq $param{'UPTOBASE'});
        # the following searches look for the specified dn component
        if ($cn ne $param{'STANDARDCN'}) {
            $cn = $param{'STANDARDCN'};
        }
        else {
            # cut off one component from actualbase each time
            $actualbase = ldap_shiftbase($actualbase);
        }
        $dn = $cn . "," . $actualbase;
    }

    # now we have either found an entry, then $entry points to
    # the Net::LDAP::Entry Object, otherwise, $entry is undef
    return $entry;
} ## end sub ldap_treesearch_cached


## @fn $ ldap_shiftbase($dn)
# Given a dn, this function returns the appropriate base by removing
# the first component of the dn string.
# More generic, from a comma separated string, this function removes
# the first part.
# @param $dn DN string
# @return DN string with the first component removed
sub ldap_shiftbase
{
    # given a dn, this function returns the appropriate base by removing
    # the first component of the dn string
    # more generic, from a comma separated string, this function removes
    # the first part
    my ($dn) = @_;
    my @base = split(/,/, $dn);
    shift(@base);
    return join(",", @base);
} ## end sub ldap_shiftbase


## @fn $ ldap_check_error($result, $die_on_error)
# check result from ldap, log an error message and eventually die
# @param $result a reference to Net::LDAP::Search object returned by search() method
# @param $die_on_error show callstack and die on error
# @return the $result argument
sub ldap_check_error
  # check LDAP result for error, send error message to syslog
  # and die with error message
{
    my $result = "";
    my ($ldap_result, $die_on_error) = @_;
    if ($ldap_result->code != 0) {
        $result = $ldap_result->error;
        log_alert_print($ldap_result->error);
        if ($die_on_error) {
            show_callstack();
            die $ldap_result->error;
        }
    }
    return $result;
} ## end sub ldap_check_error

## @fn @ ldap_splitdn($dn)
# Split a given dn into its first part (cn=...) and the remaining parts
# and return it as a list of two strings
# so that it can be used as filter and base in an LDAP search
# @param $dn DN string
# @return list of the first and second part of DN
sub ldap_splitdn
{
    # split a given dn into its first part (eg. cn=...) and the remaining parts
    # and return it as a list of two strings
    # so that it can be used as filter and base in an LDAP search
    # Example:
    # "cn=refimagetest01,ou=global,o=organisation,c=DE"
    # => ("cn=refimagetest01", "ou=global,o=organisation,c=DE")
    # Parameter: dn
    my ($dn) = @_;

    my @result = ();
    my @splitdn = split(/,/, $dn);
    push(@result, shift(@splitdn));
    push(@result, join(",", @splitdn));
    return @result;
} ## end sub ldap_splitdn


## @fn $ ldap_entry_has_value(%args)
# checks if an entry contains the given value (multivalue is checked against each particular value)
# @param args hash with the following items
# @arg @c 'ENTRY' a Net::LDAP::Entry reference
# @arg @c 'ATTR_NAME' attribute name
# @arg @c 'VALUE' the value to check for
# @return 1 if the value was found, 0 otherwise
sub ldap_entry_has_value
{
    my %param = (
        ENTRY     => undef,
        ATTR_NAME => undef,
        VALUE     => undef,
        @_,
    );

    my $result = 0;

    my $ldap_entry = $param{'ENTRY'};

    my @values = $ldap_entry->get_value($param{'ATTR_NAME'});
    my ($grep_result) = grep { $_ eq $param{'VALUE'} } @values;
    if (defined $grep_result && $grep_result eq $param{'VALUE'}) {
        $result = 1;
    }
    return $result;
} ## end sub ldap_entry_has_value


## @fn $ ldap_get_dn_entry(%args)
# get a LDAP entry for given DN
# @param args hash with the following items
# @arg @c 'LDAP' a Net::LDAP connection
# @arg @c 'DN' DN string
# @arg @c 'FILTER' LDAP filter that allows limiting the results, default: '(objectclass=*)'
# @return Net::LDAP::Entry reference or undef
sub ldap_get_dn_entry
{
    my %param = (
        LDAP   => undef,
        DN     => undef,
        FILTER => '(objectclass=*)',
        @_,
    );

    my $ldap = $param{'LDAP'};
    my $entry;
    my $result = $ldap->search(base => $param{'DN'}, scope => 'base', filter => $param{'FILTER'});
    # looking up DNs directly as base yields an error-code, if not found.
    if ($result->code and ($result->code != LDAP_NO_SUCH_OBJECT)) {
        ldap_check_error($result, 0);    # don't die
        return undef;
    }
    # we should have found an entry now
    if ($result->count > 0) {
        # there should be no or one entry
        # this will ignore possibly existing further entries found
        ($entry) = $result->all_entries;
    }
    return $entry;
} ## end sub ldap_get_dn_entry


#not used
sub ldap_get_ref_entry
{
    my %param = (
        LDAP    => undef,
        DN      => undef,
        ROOTDN  => undef,
        FILTER  => '(objectclass=*)',
        REFATTR => undef,
        @_,
    );

    my $ldap = $param{'LDAP'};
    my $entry;
    my $ref_entry;
    my $ref_dn;

    # get Entry of given DN
    my $result = $ldap->search(base => $param{'DN'}, scope => 'base', filter => $param{'FILTER'});
    # looking up DNs directly as base yields an error-code, if not found.
    if ($result->code and ($result->code != LDAP_NO_SUCH_OBJECT)) {
        my $error_msg = ldap_check_error($result, 0);    # don't die
                                                         #print STDERR "LDAP-error:$error_msg\n"; # debug
        return undef;
    }
    # check if an entry was found
    #print STDERR "Found ", $result->count, " entries\n"; # debug
    if ($result->count > 0) {
        # there should be no or one entry
        # this will ignore possibly existing further entries found
        ($entry) = $result->all_entries;

        # check if the referenceDN-attribute exists and is not empty
        if ($entry->exists($param{'REFATTR'})) {
            $ref_dn = $entry->get_value($param{'REFATTR'});
        }
    }

    # if no referenceDN was found in the given entry, do a treesearch
    # downwards
    if (!defined $ref_dn) {
        # begin search with standards in actual container
        my ($cn, $base) = ldap_splitdn($param{'DN'});
        $ref_entry = ldap_treesearch(
            LDAP    => $ldap,
            STARTDN => 'cn=standards,' . $base,
            ROOTDN  => $param{'ROOTDN'},
            FILTER  => $param{'FILTER'}
        );

        if (defined $ref_entry) {
            $ref_dn = $ref_entry->get_value($param{'REFATTR'});
        }
    }

    # we should now have a valid reference-DN
    $entry = undef;
    $entry = ldap_get_dn_entry(LDAP => $ldap, DN => $ref_dn);

    return $entry;
} ## end sub ldap_get_ref_entry


#not used
sub ldap_get_refpc_entry_cached
{
    my %param = (
        WORKSTATION_ENTRYCACHE => undef,
        DSAP_ENTRYCACHE        => undef,
        REFPC_ENTRYCACHE       => undef,
        WORKSTATIONDN          => undef,
        ROOTDN                 => undef,
        REFATTR                => undef,
        @_,
    );

    my $ws_cache    = $param{'WORKSTATION_ENTRYCACHE'};
    my $dsap_cache  = $param{'DSAP_ENTRYCACHE'};
    my $refpc_cache = $param{'REFPC_ENTRYCACHE'};
    my $entry;
    my $ref_entry;
    my $ref_dn;

    # get Entry of given DN
    if (!exists $ws_cache->{ $param{'WORKSTATIONDN'} }) {
        return undef;
    }
    $entry = $ws_cache->{ $param{'WORKSTATIONDN'} };
    # check if the referenceDN-attribute exists and is not empty
    if ($entry->exists($param{'REFATTR'})) {
        $ref_dn = $entry->get_value($param{'REFATTR'});
    }

    # if no referenceDN was found in the given entry, do a treesearch
    # downwards
    if (!defined $ref_dn) {
        # begin search with standards in actual container
        my ($cn, $base) = ldap_splitdn($param{'WORKSTATIONDN'});
        #$ref_entry = ldap_treesearch(
        #		LDAP => $ldap,
        #		STARTDN => 'cn=standards,' . $base,
        #		ROOTDN => $param{'ROOTDN'},
        #		FILTER => $param{'FILTER'}
        #);
        $ref_entry = ldap_treesearch_cached(
            ENTRYCACHE => $dsap_cache,
            STARTDN    => 'cn=standards,' . $base,
            ROOTDN     => $param{'ROOTDN'},
        );

        if (defined $ref_entry) {
            $ref_dn = $ref_entry->get_value($param{'REFATTR'});
        }
    }

    # we should now have a valid reference-DN
    $entry = undef;
    $entry = $refpc_cache->{$ref_dn};

    return $entry;
} ## end sub ldap_get_refpc_entry_cached


#not used
sub ldap_get_refimage_entry_cached
{
    my %param = (
        WORKSTATION_ENTRYCACHE => undef,
        DSAP_ENTRYCACHE        => undef,
        REFIMAGE_ENTRYCACHE    => undef,
        WORKSTATIONDN          => undef,
        ROOTDN                 => undef,
        REFATTR                => undef,
        @_,
    );

    my $ws_cache       = $param{'WORKSTATION_ENTRYCACHE'};
    my $dsap_cache     = $param{'DSAP_ENTRYCACHE'};
    my $refimage_cache = $param{'REFIMAGE_ENTRYCACHE'};
    my $entry;
    my $ref_entry;
    my $ref_dn;

    # get Entry of given DN
    if (!exists $ws_cache->{ $param{'WORKSTATIONDN'} }) {
        return undef;
    }
    $entry = $ws_cache->{ $param{'WORKSTATIONDN'} };
    # check if the referenceDN-attribute exists and is not empty
    if ($entry->exists($param{'REFATTR'})) {
        $ref_dn = $entry->get_value($param{'REFATTR'});
    }

    # if no referenceDN was found in the given entry, do a treesearch
    # downwards
    if (!defined $ref_dn) {
        # begin search with standards in actual container
        my ($cn, $base) = ldap_splitdn($param{'WORKSTATIONDN'});
        #$ref_entry = ldap_treesearch(
        #		LDAP => $ldap,
        #		STARTDN => 'cn=standards,' . $base,
        #		ROOTDN => $param{'ROOTDN'},
        #		FILTER => $param{'FILTER'}
        #);
        $ref_entry = ldap_treesearch_cached(
            ENTRYCACHE => $dsap_cache,
            STARTDN    => 'cn=standards,' . $base,
            ROOTDN     => $param{'ROOTDN'},
        );

        if (defined $ref_entry) {
            $ref_dn = $ref_entry->get_value($param{'REFATTR'});
        }
        else {
            log_notice_print("Ref Entry for " . $param{'WORKSTATIONDN'} . " not found");
        }
    }

    # we should now have a valid reference-DN
    $entry = undef;
    $entry = $refimage_cache->{$ref_dn};

    return $entry;
} ## end sub ldap_get_refimage_entry_cached


sub ldap_get_cn_from_dn
{
    my ($dn) = @_;

    my $cn;

    my ($cnpart) = ldap_splitdn($dn);
    if (defined $cnpart) {
        $cnpart = lc($cnpart);
        if ($cnpart =~ /^cn=(.+)$/) {
            $cn = $1;
        }
    }
    return $cn;
} ## end sub ldap_get_cn_from_dn


sub ldap_dn2dnsname
{
    my ($dn) = @_;
    my $dns_name;

    my @dn_parts = split(/,/, $dn);
    if (@dn_parts > 0) {
        my @dns_name_parts;
        foreach my $dn_part (@dn_parts) {
            $dn_part =~ /.+=(.+)/;
            push(@dns_name_parts, $1);
        }
        $dns_name = join('.', @dns_name_parts);
    }
    return $dns_name;
} ## end sub ldap_dn2dnsname


sub ldap_cn2addr
{
    # a workstations IP-address is defined in its cn entry,
    # consisting of "pc" + and a number padded with zeroes to 5 digits
    # this must be added to the beginning of the fixed range for this subnet
    # to get the IP.

    # Parameters:
    # - cn: contains the workstations IP relative to the fixed range
    # - firstipaddr: first IP of fixed Range

    my ($cn, $firstipaddr) = @_;
    my $newipaddr;

    #defined($cn) || die "cn2addr(): Parameter cn not defined";

    if (!defined($cn)) {
        log_alert_print('ldap_cn2addr: Parameter cn not defined');
        return undef;
    }

    #defined($firstipaddr) || die "cn2addr(): Parameter firstipaddr not defined";
    if (!defined($firstipaddr)) {
        log_alert_print('ldap_cn2addr: Parameter firstipaddr not defined');
        return undef;
    }

    # convert cn to number by ignoring the first two characters (usually "pc")
    my $cni = substr($cn, 2, 5);
    # calculate the workstations ip by adding $cni to $firstip using Net::IP
    # this is done by converting the first ip of the range to integer, adding the
    # workstation number and converting the result back to an ip address string.

    ## TODO: this is inefficient, check the IP more efficient!
    #my $firstip = new Net::IP ("$firstipaddr");
    if (!check_ip($firstipaddr)) {
        my $ip_error = Net::IP::Error();
        log_alert_printf('ldap_cn2addr: %s is invalid', $firstipaddr);
        return undef;
    }


    ## this is not very efficient:
    #$newipaddr = Net::IP::ip_bintoip( Net::IP::ip_inttobin( ($firstip->intip())+$cni ,4),4);
    #return $newipaddr;

    ## this is much faster (thanks to holgerm):
    return join(
        ".",    # Result is seperated with dots
        unpack(
            "C4",    # a networkaddress becomes a quadruple of unsigned chars
            pack(
                "N",    # an integer becomes a network address

                unpack(
                    "N",    # a nettwork address becomes an integer
                    pack("C4", split(/\./, $firstipaddr))
                  )         # ip becomes a network address
                  + $cni
            )
        )
    );

} ## end sub ldap_cn2addr


sub ldap_cn2addr_ws_stgt
{
    # a workstations IP-address is defined in its cn entry,
    # consisting of "ws0" + one hex digit + 2 decimal digits (zero padded)
    # where the hex digit is the 3rd octet and the 2 decimal digits are
    # the 4th octet of the workstations IP address. The first two octets
    # are taken from the network address.

    # Parameters:
    # - cn: contains the workstations IP relative to the fixed range
    # - network_addr: Network Address

    my ($cn, $network_addr) = @_;
    my $newipaddr;

    #defined($cn) || die "cn2addr(): Parameter cn not defined";

    if (!defined($cn)) {
        log_alert_print('ldap_cn2addr: Parameter cn not defined');
        return undef;
    }

    if (!defined($network_addr)) {
        log_alert_print('ldap_cn2addr: Parameter network_addr not defined');
        return undef;
    }

    if ($cn =~ /(ws0)([0-9A-Fa-f])(\d{2})/) {
        my @network_split = split(/\./, $network_addr);
        $newipaddr = join(".", $network_split[0], $network_split[1], hex($2), $3 + 0);
    }
    else {
        log_alert_printf('ldap_cn2addr_ws_stgt: %s in network %s is invalid', $cn, $network_addr);
        return undef;
    }

    return $newipaddr;
} ## end sub ldap_cn2addr_ws_stgt


sub ldap_get_location_entry
{
    my %param = (
        'LDAP' => undef,
        'DN'   => undef,
        @_,
    );

    my $location_entry;
    my $ldap_conn     = $param{'LDAP'};
    my $location_base = ldap_shiftbase($param{'DN'});
    my $location_dn   = 'cn=standards,' . $location_base;

    $location_entry = ldap_get_dn_entry(LDAP => $ldap_conn, DN => $location_dn, FILTER => '(objectClass=scLocation)');

    return $location_entry;
} ## end sub ldap_get_location_entry


sub ldap_calc_ip_addr
{
    my %param = (
        'DN'        => undef,
        'LOC_ENTRY' => undef,
        @_,
    );

    my $ip_addr;
    my $location_entry = $param{'LOC_ENTRY'};
    my $cn_part        = ldap_get_cn_from_dn($param{'DN'});
    my $network_addr   = $location_entry->get_value('ipNetworkNumber');
    $ip_addr = ldap_cn2addr($cn_part, $network_addr);
    if (!defined $ip_addr) {
        log_alert_printf('ldap_calc_ip_addr: could not get IP-Address for %s', $cn_part);
    }
    return $ip_addr;
} ## end sub ldap_calc_ip_addr


## note: does NOT use cache
## @fn $ ldap_get_dn_entry(%args)
# get a hash in format DN => Net::LDAP::Entry that contains all entries under the given base
# @param args hash with the following items
# @arg @c 'LDAP' a Net::LDAP connection
# @arg @c 'BASE' base DN string
# @arg @c 'OBJECTCLASS' object class of the wanted entries, i.e. 'scWorkstation'
# @arg @c 'ATTRLIST' an array reference with a list of wanted attributes, all attributes are fetched if empty
# @return hash in format DN => Net::LDAP::Entry
sub ldap_gen_entrycache
{
    my %param = (
        'LDAP'        => undef,
        'BASE'        => undef,
        'OBJECTCLASS' => undef,
        'ATTRLIST'    => undef,
        @_,
    );

    my %entrycache;
    my $ldap   = $param{'LDAP'};
    my $result = $ldap->search(
        base   => $param{'BASE'},
        scope  => 'sub',
        filter => 'objectClass=' . $param{'OBJECTCLASS'},
        attrs  => $param{'ATTRLIST'}
    );
    ldap_check_error($result, 1);

    foreach my $entry ($result->entries) {
        $entrycache{ $entry->dn } = $entry;
    }
    return \%entrycache;
} ## end sub ldap_gen_entrycache


# reads all objects in depth one from BASE to UPTOBASE consistent with FILTER
# works with cache
sub ldap_gen_entrycache_successive
{
    my %param = (
        'LDAP'     => undef,
        'CACHE'    => undef,
        'BASE'     => undef,
        'UPTOBASE' => undef,
        #        'FILTEROBJ' => undef, #simple filter, used for cache search instead of FILTER
        'FILTER'   => undef,
        'ATTRLIST' => undef,
        @_,
    );

    my %entrycache;
    my $cache    = $param{'CACHE'};
    my $ldap     = $param{'LDAP'};
    my $base     = $param{'BASE'};
    my $filter   = $param{'FILTER'};
    my $uptobase = $param{'UPTOBASE'};
    return \%entrycache if !defined $base;
    $uptobase = $base if !defined $uptobase;
    # inside the loop, all bases including the Root Base must be used.
    # CAUTION: this assummes that the Start Base and Root Base end
    # in the same string, this is not checked here.
    while (length($base) >= length($uptobase)) {
        my @objlist;
        if (defined $cache) {
            my $lc = $cache->{'LDAPCACHE'};
            #@objlist=map($lc->getLDAPentry($_), @{$lc->getChildrenByClass($base,$param{'FILTEROBJ'})});
            #my @nobjlist=map($lc->getLDAPentry($_), @{$lc->getChildren($base)});
            my $obj_ref = $lc->FilterEntries($lc->getChildren($base), $filter);
            @objlist = map($lc->getLDAPentry($_), @{$obj_ref});
        }
        else {
            my $result =
              $ldap->search(base => $base, scope => 'one', filter => $param{'FILTER'}, attrs => $param{'ATTRLIST'});
            ldap_check_error($result, 1);
            @objlist = $result->entries();
        }
        foreach my $entry (@objlist) {
            $entrycache{ $entry->dn } = $entry;
        }
        # remove first component from base
        my $shiftbase = ldap_shiftbase($base);
        last if ($shiftbase eq $base);    #failsafe
        $base = $shiftbase;
    }
    return \%entrycache;
} ## end sub ldap_gen_entrycache_successive


sub ldap_set_values
{
    my %param = (
        'LDAP'   => undef,
        'DN'     => undef,
        'VALUES' => undef,
        'ENTRY'  => undef,
        @_,
    );

    my $ldap   = $param{'LDAP'};
    my $entry  = $param{'ENTRY'};
    my $values = $param{'VALUES'};
    my $dn     = $param{'DN'};

    my $objectclass;
    my $do_replace = 0;

    # check if Values Hash specifies an objectClass
    if ($values->{'objectClass'}) {
        $objectclass = $values->{'objectClass'};
    }
    else {
        $do_replace = 1;
    }

    # if no Entry was given, we must do a search to determine
    # the objectClasses of the Entry
    if ($objectclass) {
        if (!defined $entry) {
            my $res = $ldap->search(base => $dn, scope => 'base', filter => 'objectClass=*', attrs => ['objectClass']);
            if ($res->code) {
                log_warn_printf("ldap_set_values(): %s not found in LDAP", $dn);
                return 0;
            }
            ($entry) = $res->entries;
        }
        # check if the entry already has the required objectClass
        if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => $objectclass)) {
            $do_replace = 1;
        }

    }    # if ( $objectclass )

    # now do the LDAP operation
    my $result;
    if ($do_replace) {
        # if we do a replace, possibly existing objectclass must be
        # removed from the values hash
        if ($values->{'objectClass'}) {
            delete $values->{'objectClass'};
        }
        $result = $ldap->modify($dn, replace => $values);
    }
    else {
        # add must be done
        $result = $ldap->modify($dn, add => $values);
    }
    return $result->code;
} ## end sub ldap_set_values

sub show_callstack
{
    my $i = 0;
    print "\nCallstack:\n";
    while (
        my ($package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask) =
        caller($i++)) {
        print "$package $filename $line $subroutine\n";
    }
} ## end sub show_callstack

sub check_ip
{
    ## check a given IP Address
    my $ip_addr = shift;
    my @ip_numbers = split(/\./, $ip_addr);
    if (@ip_numbers != 4) {
        $ip_addr = undef;
        return $ip_addr;
    }
    my $number = undef;
    foreach $number (@ip_numbers) {
        if ($number < 0 || $number > 255) {
            return undef;
        }
    }
    return $ip_addr;
} ## end sub check_ip

#pos11 sp1
## @fn $ check_boolean_attribute($result)
# checks if boolean parameter $attr in the $dn is TRUE or FALSE
# @param $ldap Net::LDAP connection
# @param $dn DN to  search in
# @param $attr attribute to check
# @return 0 for true, 1 for false, other value for error
sub check_boolean_attribute
{
    my %param = (
        'LDAP' => undef,
        'DN'   => undef,
        'ATTR' => undef,
        @_,
    );
    my $ldap = $param{'LDAP'};
    my $dn   = $param{'DN'};
    my $attr = $param{'ATTR'};

    #my ($ldap, $dn, $attr)  = @_;
    my $ret = 0;
    #my $ldap=$param{'LDAP'};
    ### testing if scDynamicIp is TRUE
    my $result = $ldap->search(base => $dn, filter => '(objectClass=*)', scope => 'base');
    #$self->{'LDAP'}->search(base => $dn, filter => '(objectClass=*)', scope => 'base');

    if ($result->count == 1) {
        my $entry  = $result->pop_entry();
        my $bvalue = $entry->get_value($attr);
        if (defined $bvalue) {
            if (uc($bvalue) eq 'TRUE') {
                log_info_printf('%s is set to \'TRUE\' for %s', $attr, $dn);
            }
            else {
                log_info_printf('%s set to \'FALSE\' for %s', $attr, $dn);
                $ret = 1;
            }
        }
        else {
            log_info_printf('%s not defined for %s', $attr, $dn);
            $ret = 2;
        }
    }
    else {
        log_info_printf('%s not found under %s, assuming \'FALSE\', possible LDAP misconfiguration?', $attr, $dn);
        $ret = -1;
    }
    return $ret;
} ## end sub check_boolean_attribute


#pos11 sp2
## @fn $ check_boolean_attribute($result)
# checks if boolean parameter $attr in the $dn is TRUE or FALSE, using cached LDAP tree
# @param $cldap cashed LDAP hash tree
# @param $dn DN to  search in
# @param $attr attribute to check
# @return 0 for true, 1 for false, other value for error

sub check_boolean_attribute_cached
{
    my %param = (
        'CACHE' => undef,
        'DN'    => undef,
        'ATTR'  => undef,
        @_,
    );
    my $cache = $param{'CACHE'};
    my $dn    = $param{'DN'};
    my $attr  = $param{'ATTR'};
    #my ($ldap, $dn, $attr)  = @_;
    #my $ldap=$param{'LDAP'};
    ### testing if scDynamicIp is TRUE
    #$self->{'LDAP'}->search(base => $dn, filter => '(objectClass=*)', scope => 'base');
    if (!exists $cache->{'LDAPTREE'}{$dn}) {
        log_warn_printf('dn %s not found in ldap cache, incorrect reference?', $dn);
        return -1;
    }
    my $entry  = $cache->{'LDAPTREE'}{$dn}[0];
    my $bvalue = $entry->get_value($attr);
    my $bint   = boolean_to_integer($bvalue);
    if (defined $bint) {
        if ($bint) {
            log_info_printf('%s is set to \'TRUE\' for %s', $attr, $dn);
            return 0;
        }
        else {
            log_info_printf('%s set to \'FALSE\' for %s', $attr, $dn);
            return 1;
        }
    }
    else {
        log_info_printf('%s not defined for %s', $attr, $dn);
        return 2;
    }
} ## end sub check_boolean_attribute_cached


# FALSE=0
# TRUE=1
sub boolean_to_integer
{
    my ($boolparam) = @_;
    if (defined $boolparam) {
        (uc($boolparam) eq 'TRUE') ? return 1 : return 0;
    }
    return undef;
} ## end sub boolean_to_integer


#returns list of (only not deactivated!) workstations in ldap
# entries keyed by MAC
# can create list from both cached ldap (CACHE) or LDAP object,
# if CACHE present, uses cache, if CACHE undef and LDAP present, uses direct access,
#  otherwise error
# returns reference to active (scDeactivated=FALSE), and nonactive (scDeactivated=TRUE) workstations
sub get_ws_from_ldap
{
    my %param = (
        'LDAP'   => undef,
        'BASE'   => undef,
        'ACTIVE' => 1,
        @_,
    );
    my $ldap  = $param{'LDAP'};
    my $cache = $param{'CACHE'};
    my $base  = $param{'BASE'};

    my @wslist;          #list of all workstations
    my %wshash_act;      #hash by cn
    my %wshash_deac;     #hash by cn
    my %wshash_all;      #hash by cn
    my %wshash_bymac;    #hash by $mac
    my %wshash_byip;     #hash by ip address
    if (defined $cache) {
        my $lc = $cache->{'LDAPCACHE'};
        @wslist = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($base, 'scWorkstation') });
    }
    elsif (defined $ldap) {
        my $wsresult =
          $ldap->search(base => $base, filter => '(objectClass=scWorkstation)', scope => 'one', attrs => []);
        if (defined $wsresult) {
            @wslist = $wsresult->entries();
        }
    }
    else {
        log_alert_print("Warning, neither CACHE or LDAP reference supplied, aborting.");
        return undef;
    }
    #process ws, make hash hashed by MAC for active and by cn for deactivated
    foreach my $ws (@wslist) {
        my $cname = $ws->get_value('cn');
        $wshash_all{$cname} = $ws;
        my $ip = $ws->get_value('ipHostNumber');
        $wshash_byip{$ip} = $ws if defined $ip;
        if (!$ws->get_value('macAddress')) {    #empy same as FALSE
            $wshash_deac{$cname} = $ws;
        }
        else {                                  #if more macs, multiple hashes points to same object
            $wshash_act{$cname} = $ws;
            my @mac = $ws->get_value('macAddress');
            foreach my $mac (@mac) {
                if (defined $wshash_bymac{$mac}) {
                    my $dn_dws = $wshash_bymac{$mac}->dn();
                    log_warn_printf("Warning, collision in mac $mac between '%s' and '%s', will ignore latter. ",
                        $ws->dn(), $dn_dws);
                }
                $wshash_bymac{$mac} = $ws;
            }
        }
    }
    return (\%wshash_act, \%wshash_deac, \%wshash_all, \%wshash_bymac, \%wshash_byip);
} ## end sub get_ws_from_ldap


# TODO: try to join into general subroutine with 'get_roles_from_ldap' (leave just ENTRY form)
#reads all cashregisters not under scRole, and under scRefObjectContainer
sub get_crs_from_ldap
{
    my %param = (
        'LDAP'     => undef,
        'BASE'     => undef,
        'UPTOBASE' => undef,
        @_,
    );
    my $ldap  = $param{'LDAP'};
    my $cache = $param{'CACHE'};


    my %crhash;    # put there info on ordering also
    my $results_hr =
      POS::LdapUtil::ldap_gen_entrycache_successive( #search all subtree..  from LdapUtil.pmother scLocations should not be visible! check..
        'LDAP'     => $ldap,
        'CACHE'    => $cache,
        'BASE'     => $param{'BASE'},
        'UPTOBASE' => $param{'UPTOBASE'},
        'FILTER'   => '(objectClass=scRefObjectContainer)',
        'ATTRLIST' => []                                      #dn is enough
      );
    foreach my $resultdn (keys %{$results_hr}) {
        my @contdepth = split(',', $resultdn);
        my $contdepth = @contdepth;
        my @crentries;
        if (defined $cache) {
            my $lc = $cache->{'LDAPCACHE'};
            @crentries = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($resultdn, 'scCashRegister') });
        }
        elsif (defined $ldap) {
            my $crresult = $ldap->search(base => $resultdn, filter => '(objectClass=scCashRegister)', scope => 'one');
            @crentries = $crresult->entries();
        }
        foreach my $cr (@crentries) {
            my $dn = $cr->dn();
            $crhash{$dn}{'ENTRY'} = $cr;    #Net::LDAP::Entry object
            $crhash{$dn}{'DEPTH'} =
              $contdepth;    #insert depth of the container into the hash for potential priority assistance
            $crhash{$dn}{'dn'} = $dn;
            $crhash{$dn}{'cn'} = $cr->get_value('cn');    #insert other data directly to simplify parsing
        }
    }
    return (\%crhash);
} ## end sub get_crs_from_ldap


# mainly used by posLeaseMonitor.pm
# returns two list, (keyed by ids and indices)
# params: LDAP,ROOTBASE
# index
sub get_ids_from_ldap
{
    my %param = (
        'LDAP' => undef,
        'BASE' => undef,
        @_,
    );
    my $ldap  = $param{'LDAP'};
    my $cache = $param{'CACHE'};
    my $base  = $param{'BASE'};


    my %ideshash;
    my %invidhash;
    my $entry;


    if (defined $cache) {
        my $lc = $cache->{'LDAPCACHE'};
        $entry = $lc->getLDAPentry($base);
    }
    elsif (defined $ldap) {
        my $idpoolresult =
          $ldap->search(base => $base, filter => '(objectClass=scLocation)', scope => 'base', attrs => ['scIdPool']);
        if (defined $idpoolresult) {    #assume only one scLoc exists
            $entry = $idpoolresult->pop_entry();
        }
    }
    my @idpool = $entry->get_value('scIdPool');
#id's be put into hash, with each line as a hash value (ordering of lines is of course not guarantied to correspond to LDAP)
    my $multilineidx = 0;
    my $index        = 0;               #only different
    my $fullindex    = 0;               #including multiplicities
    foreach my $entry (@idpool) {
        my @ids = split(';', $entry);
        foreach my $id (@ids) {
            if (defined $ideshash{$id}) {
                $ideshash{$id}{'COUNT'}++;
            }
            else {
                $ideshash{$id}{'INDEX'}        = $index;
                $ideshash{$id}{'MULTILINEIDX'} = $multilineidx;
                $index++;
                $ideshash{$id}{'COUNT'} = 1;
            }
            $invidhash{$fullindex}{'ID'}           = $id;
            $invidhash{$fullindex}{'FULLINDEX'}    = $fullindex;
            $invidhash{$fullindex}{'MULTILINEIDX'} = $multilineidx;
            $fullindex++;
        }
        $multilineidx++;
    }
    return (\%invidhash, \%ideshash);
} ## end sub get_ids_from_ldap


sub ldap_search_parent
{
    my %param = (
        LDAP    => undef,
        STARTDN => undef,
        ROOTDN  => undef,
        FILTER  => undef,
        @_,
    );

    # todo: params should be checked here

    my $dn   = $param{'STARTDN'};
    my $ldap = $param{'LDAP'};
    my $result;
    my $entry;

    while (1) {
        # search LDAP for the actual dn
        $result = $ldap->search(base => $dn, scope => 'base', filter => $param{'FILTER'});
        if ($result->code and ($result->code != LDAP_NO_SUCH_OBJECT)) {
            ldap_check_error($result, 0);    # don't die
            return undef;
        }
        last if ($result->count > 0);
        last if ($dn eq $param{'ROOTDN'});
        $dn = ldap_shiftbase($dn);
    }

    # we should have found an entry now
    if ($result->count > 0) {
        # there should be no or one entry
        # this will ignore possibly existing further entries found
        ($entry) = $result->all_entries;
    }
    return $entry;

} ## end sub ldap_search_parent


#not to be exported, auxiliary to refresh_roleconfig
#params LDAP,ROOTBASE,STARTBASE
#requires LDAP or CACHE
sub get_roles_from_ldap
{
    my %param = (
        'LDAP'     => undef,
        'CACHE'    => undef,
        'BASE'     => undef,
        'UPTOBASE' => undef,
        'SCLOCDN'  => undef,    #dn of scLocation, defines the dn divide to mark roles as local or global
        @_,
    );
    my $ldap  = $param{'LDAP'};
    my $cache = $param{'CACHE'};
    my $scloc = $param{'SCLOCDN'};
    $scloc = $param{'BASE'} if !defined $scloc;

    my %roleshash;              # put there info on ordering also
                                #read all accresible role references from ldap
                                #log_alert_print('role parsing ldap now');
    my $results =
      POS::LdapUtil::ldap_gen_entrycache_successive( #search all subtree..  from LdapUtil.pmother scLocations should not be visible! check..
        'LDAP'  => $ldap,
        'CACHE' => $cache,
        'BASE'  => $param{'BASE'},                   #deepest level (scLocation (==ROOTBASE))
        'UPTOBASE' =>
          $param{'UPTOBASE'},    #most upper level (ALL_LDAPBASE if ALLOWGLOBALROLES, otherwise scLocation (ROOTBASE))
        'FILTER'   => '(objectClass=scRefObjectContainer)',
        'ATTRLIST' => []                                      #dn is enough
      );
    #my @sresults=sort { &POS::LdapUtil::ldap_comparedn($a,$b)} keys %{$results};
    foreach my $resultdn (keys %{$results}) {
        my @contdepth = split(',', $resultdn);
        my $contdepth = @contdepth;
        my @roles;
        #printf $result;
        if (defined $cache) {
            my $lc = $cache->{'LDAPCACHE'};
            @roles = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($resultdn, 'scRole') });
        }
        else {
            my $roleresult = $ldap->search(base => $resultdn, filter => '(objectClass=scRole)', scope => 'one');
            @roles = $roleresult->entries();
        }
        foreach my $role (@roles) {
            my $dn = $role->dn();
            $roleshash{$dn}{'ENTRY'} = $role;    #Net::LDAP::Entry object
            $roleshash{$dn}{'DEPTH'} =
              $contdepth;    #insert depth of the container into the hash for potential priority assistance
            $roleshash{$dn}{'dn'} = $dn;                       #useful, e.g. for rolefile write
            $roleshash{$dn}{'cn'} = $role->get_value('cn');    #insert other data directly to simplify parsing
            $roleshash{$dn}{'scRoleName'}        = $role->get_value('scRoleName');
            $roleshash{$dn}{'scRoleDescription'} = $role->get_value('scRoleDescription');
            $roleshash{$dn}{'scRoleDescription'} = '' if !defined $roleshash{$dn}{'scRoleDescription'};

            $roleshash{$dn}{'GLOBAL'} = 1;
            $roleshash{$dn}{'GLOBAL'} = 0 if $dn =~ /$scloc$/;

            #!ldap_search_parent(
            #    'LDAP'    => $ldap,
            #    'STARTDN' => $dn,
            #    'ROOTDN'  => $rootbase,
            #    'FILTER'  => '(objectClass=scLocation)'
            #);

        }
    }
    #log_alert_print('role parsing done');

    #compute ordering ({'INDWEX'})
    my %depthhash;    #hash by depth
    foreach my $dn (keys %roleshash) {    #create intermediate hash with keys being the depth level
        my $depth = $roleshash{$dn}{'DEPTH'};
        my %tmphash;
        $tmphash{'dn'}                = $dn;
        $tmphash{'cn'}                = $roleshash{$dn}{'cn'};
        $tmphash{'scRoleName'}        = $roleshash{$dn}{'scRoleName'};
        $tmphash{'scRoleDescription'} = $roleshash{$dn}{'scRoleDescription'};
        push @{ $depthhash{$depth} }, \%tmphash;
    }
    my $index = 0;
    foreach my $key (sort { $b <=> $a } (keys(%depthhash)))
    {                                     #lowest level to highest, and within each level by alphabet
        my @svallist = sort {             #within each depth lever, sort by names by alphabet
            $a->{'scRoleName'} cmp $b->{'scRoleName'}
        } @{ $depthhash{$key} };
        foreach my $role (@svallist) {
            $roleshash{ $role->{'dn'} }{'INDEX'} = $index;
            $index++;
        }
    }


    return (\%roleshash);
} ## end sub get_roles_from_ldap


1;
