#!/usr/bin/perl -w
## @file
# Implementation of POS::Pid

## @class
# Create and maintain a PID file in /var/run/Prog.pid
# @param 'PID' process PID, default: $$,
package POS::Pid;

use strict;
use warnings;
use POS::Syslog;

# version of the package
our $VERSION = 0.1;

#main code

##### globals

my $DEFAULT_TIME      = 1;
my $DEFAULT_TIMEOUT   = 10;
my $DEFAULT_DIRECTORY = "/var/run/";


## @cmethod object new(%args)
# @param args hash with the following items
# @arg @c 'PID' process PID, default: $$,
# @arg @c 'TIME' sleep time between retries, default: 1s
# @arg @c 'TIMEOUT' total time for retries, default: 10s
# @arg @c 'FILE' PID file name, default: get_filename()
# @arg @c 'DIRECTORY' pid file directory, default: "/var/run/"
# @arg @c 'PROCNAME' process name, default: get_procname()
# @arg @c 'TOUCH_TIMEOUT' default: 1800
# @return A new object
sub new
{
    my $package = shift;
    $package = ref($package) || $package;
    my $data    = {
        'PID'           => $$,
        'TIME'          => undef,
        'TIMEOUT'       => undef,
        'ERROR'         => 0,
        'ERROR_MSG'     => "",
        'FILE'          => undef,
        'DIRECTORY'     => undef,
        'PIDFILE'       => undef,
        'PROCNAME'      => undef,
        'EXIT'          => 0,
        'TOUCH_TIMEOUT' => 1800,
        'ATIME'         => undef,
        'MTIME'         => undef,
        'CTIME'         => undef,
        'BTIME'         => undef,
    };
    while (@_) {
        my $field = shift;
        if (@_) {
            $data->{$field} = shift;
        }
    }
    if (!$data->{'TIME'}) {
        $data->{'TIME'} = $DEFAULT_TIME;
    }

    if (!defined($data->{'TIMEOUT'})) {
        $data->{'TIMEOUT'} = $DEFAULT_TIMEOUT;
    }
    if (!$data->{'DIRECTORY'}) {
        $data->{'DIRECTORY'} = $DEFAULT_DIRECTORY;
    }

    if (!$data->{'FILE'}) {
        $data->{'FILE'} = get_filename();
    }

    if (!$data->{'PIDFILE'}) {
        $data->{'PIDFILE'} = $data->{'DIRECTORY'} . $data->{'FILE'} . ".pid";
    }

    if (!$data->{'PROCNAME'}) {
        $data->{'PROCNAME'} = get_procname();
    }
    $data->{'BTIME'} = time();
    stat_pid($data);
    create_file($data);
    if ($data->{'ERROR'} && $data->{'EXIT'}) {
        die $data->{'ERROR_MSG'}, "\n";
    }

    return bless($data, $package);
} ## end sub new($;@)


## @fn $ get_filename()
# Returns pid file name derived from executable name $0. Path and file extension are not included.
# @return pid file name or undef
sub get_filename
{
    my @a = split(/\//, $0);

    my ($str, @strs) = split(/\./, $a[@a - 1]);

    if ($str) {
        return $str;
    }
    return;
} ## end sub get_filename ()


## @fn $ get_procname()
# Returns process file name derived from executable name $0. Path is not included, file extension is included.
# @return process file name or undef
sub get_procname
{
    my @a = split(/\//, $0);
    my $procname = pop(@a);

    if ($procname) {
        return $procname;
    }
    return;
} ## end sub get_procname ()


## @method bool create_file()
# Attempts to create a PID file. Sets ERROR to -1 if timeout is reached.
# @return 1 on success, 0 on failure
sub create_file
{
    my $self = shift;

    if (!$self->{'FILE'}) {
        $self->{'ERROR'}     = -2;
        $self->{'ERROR_MSG'} = "No Pid-File defined";
        return 0;
    }

    use Fcntl qw(:DEFAULT :flock);

    my $timebeg = time();
    while (1) {
        if (sysopen(FH, $self->{'PIDFILE'}, O_RDWR | O_CREAT | O_EXCL)) {
            print FH $self->{'PID'};
            close(FH);
            $self->{'ERROR'}     = 0;
            $self->{'ERROR_MSG'} = "";
            log_debug_print "created PID File ", $self->{'PIDFILE'};
            return 1;
        }
        else {
            if (sysopen(FE, $self->{'PIDFILE'}, O_RDONLY)) {
                log_debug_print "Can open PID File, pid:$$";
                my $pid = <FE>;
                close(FE);
                if (-f "/proc/$pid/cmdline") {
                    my $fhcmd;
                    if (open($fhcmd, '<',"/proc/$pid/cmdline")) {
                        my $pcmdline = <$fhcmd>;
                        close($fhcmd);
                        my $interpreter;
                        my $pname;
                        # parse /proc/PID/cmdline of perl scripts to be able to handle
                        # 'perl <options> script <options>' forms
                        if ($pcmdline =~ /^([^\0]*).*?\0((--\0(-[^\0]+))|([^-\0][^\0]*))/) {
                            $interpreter = $1;
                            (defined $4) ? $pname = $4 : $pname = $5;
                        }
                        my $procname = $self->{'PROCNAME'};
                        log_debug_print "interpreter: '$interpreter', scriptfile: '$pname', procname: '$procname'";

                        if ($pname =~ /^(.*\/|)$procname(|\.[a-zA-Z0-9]+)$/) {
                            ## Programm is runnig
                            ## print "Program is running\n";
                            my $timenow = time();
                            if (($timenow - $timebeg) > $self->{'TIMEOUT'}) {
                                $self->{'ERROR'}     = -1;
                                $self->{'ERROR_MSG'} = "Global Timeout, $procname is already running, pid:$pid";
                                log_debug_print $self->{'ERROR_MSG'};
                                return 0;
                            }
                            sleep($self->{'TIME'});
                            next;
                        }
                        else {
                            ## programm is not running;
                            log_debug_print "Process is not running, pid:$$";
                            unlink($self->{'PIDFILE'});
                            if (unlink($self->{'PIDFILE'})) {
                                next;
                            }
                        }
                    }
                }
                else {
                    ## programm is not running;
                    log_debug_print "Programm is not running, pid:$$";
                    if (unlink($self->{'PIDFILE'})) {
                        next;
                    }
                }
            }
        }
        $self->{'ERROR'}     = -3;
        $self->{'ERROR_MSG'} = "Unknown Error";
        return 0;
    }
} ## end sub create_file

## @method $ get_pid()
# Returns PID number
# @return PID number
sub get_pid
{    #get pid of registered process
    my $self = shift;
    return $self->{'PID'};
} ## end sub get_pid ($)

## @method $ get_pidf()
# Returns PID number that is read from the PID file
# @return PID number or -1 on error
sub get_pidf
{    #get pid from pidfile
    my $self = shift;
    my $pid  = undef;
    if (sysopen(FE, $self->{'PIDFILE'}, O_RDONLY)) {
        #print "Can open File, $$\n";
        $pid = <FE>;
        close(FE);
    }
    if (!$pid) {
        return -1;    #should return a numeric value
    }
    return $pid;
} ## end sub get_pidf ($)


## @method void delete()
# Delete the PID file
sub delete
{
    my $self = shift;

    unlink($self->{'PIDFILE'});
} ## end sub delete ($)

## @method void touch()
# open the PID file and write the PID again
sub touch
{
    my $self = shift;
    $self->{'PID'} = $$;
    if (sysopen(FH, $self->{'PIDFILE'}, O_RDWR | O_CREAT | O_EXCL)) {
        print FH $self->{'PID'};
        close(FH);
    }
    else {
        log_error_print "Update PID file " . $self->{'PIDFILE'} . " failed, pid $$";
    }

} ## end sub touch ($)

## @method void stat_pid()
# Fetch the pid file time information
sub stat_pid
{
    my $self = shift;
    my @a    = stat($self->{'PIDFILE'});
    if (@a) {
        #print join("\t",@a), "\n";
        $self->{'CTIME'} = $a[10];
        $self->{'MTIME'} = $a[9];
        $self->{'ATIME'} = $a[8];
    }
} ## end sub stat_pid ($)

## @method void kill_pid()
# Kill the process who's PID is read from the PID file
# @return undef or a hash containing 'PID' => pid of the killed process, 'KILL' => output of kill command
sub kill_pid
{
    my $self = shift;
    my $pid  = undef;
    my %ret;

    $pid = get_pidf($self);
    if (!defined $pid) {
        return;
    }
    $ret{'PID'} = $pid;
    if (($self->{'BTIME'} - $self->{'TOUCH_TIMEOUT'}) > $self->{'MTIME'}) {
        my $kill = `kill -9 $pid`;
        $ret{'KILL'} = $kill;
        return %ret;
    }

    return %ret;
} ## end sub kill_pid ($)


