#!/usr/bin/perl -w
## @file
# Implementation of image file functions

## @class
package POS::SyncImages;

use strict;

use POS::Image;
use POS::Syslog;
use POS::LdapUtil;
use POS::posDataStatic;
use POS::posDataCache;

# version of the package
our $VERSION = 0.1;


## @cmethod object new()
# @return A new object
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ## config data
        'CACHE'  => undef,    ## cached ldap data
    };

    %$self = (%$self, @_);

    return if !defined $self->{'STATIC'};
    return if !defined $self->{'CACHE'};

    return bless($self);
} ## end sub new

sub sync_command
{
    my @command = @_;
    log_info_print("Running ", join(' ', @command));
    my $val = system(@command);
    if ($val != 0) {
        log_error_print("Error Excuting ", join(' ', @command));
        return 0;
    }
    return 1;
} ## end sub sync_command

sub adminserver
{
    my ($self) = @_;

    my $adminserver = $self->{'STATIC'}{'CONFIG'}{'ALL_LDAPHOST'};
    $adminserver =~ s/.*:\/\///;
    return $adminserver;
} ## end sub adminserver

sub remote_sync_boot
{
    my ($self) = @_;

    my $src     = 'rsync://' . $self->adminserver() . "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_BOOT_BASE'};
    my $dest    = $self->{'STATIC'}{'SYNC_BOOT_DIR'};
    my @command = ('rsync', '-axzv', $src, $dest, '--delete', '--exclude', '/pxelinux.cfg');

    return sync_command(@command);
} ## end sub remote_sync_boot

sub remote_sync_images
{
    my ($self) = @_;

    my $src     = 'rsync://' . $self->adminserver() . "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_IMAGE_BASE'};
    my $dest    = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    my @command = ('rsync', '-axzv', $src, $dest, '--delete');

    return sync_command(@command);
} ## end sub remote_sync_images

sub remote_sync_md5s
{
    my ($self) = @_;

    my $src     = 'rsync://' . $self->adminserver() . "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_IMAGE_BASE'};
    my $dest    = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    my @command = ('rsync', '-axzv', '--include=*.md5', '--exclude=*', $src, $dest, '--delete');

    return sync_command(@command);
} ## end sub remote_sync_md5s

sub remote_sync_image
{
    my ($self, $image) = @_;

    my $name = $image->{'FILENAMEVER'};
    $name .= ".delta" if $image->{'DELTA'};
    my $src     = 'rsync://' . $self->adminserver() . "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_IMAGE_BASE'};
    my $dest    = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    my @command = (
        'rsync',               '-axzv',       "--include=$name", "--include=$name.gz",
        "--include=$name.md5", '--exclude=*', $src,              $dest,
        '--delete'
    );
    return sync_command(@command);
} ## end sub remote_sync_image

sub local_sync_boot
{
    my ($self) = @_;
    my $src    = $self->{'STATIC'}{'SYNC_BOOT_DIR'} . '/';
    my $dest   = $self->{'STATIC'}{'TFTP_BOOT_DIR'} . '/';
    # exclude everything under pxelinux.cfg, except pxelinux.cfg/default, bnc#599195
    my @command = ('rsync', '-axzv', $src, $dest, '--delete', '--exclude', '/pxelinux.cfg');
    return sync_command(@command);
} ## end sub local_sync_boot

sub local_sync_image
{
    my ($self, $image) = @_;
    my $installed = $image->install($self->{'STATIC'}{'TFTP_IMAGE_DIR'});
    return $installed;
} ## end sub local_sync_image

sub image_in_use
{
    my ($self, $image) = @_;

    foreach my $ws (keys %{ $self->{'CACHE'}{'WS'}{'ACT'} }) {
        # return true for current image and also for several previous images
        my @values = $self->{'CACHE'}{'WS'}{'ACT'}{$ws}->get_value('scNotifiedImage');
        foreach my $value (@values) {
            $value =~ s/;.*//;
            $value =~ s/.*\///;
            if ($value eq $image->{'FILENAMEVER'}) {
                log_info_print "found image in use: " . $value;
                return 1;
            }
        }
    }
    return 0;
} ## end sub image_in_use

sub image_in_scSynchronizedImagesDn
{
    my ($self, $image) = @_;
    # no list -> sync all images
    my $scLocation = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($self->{'STATIC'}{'BRANCHBASE'});
    if (!defined $scLocation ||
        !$scLocation->exists('scSynchronizedImagesDn')) {
        return 1;
    }

    foreach my $dn ($scLocation->get_value('scSynchronizedImagesDn')) {
        my $entry     = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($dn);
        my $ver_dn    = "scVersion=" . $self->{'VERSION'} . ",$dn";
        my $ver_entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($ver_dn);
        next unless defined $entry;
        if (defined $entry->get_value('scPosImageDn')) {
            my $image_dn = $entry->get_value('scPosImageDn');
            $entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($image_dn);
            next unless defined $entry;
        }


        if (ldap_entry_has_value('ENTRY' => $entry, 'ATTR_NAME' => 'scImageFile', 'VALUE' => $image->{'FILENAME'}) &&
            (   ldap_entry_has_value(
                    'ENTRY'     => $entry,
                    'ATTR_NAME' => 'scPosImageVersion',
                    'VALUE'     => $image->{'VERSION'} . ";active"
                ) ||
                (defined $ver_entry &&
                    !ldap_entry_has_value('ENTRY' => $ver_entry, 'ATTR_NAME' => 'scDisabled', 'VALUE' => "TRUE"))
            )
          ) {

            log_info_print "found listed in scSynchronizedImagesDn: " . $image->{'FILENAMEVER'};
            return 1;
        }
    }
    return 0;
} ## end sub image_in_scSynchronizedImagesDn

sub image_is_default
{
    my ($self, $image) = @_;

    my $default_cr;
    foreach my $cr (
        @{  $self->{'CACHE'}{'LDAPCACHE'}
              ->getChildrenByClass("cn=global," . $self->{'STATIC'}{'LDAPBASE'}, 'scCashRegister')
        }
      ) {
        if (ldap_entry_has_value(
                'ENTRY'     => $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($cr),
                'ATTR_NAME' => 'scCashRegisterName',
                'VALUE'     => 'default'
            )
          ) {
            $default_cr = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($cr);
            last;
        }
    }

    if (!defined $default_cr) {
        log_error_print("No default cash register found");
        return 0;
    }

    my $default_image_dn = $default_cr->get_value('scPosImageDn');

    if (!defined $default_image_dn || !defined $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($default_image_dn)) {
        log_error_print("Default image $default_image_dn does not exist");
        return 0;
    }

    my $default_image = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($default_image_dn);

    if (ldap_entry_has_value('ENTRY' => $default_image, 'ATTR_NAME' => 'scImageFile', 'VALUE' => $image->{'FILENAME'})
        &&
        ldap_entry_has_value(
            'ENTRY'     => $default_image,
            'ATTR_NAME' => 'scPosImageVersion',
            'VALUE'     => $image->{'VERSION'} . ";active"
        )
      ) {

        log_info_print "found default: " . $image->{'FILENAMEVER'};
        return 1;
    }
    return 0;
} ## end sub image_is_default

sub image_needed
{
    my ($self, $image) = @_;

    return 1 if $self->image_in_use($image);
    return 1 if $self->image_is_default($image);
    return 1 if $self->image_in_scSynchronizedImagesDn($image);
    return 0;
} ## end sub image_needed

sub sync_directory_cleanup
{
    my ($self) = @_;
    my $dir = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    opendir(DIR, $dir) || die "can't opendir $dir: $!";
    my @files = grep { $_ !~ /\.md5$/ && -f "$dir/$_" } readdir(DIR);
    closedir DIR;

    foreach my $file (@files) {
        my $image = POS::Image->new();
        $image->set("$dir/$file");
        if (!$image->valid() && defined $image->{'PATH'}) {
            # the file matches the image naming conventions but
            # the corresponding md5 no longer exist (on AS) -> delete
            log_info_print "deleting " . $image->{'PATH'};
            $image->remove();
        }
    }
} ## end sub sync_directory_cleanup

sub local_directory_cleanup
{
    my ($self)     = @_;
    my $dir        = $self->{'STATIC'}{'TFTP_IMAGE_DIR'};
    my $remote_dir = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    opendir(DIR, $dir) || die "can't opendir $dir: $!";
    # images that do not exist on AS
    my @files = grep { $_ !~ /\.md5$/ && -f "$dir/$_" } readdir(DIR);
    closedir DIR;

    foreach my $file (@files) {
        my $image = POS::Image->new();
        $image->set("$dir/$file");
        # unused images that do not exist on AS or invalid images
        if (!$image->valid() ||
            !-f "$remote_dir/$file" && !$self->image_in_use($image)) {
            log_info_print "deleting " . $image->{'PATH'};
            $image->remove();
        }
    }
} ## end sub local_directory_cleanup


sub get_dl_md5_list
{
    my ($self) = @_;
    my $dir = $self->{'STATIC'}{'SYNC_IMAGE_DIR'};
    opendir(DIR, $dir) || die "can't opendir $dir: $!";
    my @files = grep { /\.md5$/ && -f "$dir/$_" } readdir(DIR);
    closedir DIR;

    my @images;
    foreach my $file (@files) {
        my $image = POS::Image->new();
        if ($image->set_md5_file("$dir/$file")) {
            push @images, $image;
            log_info_print "found md5 for " . $image->{'FILENAMEVER'};
        }
    }
    return @images;
} ## end sub get_dl_md5_list

sub get_dl_image_list
{
    my ($self) = @_;

    my @images = $self->get_dl_md5_list();
    my @deltas;
    my %h_images;
    foreach my $image (@images) {
        next if !$self->image_needed($image);

        if ($image->{'DELTA'}) {
            push @deltas, $image;
        }
        else {
            $h_images{ $image->{'FILENAMEVER'} } = $image;
        }
    }

    # use delta where possible
    foreach my $delta (@deltas) {
        if (defined $h_images{ $delta->{'MD5'}{'BASE'} } &&
            !$h_images{ $delta->{'MD5'}{'BASE'} }{'DELTA'}) {
            $h_images{ $delta->{'FILENAMEVER'} } = $delta;

            log_info_print "using delta for  " . $delta->{'FILENAMEVER'};
        }
    }

    my @base_first = sort { $a->{'DELTA'} <=> $b->{'DELTA'} } values(%h_images);
    return @base_first;
} ## end sub get_dl_image_list

sub sync
{
    my ($self) = @_;
    if ($self->{'REMOTE'} && !$self->remote_sync_boot()) {
        log_error_print "Can't sync boot images from admin server";
        return 0;
    }
    if ($self->{'LOCAL'} && !$self->local_sync_boot()) {
        log_error_print "Can't install boot images";
        return 0;
    }

    if ($self->{'REMOTE'} && !$self->remote_sync_md5s()) {
        log_error_print "Can't sync md5 files from admin server";
        return 0;
    }


    my @deltas;
    foreach my $image ($self->get_dl_image_list()) {
        if ($self->{'REMOTE'} && !$self->remote_sync_image($image)) {
            log_error_print "Can't sync '" . $image->{'NAMEVER'} . "' from admin server";
            return 0;
        }
        $image->reread();    #use up-to-date info
        if ($self->{'LOCAL'} && !$self->local_sync_image($image)) {
            log_error_print "Can't locally install '" . $image->{'NAMEVER'};
            return 0;
        }

        if ($image->{'DELTA'}) {
            my $full = $image->apply_delta();
            if (!$full) {
                log_error_print "Can't apply delta '" . $image->{'NAMEVER'};
                return 0;
            }

            if ($self->{'LOCAL'} && !$self->local_sync_image($full)) {
                log_error_print "Can't locally install '" . $image->{'NAMEVER'};
                return 0;
            }
        }
    }
    $self->sync_directory_cleanup()  if ($self->{'REMOTE'});
    $self->local_directory_cleanup() if ($self->{'LOCAL'});
    return 1;
} ## end sub sync


1;
