#!/usr/bin/perl -w
package POS::Syslog;

use strict;
use Unix::Syslog qw( :macros :subs);
use Getopt::Long;
use POSIX qw(strftime);
use POSIX qw(locale_h);

require Exporter;
our @ISA = qw(Exporter);

our $VERSION = 0.1;

our @EXPORT = qw(log_debug_printf  log_debug_print
  log_info_printf   log_info_print
  log_notice_printf log_notice_print
  log_warn_printf   log_warn_print
  log_error_printf  log_error_print
  log_crit_printf   log_crit_print
  log_alert_printf  log_alert_print
  log_emerg_printf  log_emerg_print
  log_parse_argv    log_print_usage);
#
# Example
#
#  use POS::Syslog;
#  use Unix::Syslog(:macros);      ## for Syslog priority constants only
#
#  $log = POS::Syslog->new(FORMAT => "Missing value %s in object %s",
#				    ARGS   => ["image","scPosImage"]);
#  $log = POS::Syslog->new(PRIORITY =>LOG_INFO,
#				    FORMAT => "Missing value %s in object %s",
#				    ARGS   => ["image","scPosImage"]);
#

my $log_priority_stderr    = LOG_WARNING;
my $log_priority_syslog    = LOG_INFO;
my $log_priority_threshold = $log_priority_stderr < $log_priority_syslog ? $log_priority_syslog : $log_priority_stderr;


# constructor

sub new
{
    my $package = shift;
    # setting default Values
    my $data = { 'IDENT' => $0, 'OPTION' => LOG_PID, 'FACILITY' => LOG_USER, 'PRIORITY' => LOG_ERR, };

    my @prio_str = ('EMERG', 'ALERT', 'CRIT', 'ERROR', 'WARN', 'NOTICE', 'INFO', 'DEBUG');

    # enumerate possible options for little optimized behavior
    $data->{'PRIORITY'} = $_[1] || '';
    $data->{'FORMAT'}   = $_[3] || '';
    $data->{'ARGS'}     = $_[5] || [];
    $data->{'CALLER'}   = $_[7] || [caller(0)];

    $data->{'PRIORITY'} = LOG_EMERG if $data->{'PRIORITY'} < LOG_EMERG;
    $data->{'PRIORITY'} = LOG_DEBUG if $data->{'PRIORITY'} > LOG_DEBUG;

    my $file = $data->{'CALLER'}[1];
    my $line = $data->{'CALLER'}[2];

    $file =~ s/.*\///;

    my $fileline = " $file:$line:";
    # print files and lines to stderr only at verbosity>=debug
    my $fileline_stderr = ($log_priority_stderr >= LOG_DEBUG) ? $fileline : '';
    my $date = '';

    # print syslog timestamp (RFC 3164) when STDERR is redirected from terminal
    if (! -t STDERR) {
        my $old_locale = setlocale(LC_TIME);
        setlocale(LC_TIME, 'C') if ($old_locale ne 'C');
        $date = strftime('%b %e %H:%M:%S ', localtime);
        setlocale(LC_TIME, $old_locale) if ($old_locale ne 'C');
    }

    if ($data->{'PRIORITY'} <= $log_priority_stderr) {
        printf(STDERR "%s[%s]%s $data->{'FORMAT'}\n",
            $date, $prio_str[$data->{'PRIORITY'}],
            $fileline_stderr, @{ $data->{'ARGS'} }
        );
    }
    if ($data->{'PRIORITY'} <= $log_priority_syslog) {
        openlog($data->{'IDENT'}, $data->{'OPTION'}, $data->{'FACILITY'});
        syslog(
            $data->{'PRIORITY'},
            "[%s]%s $data->{'FORMAT'}",
            $prio_str[$data->{'PRIORITY'}],
            $fileline, @{ $data->{'ARGS'} }
        );
        closelog();
    }
    return;
} ## end sub new


sub log_debug_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_DEBUG, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_DEBUG);
} ## end sub log_debug_printf ($;@)

sub log_debug_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_DEBUG, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_DEBUG);
} ## end sub log_debug_print (@)

sub log_info_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_INFO, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_INFO);
} ## end sub log_info_printf ($;@)

sub log_info_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_INFO, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_INFO);
} ## end sub log_info_print (@)

sub log_notice_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_NOTICE, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_NOTICE);
} ## end sub log_notice_printf ($;@)

sub log_notice_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_NOTICE, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_NOTICE);
} ## end sub log_notice_print (@)

sub log_warn_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_WARNING, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_WARNING);
} ## end sub log_warn_printf ($;@)

sub log_warn_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_WARNING, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_WARNING);
} ## end sub log_warn_print (@)

sub log_error_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_ERR, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_ERR);
} ## end sub log_error_printf ($;@)


sub log_error_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_ERR, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_ERR);
} ## end sub log_error_print (@)

sub log_crit_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_CRIT, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_CRIT);
} ## end sub log_crit_printf ($;@)


sub log_crit_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_CRIT, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_CRIT);
} ## end sub log_crit_print (@)

sub log_alert_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_ALERT, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_ALERT);
} ## end sub log_alert_printf ($;@)


sub log_alert_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_ALERT, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_ALERT);
} ## end sub log_alert_print (@)

sub log_emerg_printf ($;@)
{
    my $format = shift;
    POS::Syslog->new(PRIORITY => LOG_EMERG, FORMAT => $format, ARGS => \@_, CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_EMERG);
} ## end sub log_emerg_printf ($;@)


sub log_emerg_print (@)
{
    POS::Syslog->new(PRIORITY => LOG_EMERG, FORMAT => "%s", ARGS => ["@_"], CALLER => [caller(0)])
      if ($log_priority_threshold >= LOG_EMERG);
} ## end sub log_emerg_print (@)

sub log_parse_prio ($)
{
    my ($prio) = @_;

    return $prio       if $prio =~ /^[0-9]+$/;
    return LOG_EMERG   if $prio =~ /emerg/i;
    return LOG_ALERT   if $prio =~ /alert/i;
    return LOG_CRIT    if $prio =~ /crit/i;
    return LOG_ERR     if $prio =~ /err/i;
    return LOG_WARNING if $prio =~ /warn/i;
    return LOG_NOTICE  if $prio =~ /notice/i;
    return LOG_INFO    if $prio =~ /info/i;
    return LOG_DEBUG   if $prio =~ /debug/i;
    return LOG_ERR;    #default
} ## end sub log_parse_prio ($)

sub verbose
{
    my $vval = $_[1];
    if ($vval) {
        $log_priority_stderr = log_parse_prio($vval);
    }
    else {
        $log_priority_stderr += 1;
    }
} ## end sub verbose


sub log_parse_argv ()
{
    # first parse environment
    $log_priority_stderr = log_parse_prio($ENV{'POS_FORCE_VERBOSITY'}) if ($ENV{'POS_FORCE_VERBOSITY'});
    # then command line so it has preference
    my $p = new Getopt::Long::Parser;
    $p->configure('pass_through');
    my $ret = $p->getoptions('verbose|v:s' => \&verbose, 'syslog:s' => \$log_priority_syslog);

    $log_priority_syslog = log_parse_prio($log_priority_syslog);

    $log_priority_threshold = $log_priority_stderr < $log_priority_syslog ? $log_priority_syslog : $log_priority_stderr;

    log_info_print "log priority set to $log_priority_stderr $log_priority_syslog";
} ## end sub log_parse_argv ()

sub log_print_usage ()
{
    print STDERR "  -v level, --verbose=level   set the stderr verbosity level\n";
    print STDERR "  --syslog=level              set the syslog verbosity level\n";
    print STDERR "                              verbosity level can be specified by a number\n";
    print STDERR "                              or as one of:\n";
    print STDERR "                              [emerg alert crit err warn notice info debug]\n";
} ## end sub log_print_usage ()

1;

