#!/usr/bin/perl -w
## @file
# Implementation of POS::posDhcp


## @class
# Generates the DNS configuration and zonedata from LDAP.
# @par
# @note @em non-class perl module
package POS::posDhcp;

use strict;

# version  of the package
our $VERSION = 0.1;

use Exporter;

our @ISA = qw( Exporter );

# Exported subroutines, strings, arrays, hashes
our @EXPORT = qw(getchecktftpserver);

use POS::LDAP;
use Net::IP;
use Net::IPv4Addr qw( :all);
use Socket;
use File::Path;
use File::Copy;
use File::Compare;
use IO::File;
use POSIX qw(tmpnam floor);
use Time::localtime;
use URI::Escape;
use POS::Config;
use POS::Syslog;
use POS::Pid;
use POS::IP qw(:DEFAULT is_in_range);
use POS::LdapUtil;
use POS::DhcpEntries;
use strict;

## @cmethod object new()
# @return A new object
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ## config data
        'CACHE'  => undef,    ## cached ldap data
    };

    %$self = (%$self, @_);

    return if !defined $self->{'STATIC'};
    return if !defined $self->{'CACHE'};

    return bless($self);
} ## end sub new


sub gensubnets()
{
    # generate a DHCP subnet declaration for each given base
    my ($self, $tmp_fh) = @_;
    my $Not_Authoritative = $self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_NOTAUTHORITATIVE'};
    my $No_Dyn_Range      = $self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_NODYNRANGE'};
    my $entry             = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($self->{'STATIC'}{'BRANCHBASE'});
    # in LDAP the range is given as a string consisting of two IP-adresses
    # separated by comma.
    my $tmprange = $entry->get_value('scDhcpRange');
    # remove blanks from range string
    $tmprange =~ s/\s+//g;
    my @range   = split(/,/, $tmprange);
    my $network = $entry->get_value('ipNetworkNumber');
    my $netmask = $entry->get_value('ipNetmaskNumber');
    # write the dn of the current entry as a comment (useful for finding errors in LDAP entries)
    print $tmp_fh "  # dn: ", $entry->dn, "\n";
    print $tmp_fh "  subnet $network netmask $netmask {\n";
    # if $Not_Authoritative is set, write the "not authoritative" flag
    # into the config file to avoid sending DHCPNAKs.
    if ($Not_Authoritative) {
        print $tmp_fh "  not authoritative;\n";
    }
    # the broadcast address can be calculated from network address and netmask
    print $tmp_fh "    option broadcast-address ", Net::IPv4Addr::ipv4_broadcast($network, $netmask), ";\n";
    print $tmp_fh "    option routers ", $entry->get_value('scDefaultGw'), ";\n";
    print $tmp_fh "    option domain-name \"" . get_searchlist($entry) . "\";\n";
    #### try to get default boot_option
    my $default_boot_option = get_standard_bootfile($self, $entry->dn);
    if ($default_boot_option) {
        #my $default_boot_option = get_boot_options($self,$default_imagedn);
        #if ($default_boot_option) {
        print $tmp_fh "    filename \"$default_boot_option\";\n";
    }
    else {
        log_warn_printf('Can\'t create default boot-option: %s', $entry->dn);
    }

    # generate netboot entries (for the whole subnet)
    #####????genbootopts($entry->dn);???'

    # a range is only generated if scDynamicIp is set to "TRUE"
    # but if $No_Dyn_Range is true, never write a range.
    if ($entry->get_value('scDynamicIp') eq 'TRUE' &&
        !$No_Dyn_Range) {
        print $tmp_fh "    range ", $range[0], " ", $range[1], ";\n";
    }
    my $leasetime = $self->{'CACHE'}{'DHCP_DYNLEASETIME'};
    if ($leasetime and $leasetime > 0) {
        print $tmp_fh "    default-lease-time " . $leasetime . ";\n";
        print $tmp_fh "    max-lease-time " . $leasetime . ";\n";
    }

    my $next_ip = getnextserver($self, $entry->dn);
    print $tmp_fh "    next-server " . $next_ip . ";\n";

    # generate fixed-address entries
    # may add a tftp next-server or put this information in the branch

    if (!genfixed($self, $entry->dn, $network, getnextserver($self, $entry->dn), $tmp_fh)) {
        log_info_printf('no scWorkstation object found under %s, no hosts written into %s',
            $entry->dn, $self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_DHCPCONFFILE'});
    }
    print $tmp_fh "  }\n\n";
} ## end sub gensubnets()

sub _dnsname2dhcphost
{
   my ($self, $dnsname, $mac) = @_;
   $mac =~ s/://g;
   $dnsname  =~ s/^(.*?)(\..*)/$1-$mac$2/;    #make dhcp host section name unique
   return $dnsname;
}

sub _generate_dhcp_block
{
   my ($self, $ws_obj, $mac) = @_;
   my $hostblock;
   my $fixedipaddr  = $ws_obj->get_value('ipHostNumber');
   my $dhcp_dns= ldap_dn2dnsname($ws_obj->dn);
   my $dhcp_host_id = $self->_dnsname2dhcphost($dhcp_dns,$mac);
   $hostblock = "      host ".$dhcp_host_id."{\n";
   $hostblock.= "        hardware ethernet ".$mac.";\n";
   $hostblock.= "        fixed-address ".$fixedipaddr.";\n";
#            option host-name is redundant with DNS and might cause problems
#            print $tmp_fh "        option host-name ",  $dhcp_host,    ";\n";
   $hostblock.= $self->writebootopts($ws_obj);
   $hostblock.= "      }\n";


   return $hostblock;
}

sub genfixed()
{
    # generate fixed-address entries for a given base.
    # for calculating the clients IP, which is coded in his cn entry,
    # we need the fixed range for his subnet.
    # if no boot schema (scDoTftpBoot) is set for a workstation,
    # the default lease time is set according to the scDoTftpBoot entry
    # from the location.

    # Parameters:
    # - base: where to search scWorkstation entries
    # - network_addr: network address of Location for IP calculation
    # - value of scDoTftpBoot for this location
    my ($self, $base, $network_addr, $boot_opt, $tmp_fh) = @_;

    my $wshash_r = $self->{'CACHE'}{'WS'}{'ACT'};

    #if results contains no entries, do nothing
    #if (!$wshash_r) {
    #    return 0;
    #}

    print $tmp_fh "    group {\n";

    # set different lease time for fixed-address entries
    my $leasetime = $self->{'CACHE'}{'DHCP_FIXLEASETIME'};
    if ($leasetime and $leasetime > 0) {
        print $tmp_fh "      default-lease-time " . $leasetime . ";\n";
        print $tmp_fh "      max-lease-time " . $leasetime . ";\n";
    }
    print $tmp_fh "$self->{'STATIC'}{'CONFIG'}{'POS_DHCP_GROUP_START_ID'}\n";

##TODO: specifc procedure ws-add
    foreach my $cn (keys %{$wshash_r}) {
        my $entry = $wshash_r->{$cn};
        my @macs  = $entry->get_value('macAddress');
        foreach my $mac (@macs) {
            #print $tmp_fh "$self->{'STATIC'}{'CONFIG'}{'POS_DHCP_WS_START_ID'}$mac\n";
            my $hostblock = $self->_generate_dhcp_block($entry,$mac);
            print $tmp_fh $hostblock;
        }
    }

    # close group
    print $tmp_fh "$self->{'STATIC'}{'CONFIG'}{'POS_DHCP_GROUP_END_ID'}\n";
    print $tmp_fh "    }\n\n";

} ## end sub genfixed()

sub dhcp_write_header($$)
{
    # read the given header file and write the content to the new config file
    my ($self, $headerfile, $tmp_fh) = @_;

    # read all lines from the header file
    if (!open(header_fh, '<', $headerfile)) {
        log_alert_printf('can\'t open %s', $headerfile);
        die "can't open $headerfile";
    }
    my @headerlines = <header_fh>;
    close(header_fh);

    foreach my $line (@headerlines) {
        print $tmp_fh $line;
    }
    print $tmp_fh "\n# end of header\n\n";
} ## end sub dhcp_write_header($$)


sub writebootopts()
{
    ##
    ##
    my ($self, $entry, $tmp_fh) = @_;
    my $bootopts;
    my $nextserver=$self->getnextserver($entry->dn);

    if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => 'scWorkstation') &&
        $entry->get_value('macAddress')) {
        my $scPosImageDn = $entry->get_value('scPosImageDn');
        my $scRefPcDn    = $entry->get_value('scRefPcDn');
        if (!$scPosImageDn) {
            if ($scRefPcDn) {
                $scPosImageDn = get_default_image_dn($self, $scRefPcDn);
            }
            else {
                log_warn_printf('Entry %s: missing ImageDn and scRefPcDn', $entry->dn);
            }
        }
        if ($scPosImageDn) {
            if (defined($self->{'DHCPCACHE'}{'BOOTOPTIONS'}{$scPosImageDn})) {
                $bootopts.="        next-server ".$nextserver.";\n";
                $bootopts.="        filename \"".$self->{'DHCPCACHE'}{'BOOTOPTIONS'}{$scPosImageDn}."\";\n";
            }
            else {
                my $boot_option = $self->get_boot_options($scPosImageDn);
                if ($boot_option) {
                    $bootopts.="       next-server ".$nextserver.";\n";
                    $self->{'DHCPCACHE'}{'BOOTOPTIONS'}{$scPosImageDn} = $boot_option;
                    $bootopts.="       filename \"".$self->{'DHCPCACHE'}{'BOOTOPTIONS'}{$scPosImageDn}."\";\n";
                }
                else {
                    log_warn_printf('Entry %s: missing bootoptions', $entry->dn);
                }
            }
        }
        else {
            log_warn_printf('Entry %s: missing ImageDn', $entry->dn);
        }
    }
    return $bootopts;
} ## end sub writebootopts()


## new general subroutine to find TFTP server
## warning!! this method is called from bash, which parses stdout
## => must only print $tftp_ip if all ok, or nothing if error
sub getchecktftpserver
{
    #
    # return ip of tftp-server
    # todo: check if tft-server belongs to the branch
    my ($config, $Ldap_Conn, $base) = @_;

    # try to generate new if not supplied
    $config = POS::Config->new() if (!defined $config);
    $Ldap_Conn = POS::LDAP->new(
        'CONFIG'     => $config,
        'NEEDCAP'    => 'ro',
        'LDAPBINDDN' => $config->{'POS_ADMIN_DN'},
        'LDAPBINDPW' => $config->{'POS_ADMIN_PASSWORD'}
    ) if (!defined $Ldap_Conn);
    $base = $config->{'BRANCH_LDAPBASE'} if (!defined $base);

    my $tftp_name = $config->{'TFTP_SERVER'};
    my $tftp_data;


    my $result = $Ldap_Conn->search(
        base   => $base,
        scope  => 'sub',
        filter => '(&(objectClass=scService)(scServiceName=tftp))',
        attrs  => ['ipHostNumber']
    );
    if ($result->count == 1) {
        my $entry   = $result->pop_entry();
        my $tftp_ip = $entry->get_value('ipHostNumber');
        if (!defined($tftp_ip)) {
            log_error_printf("No ipHostNumber of the TFTP-Server under $base defined");
            return (undef);
        }
        #flush dns cache
        system('rcnscd restart > /dev/null');
        #hostname resolv check here, cmp to tftp ip received above
        my @hostret = gethostbyname($tftp_name);
        if (@hostret) {
            splice(@hostret, 0, 4);
            if (@hostret > 1) {
                log_notice_printf(
                    "More than one ip addresses of the TFTP-Server '$tftp_name' resolves via dns, may cause problems.");
            }
            my $tftp_ip_ok = 0;
            my @rhostret;
            foreach my $chret_ip (@hostret) {
                my $hret_ip = inet_ntoa($chret_ip);
                push @rhostret, $hret_ip;
                if ($hret_ip eq $tftp_ip) {
                    $tftp_ip_ok = 1;
                    last;
                }
            }
            my $rhostretstr = join(', ', @rhostret);
            if (!$tftp_ip_ok) {
                log_error_printf(
"ipHostNumber '$tftp_ip' of the LDAP TFTP-Server under $base differs from tftp ip(s) '$rhostretstr' from dns"
                );
                return (undef);
            }
            return $tftp_ip;
        }
        else {
            log_error_printf("Cannot resolve '$tftp_name'");
            return (undef);
        }
    }
    elsif ($result->count > 1) {
        log_error_printf("More than one TFTP-Server defined under $base");
        return (undef);
    }
    else {
        log_error_printf("No TFTP-Server defined under $base");
        return (undef);
    }
} ## end sub getchecktftpserver

sub getnextserver($$)
{
    #
    # return ip of tftp-server
    # todo: check if tft-server belongs to the branch
    my ($self, $base) = @_;
    my $tftp_server = $self->{'TFTP'};
    return $tftp_server if defined $tftp_server;
    return 'tftp';
} ## end sub getnextserver($$)

sub get_boot_options($$)
{
    #
    # return scDhcpOptionsRemote
    #
    my ($self, $scPosImageDn) = @_;
    if (!$scPosImageDn) {
        return undef;
    }
    my $result = $self->{'CACHE'}{'LDAP'}->search(
        base   => $scPosImageDn,
        scope  => 'base',
        filter => '(objectClass=scPosImage)',
        attrs  => ['scDhcpOptionsRemote']
    );
    if ($result->count == 1) {
        my $entry  = $result->pop_entry;
        my $option = $entry->get_value('scDhcpOptionsRemote');
        if ($option) {
            return $option;
        }
    }
    log_warn_printf('Can\'t find scDhcpOptionsRemote in  %s', $scPosImageDn);
    return undef;
} ## end sub get_boot_options($$)


sub get_default_image_dn($$)
{
    ## getting default image for a Cashregister
    my ($self, $scRefPcDn) = @_;
    if (!$scRefPcDn) {
        return undef;
    }
    my $result = $self->{'CACHE'}{'LDAP'}->search(
        base   => $scRefPcDn,
        scope  => 'base',
        filter => '(objectClass=scCashRegister)',
        attrs  => ['scPosImageDn']
    );
    if ($result->count == 1) {
        my $entry  = $result->pop_entry;
        my $option = $entry->get_value('scPosImageDn');
        if ($option) {
            return $option;
        }
    }
    log_warn_printf('Can\'t find scPosImageDn in scCashRegister:  %s', $scRefPcDn);
    return undef;
} ## end sub get_default_image_dn($$)

sub get_standard_bootfile($$)
{
    ## return boot file name from scHardware
    my ($self, $base) = @_;

    my @a = split(/,/, $base);

    while (@a) {
        my $result = $self->{'CACHE'}{'LDAP'}->search(
            base   => join(",", @a),
            scope  => 'sub',
            filter => '(objectClass=scHardware)',
            #attrs => ['scPosImageDn'] );
            attrs => ['scDhcpOptionsRemote']
        );

        if ($result->count == 1) {
            my $entry  = $result->pop_entry;
            my $option = $entry->get_value('scDhcpOptionsRemote');
            if ($option) {
                return $option;
            }
        }
        shift @a;
    }
    log_warn_printf('Can\'t find any scHardware-Object in treesearch for  %s', $base);
    return undef;
} ## end sub get_standard_bootfile($$)

sub get_searchlist($)
{
    my $ldap_entry = shift;
    my $zonename   = "";

    if (!($zonename = $ldap_entry->get_value('associatedDomain'))) {
        my $dn = $ldap_entry->dn();
        # generate string with list of domains to search including
        # the domain for a subnet given by dn
        my @search;
        my @aa = split(/,/, $dn);
        foreach my $a (@aa) {
            my ($cn, $name) = split(/=/, $a);
            push(@search, $name);
        }

        $zonename = join(".", @search);
    }
    return $zonename;
} ## end sub get_searchlist($)


sub dhcp_conf_syntax_check
{
    # 20010806 stephan.duehr@suse.de
    # Before activating the new config, do a syntax check by
    # running dhcpd -t on the tmpfile. If the file is ok, it
    # will return exit code 0 and 1 on failure.
    # This function returns 1 (true) if syntax is ok and
    # 0 (false) if syntax check found errors. In case of any
    # errors, a syslog message is generated and the output
    # with detailed information on the errors is written to
    # a temporary file, as this is too much data for writing
    # it to syslog.
    #
    # Parameters:
    # - Filename of the generated dhcpd.conf (temporary filename)
    # - Filename for the debug info output (eg. /var/tmp/dhcpconferr.out)

    my ($self, $tmp_filename) = @_;
    my $debug_filename = "/var/log/pos/dhcpconferr.out";
    my $ret_code       = 0;
    my $dhcpd_exitcode = system("/usr/sbin/dhcpd -t -cf $tmp_filename 2> $debug_filename") >> 8 ;
    if ($dhcpd_exitcode == 0) {
        # syntax is ok, return true
        $ret_code = 1;
    }
    else {
        # errors were found: generate syslog message
        log_alert_print('generated dhcpd.conf has errors, keeping old config');
        my $badcfg = $self->{'CACHE'}{'DHCP'}{'CONFIG'}{'PATHNAME'}.".with-errors";
        copy($tmp_filename,   $badcfg);
        log_alert_print("see $debug_filename for error log");
        log_alert_print("and the invalid dhcp configuration $badcfg ");
        $ret_code = 0;
    }
    return $ret_code;
} ## end sub dhcp_conf_syntax_check



sub dhcp_add_remove_hosts
{
    my ($self, $tmp_fh, $dhcpcfg, $add_hr, $remove_hr) = @_;
    if (!open(dhcp_fh, '<', $dhcpcfg)) {
        log_alert_printf('can\'t open %s', $dhcpcfg);
        die "can't open $dhcpcfg";
    }
    my @dhcplines = <dhcp_fh>;
    close(dhcp_fh);
    
    my $moddhcp;
    my $block;
    my $blockmode=0;
    my $nbracket=0;
    my $mac;    
    foreach my $line (@dhcplines) {
        
        $blockmode = 0 if ( $line =~ $self->{'STATIC'}{'CONFIG'}{'POS_DHCP_GROUP_END_ID'}   );
        $moddhcp.=$line unless $blockmode;
        if ( $line =~ $self->{'STATIC'}{'CONFIG'}{'POS_DHCP_GROUP_START_ID'} ) {
           #write addon lines
           foreach my $mac (keys %{$add_hr})  {
              $moddhcp.=$self->_generate_dhcp_block($add_hr->{$mac}{'obj'},$mac);
           }
           $blockmode = 1 ;
           next; #so we do not write tagline again
        }
        if ($blockmode) {
           if ($line=~/\s*host\s+/) {
              $nbracket=0; #should not be normaly needed
              $block="";
           }
           $block.=$line;
           if ($line=~/\s*hardware\s*ethernet\s*(.*?)\s*;/) {
             $mac=uc($1);  
           }
           #get brackets from line (last one must be alone)
           my $close_b=$line;
           $close_b=~s/[^}]//g;
           my $open_b=$line;
           $open_b=~s/[^{]//g;
           $nbracket=$nbracket+length($open_b)-length($close_b);  
           if ($nbracket == 0) {
              $moddhcp.=$block unless defined $remove_hr->{$mac};               
           }     
        }

    }
    print $tmp_fh $moddhcp;
     
    

}




sub _process_ws_hash
{
    my ($self, $ws_in_hr) = @_;
    my %ws_out_h;
    foreach my $mac (keys %{$ws_in_hr}) {
        my $ws_obj=$ws_in_hr->{$mac};
        $ws_out_h{$mac}{'IP'}=$ws_obj->get_value('ipHostNumber');
        $ws_out_h{$mac}{'MAC'}=$mac;
        $ws_out_h{$mac}{'obj'}=$ws_obj;
    }
    return \%ws_out_h;    
}

#TODO: if needed somewhere else too, put it in a common module
sub substract_hash
{
    my ($self, $hr1, $hr2, @cmpvals) = @_;
    my %diff_h;
    foreach my $key (keys %{$hr1}) {
        my $isinboth=0;
        if (defined $hr2->{$key}) {
          $isinboth=1;
          foreach my $val (@cmpvals) {
            if ($hr1->{$key}{$val} ne $hr2->{$key}{$val}) {
                $isinboth=0;
                last;
            }
          }
        }
        unless ($isinboth) {
          foreach my $key2 (keys %{$hr1->{$key}}) {
             $diff_h{$key}{$key2}=$hr1->{$key}{$key2};
          }
        }
    }
    return \%diff_h;
 
}


sub ldap2dhcp
{
    my ($self, $reuse) = @_;
    my $ret = 0;
 

    my $tmp_filename;
    my $tmp_fh;

    # DHCP config file to be written
    my $outfile = $self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_DHCPCONFFILE'};

    # Open temporary file
    do { $tmp_filename = POSIX::tmpnam() } until $tmp_fh = IO::File->new($tmp_filename, O_WRONLY | O_CREAT | O_EXCL);

    if ( $reuse ) { #rewrite all
        my $ws_hr=$self->_process_ws_hash($self->{'CACHE'}{'WS'}{'BY-MAC'});
        log_debug_printf('have %s ws entries in LDAP: %s', int keys %{$ws_hr}, join(', ',keys %{$ws_hr}));  
        my $addws_hr   = $self->substract_hash($ws_hr, $self->{'CACHE'}{'DHCP'}{'CONFIG'}{'HASH'}  ,   'IP');
        log_debug_printf('adding %s host entries to dhcp.conf : %s', int keys %{$addws_hr}, join(', ',keys %{$addws_hr}));   
        my $removews_hr= $self->substract_hash($self->{'CACHE'}{'DHCP'}{'CONFIG'}{'HASH'}, $ws_hr  ,   'IP'); 
        log_debug_printf('removing %s host entries to dhcp.conf : %s', int keys %{$removews_hr}, join(', ',keys %{$removews_hr}));  
        
        $self->dhcp_add_remove_hosts( $tmp_fh, $outfile, $addws_hr, $removews_hr);
    }
    else {
        # write header
        $self->dhcp_write_header($self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_TEMPLATEFILE'}, $tmp_fh);
        # generate subnet declaration for the list of bases
        $self->gensubnets($tmp_fh);
    } 
    
    # cleanup
    close($tmp_fh);


    # compare temporary file with dhcp config file
    if (compare($tmp_filename, $outfile) != 0) {
        # 20010806 stephan.duehr@suse.de
        # before activating the new config, do a syntax check
        if ($self->dhcp_conf_syntax_check($tmp_filename)) {

            # ret_code=1 <=> files differ and syntax is ok: copy tmpfile over outfile
            # but keep the old one with .old extension before
            if (!-e $outfile) {
                log_info_printf('No previous dhcp configuration (%s) found.. skipping backup', $outfile);
            }
            elsif (!copy($outfile, $outfile . '.old')) {
                log_warn_printf('Can\'t backup dhcp config into %s reason %s', $outfile . '.old', $!);
            }
            if (!copy($tmp_filename, $outfile)) {
                log_alert_printf('Can\'t write new %s reason %s', $outfile, $!);
                $ret = 2;
            }
            else {
                $ret = 0;    #success

                system "/etc/init.d/dhcpd restart";
            }
        }    #fi of syntax check
        else {
            $ret = 3;    #messages already generated by dhcp_conf_syntax_check subroutine
        }
    }    #fi of compare temporary file with dhcp config file

    # remove temporary file when program ends
    if (!unlink($tmp_filename)) {
        log_warn_printf('couldn\'t remove temporary dhcp config file %s', $tmp_filename);
    }
    return $ret;
} ## end sub ldap2dhcp

