#!/bin/bash

#
#==========================================================================
# FILE        : migration_common.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2013 SUSE LINUX Products GmbH, Nuernberg, All rights reserved
#               (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Ondrej Holecek <oholecek@suse.com>
#               Petr Ostadal <postadal@suse.com>
#
# DESCRIPTION : Common functions used in migration scripts.
#               Logging facility, Error handling, Copying/Extracting tared
#               file and displaying progress bar
#
#==========================================================================


NLPOSMIGRATIONDIR="NLPOS9-migration"
SLEPOSMIGRATIONDIR="SLEPOS-migration"
SUSERELEASE="etc/SuSE-release"


############# Logging facility #####################

#LOGLEVElS
LOG_OFF=0
LOG_INFO=1
LOG_VERBOSE=2
#default LOGLEVEL
LOGLEVEL=$LOG_INFO
#default LOGFILE
LOGFILENAME="/var/log/slepos-migration.log"
#enabled logging targets
LOGSCREEN=1
LOGFILE=0
PROGRESSBAR=0

function log
{
    MSG=$1
    LEVEL=$2
    [ -z "$MSG" ] && return
    [ -z "$LEVEL" ] && LEVEL=$LOG_INFO
    [ $LEVEL -gt $LOGLEVEL ] && return
    [ "$LOGFILE" = "1" ] && echo -n "`date '+%b %d %H:%M:%S'` "
    echo -e "$MSG"
}

function logChapter
{
    MSG=$1
    log "========== $MSG =========="
}

function logSuccess
{
    MSG=$1
    log "$MSG        [OK]"
}

function logFailure
{
    MSG=$1
    log "$MSG        [FAIL]"
}

function setLogLevel
{
    NEWLEVEL=$1
    if [ -z "$NEWLEVEL" ]; then
        warning "Trying to switch logging level without new level set. Ignoring"
    elif [ "$NEWLEVEL" -gt 2 ]; then
        warning "New logging level out of range. Ignoring"
    else
        LOGLEVEL=$NEWLEVEL
        log "New logging level set to $NEWLEVEL" $LOG_VERBOSE
    fi
}

function SetFileOutput
{
    if [ $1 -eq 0 ]; then
        LOGFILE=0
    elif [ $1 -eq 1 ]; then
        LOGFILE=1
        PROGRESSBAR=1
        exec 3>&1
        exec 1>> "$LOGFILENAME"
        exec 2>> "$LOGFILENAME"
    else
        warning "SetFileOutput called with bad parameter. Ignoring."
    fi
}

################## Error handling ###################

function err
{
    MSG=$1
    if [ -z "$MSG" ]; then
        log "[ERROR] unknown error, check log files for errors."
    else
        log "[ERROR] $MSG"
    fi
    #add some more error handling
    exit 255
}

function warning
{
    MSG=$1
    [ -z "$MSG" ] && return
    log "[WARNING] $MSG"
}

################## Tar progress support #############


## evaluate a floating point number expression
# param: expresion for bc
# return: 0 successful
# output: expression result
function FloatEval
{
    local stat=0
    local result=0.0
    if [ $# -gt 0 ]; then
        result=$(echo "scale=$float_scale; $*" | bc -q 2>/dev/null)
        stat=$?
        if [ $stat -eq 0 ] && [ -z "$result" ]; then
            stat=1
        fi
    fi
    echo $result
    return $stat
}

## evaluate a floating point number conditional expression
# param: expression for bc
# return: 0 - true, 1 - false
function FloatCond
{
    local ret=1
    if [ $# -gt 0 ]; then
        cond=$(echo "$*" | bc -q 2>/dev/null)
        if [ "$cond" = "1" ]; then
            ret=0
        else
            ret=1
        fi
    fi
    return $ret
}

## send position: $1% on output
# params
# $1 - number
function ProgressPosition
{
    if [ $PROGRESSBAR -eq 1 ]; then
        printf "position: %1.0f%%\n" $1 >&3
    fi
}

## send title: on output
# params
# $1 - title
function ProgressTitle
{
    if [ $PROGRESSBAR -eq 1 ]; then
        echo "title: ${1}" >&3
    fi
}

## send titles: on output
# params
# $1 - number of all prepared titles of progressbar
function ProgressTitles
{
    if [ $PROGRESSBAR -eq 1 ]; then
        echo "titles: ${1}" >&3
    fi
}

################## Migration facilities ###################

## copy file from tar or backup directory
# params:
# $1: src
#     -use for relativ src path "aa/bb" instead of "./aa/bb"
#     -trailing slash avoid creating an additional directory level at the
#      destination aa/bb/ -> bb or aa/bb -> aa/bb
# $2: dst
# $3: progress
# return: 0 successful
function CopyFileFromBackup
{
    if [ ! "$BACKUPDIR" ] && [ ! "$BACKUPFILE" ]; then
        err "internal error"
        exit 254
    fi

    # default scale used by float functions
    float_scale=2
    local src=$1
    local dst=$2
    local progressbar=$3
    local exitcode=1
    local tar_src="$src"
    local dir=""
    local name=""
    local cmd=""

    log "Copying/Migrating data '$src' -> '$dst'." $LOG_VERBOSE

    if [ "$BACKUPDIR" ]; then
        tar_src=""
        # for maintaining the last '/' in path
        dir=$src
        name="."
        if ! echo "$src" | grep -q "/$"; then
            dir=`dirname $src`
            name=`basename $src`
        fi
        cmd="tar -c -S -C \"${BACKUPDIR}/${dir}\" "$name" | tar"
    else
        # $BACKUPFILE is set
        # count '/' in $src
        strip=$(echo -n "$src"| sed -e 's/\/\/\+/\//g' -e 's/[^/]//g' |wc -c)
            cmd="tar -f \"$BACKUPFILE\" --strip-components=$strip"
    fi

    if [ "$progressbar" = "1" ]; then
        if [ "$BACKUPDIR" ]; then
            file_size=$(du -bc "$BACKUPDIR/$src" | tail -n 1 | cut -f 1)
        else
            # compute extracted filesize
            file_size=$(tar -tvf "$BACKUPFILE" | awk '{ c=c + $3;} END { print c }')
        fi

        # record size for tar, has to be multiplier of 512
        record_size=10240
        # 1% equals P file size
        p=$(FloatEval "$file_size / 100")

        # computing  checkpoint to maintain 1%/25%/50% increase
        for pi in 1 25 50; do
            log "[debug] trying PI=$pi" $LOG_VERBOSE
            pp=$(FloatEval "$p * $pi")
            checkpoint=$(FloatEval "$pp / $record_size")
            #if CHECKPOINT == 0, try next PI
            if FloatCond "$checkpoint > 1"; then
                 # computing correction
#    (("CHECKCORRECTION=($PP *10)/(($PP%$RECORD_SIZE)*10)"))
                log "[debug] selected PI=$pi" $LOG_VERBOSE
                log "[debug] PP=$pp" $LOG_VERBOSE
                break;
            fi
        done

        # if CHECKPOINT is still 0, then file is small enough to one 100% PI

        # correction
        int_checkpoint=$( echo "scale=0;$checkpoint/1" | bc -q)
        int_checkpoint=$(printf "%1.0f" "$checkpoint")
        pi=$(FloatEval "$pi * $int_checkpoint / $checkpoint")

        log "[debug] FILESIZE:   $file_size" $LOG_VERBOSE
        log "[debug] RECORDSIZE: $record_size" $LOG_VERBOSE
        log "[debug] CHECKPOINT: $checkpoint" $LOG_VERBOSE
        log "[debug] INT_CHEKPT: $int_checkpoint" $LOG_VERBOSE
        log "[debug] PERCENTINCL $pi" $LOG_VERBOSE

        # position counter
        position=0

        ( eval $cmd -xSk --no-overwrite-dir -C "$dst" \
            "--checkpoint-action=\"exec=echo checkpoint: \\\$TAR_CHECKPOINT\"" \
            --checkpoint=$int_checkpoint --record-size=$record_size \
            -- "$tar_src" 2>&1 ; echo "exitcode: $?" ) \
        | while read LINE; do

            if echo "$LINE" | grep -q "^checkpoint: \+[0-9]\+$"; then
                if FloatCond "$position <= 100.5"; then
                    ProgressPosition "$position"
                    position=$(FloatEval "$position + $pi")
                fi
            elif echo "$LINE" | grep -q "^exitcode: \+[0-9]\+$"; then
                local_exitcode=$(echo "$LINE" | cut -d' ' -f 2)
                # return from while loop/fork
                return $local_exitcode
            else
                log "$LINE" $LOG_VERBOSE
            fi
        done
        exitcode=$?
        # assuring 100% progress
        ProgressPosition 100
    else
        eval $cmd -xSk --no-overwrite-dir -C \"$dst\" -- "$tar_src" 2>/dev/null
        exitcode=$?
    fi
    if [ $exitcode -eq 0 ]; then
        log "'$src' -> '$dst' succeed." $LOG_VERBOSE
    else
        log "'$src' -> '$dst' failed ($exitcode) !"
    fi

    return $exitcode
}

## returns 0 if this file exist in backup file
# params:
#    $1 - testing file
function CheckFileInBackup
{
    local f=$1
    if [ "$BACKUPDIR" ]; then
        test -f "${BACKUPDIR}/$f"
        return $?
    elif [ "$BACKUPFILE" ]; then
        tar -tf "$BACKUPFILE" "$f" 2>/dev/null
        return $?
    fi
    return 1;
}

# extracts etc/SLEPOS/branchserver.conf and returns path to it
# set CONFFILE variable on success
function GetOldConfFile
{
    local f
    local ret
    local type=$1
    if [ ! "$tmpdir" ]; then
        err "internal error"
        exit 254
    fi
    if [ "$SUSEVERSION" = "9" ]; then
        CopyFileFromBackup etc/opt/SLES/POS/branchserver.conf "$tmpdir" 0
        ret=$?
        f="${tmpdir}/branchserver.conf"
    elif [ "$SUSEVERSION" = "10" -a -f "/etc/SLEPOS/branchserver.conf" ]; then
        mv "/etc/SLEPOS/branchserver.conf" "$f"
        ret=$?
        f="${tmpdir}/branchserver.conf"
    elif [ "$SUSEVERSION" = "11" ]; then
        if (($type & $DEPLOY_TYPE_AS )); then
            CopyFileFromBackup etc/SLEPOS/adminserver.conf "$tmpdir" 0
            ret=$?
            f="${tmpdir}/adminserver.conf"
        fi
        if (($type & $DEPLOY_TYPE_BS )); then
            CopyFileFromBackup etc/SLEPOS/branchserver.conf "$tmpdir" 0
            ret=$?
            f="${tmpdir}/branchserver.conf"
        fi
    fi
    if [ -f "$f" ]; then
        CONFFILE="$f"
        return $ret;
    else
        return 1;
    fi
}

## returns /etc/SuSE-release form backuped file/dir
# set OLDSUSERELEASE on success
function GetOldSuseRelease
{
    if [ ! "$tmpdir" ]; then
        err "internal error"
        exit 254
    fi
    CopyFileFromBackup "${SUSERELEASE}" "${tmpdir}" 0
    ret=$?
    f="${tmpdir}/SuSE-release"
    if [ -f "$f" ]; then
        OLDSUSERELEASE="$f"
        return $ret;
    else
        return 1;
    fi
}

## returns SuSE product release version
# params:
#      $1 - SuSE-release file
function GetSuseVersion
{
    local suserelease=$1
    if [ -f "$suserelease" ]; then
        grep "VERSION" "$suserelease" | awk '{ print $3 }'
    fi
}

## extracts ldif or xml from dump
# set LDAPFILE variable on success
function GetLDAPfile
{
    local f
    if [ ! "$tmpdir" ]; then
        err "internal error"
        exit 254
    fi
    if [ "$SUSEVERSION" = "9" ]; then
        CopyFileFromBackup "${NLPOSMIGRATIONDIR}/nlpos9-ldap.ldif" "$tmpdir" 0
        ret=$?
        f="${tmpdir}/nlpos9-ldap.ldif"
    elif [ "$SUSEVERSION" = "10" ]; then
        f="${tmpdir}/slepos-ldap.ldif"
        slapcat -d 0 -l "$f"
        ret=$?
    elif [ "$SUSEVERSION" = "11" ]; then
        CopyFileFromBackup "${SLEPOSMIGRATIONDIR}/slepos-ldap.ldif" "$tmpdir" 0
        # for now we use ldif from slapcat, but in future xml may come handy
        if [ -f "${tmpdir}/slepos-ldap.xml" ]; then
            f="${tmpdir}/slepos-ldap.xml"
        elif [ -f "${tmpdir}/slepos-ldap.ldif" ]; then
            f="${tmpdir}/slepos-ldap.ldif"
        fi
    fi
    if [ -f "$f" ]; then
        LDAPFILE="$f"
        return $ret;
    else
        return 1;
    fi
}
