#!/usr/bin/perl -w
#
#==========================================================================
# FILE        : posLdapMigration.pl
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : Copyright (c) 2013 SUSE LINUX Products GmbH, Nuernberg, Germany.
#
# Author      : Pavel Sladek <psladek@suse.com>
#               Ondrej Holecek <oholecek@suse.com>
#
# DESCRIPTION : LDIF data migration from NLPOS9, SLEPOS10, SLEPOS11 to SLEPOS11 SP2
#==========================================================================
## @file
# SLEPOS BranchServer network configuration
# @par Authors
# - Pavel Sladek
# - Ondrej Holecek
# @copy 2013, SUSE LINUX Products GmbH, Nuernberg, Germany.
#

# input file pathname - database output generated by slapcat
# outputfile pathname - upgraded database ready to be imported by ldapadd

use strict;
use warnings;

use POS::Config;
use Net::LDAP::LDIF;
use File::Temp qw/tempfile/;
use Getopt::Long;

our $nlpos9_s   = "v9";
our $slepos10_s = "v10";
our $slepos11_s = "v11";
our $pxedefault = "/boot/pxelinux.0";
my $verbose = 0;
my $configFile;

sub print_help
{
    print("$0 - LDIF data migration from NLPOS9, SLEPOS10, SLEPOS11 to SLEPOS11 SP2\n\n");
    print("usage:\n");
    print(
"$0 [-v] [-f|--config-file=<config-file>] $nlpos9_s|$slepos10_s|$slepos11_s <input-ldap-file> <output-ldap-file>\n\n"
    );
    exit 0;
} ## end sub print_help

# we take one mandatory argument input file and optional argument output file. If no output file -> stdout
# ignore slepos version as we'll detect whats missing automatically. Left there for compatibility reasona
my $optres         = GetOptions('help|usage|h' => \&print_help, 'v' => \$verbose, 'config-file|f=s' => \$configFile);
my $source_version = shift @ARGV;
my $input          = shift @ARGV;
my $output         = shift @ARGV;

if (not $input) {
    print STDERR "[ERROR] Missing input file!\n";
    exit 255;
}
if (not -f $input) {
    print STDERR "[ERROR] Cannot find input file $input\n";
    exit 255;
}
if (not $output) {
    print STDERR "[ERROR] Missing output file!\n";
    exit 255;
}

my $config;
eval {
    if ($configFile) {
        $config = POS::Config->new($configFile);
    }
    else {
        $config = POS::Config->new();
    }
};
if ($@) {
    if ($source_version !~ /$slepos11_s/) {
        print STDERR
          "[ERROR] Cannot load current SLEPOS configuration. Migrating from SLEPOS10 and older is not possible!\n";
        print STDERR "[ERROR] Initialize AdminServer first\n";
        exit 255;
    }
    #add empty placeholder, it won't be used on SLEPOS11+ anyway
    $config = { 'POS_ADMIN_PASSWORD' => '' };
}
my $in = Net::LDAP::LDIF->new($input, 'r');
my ($tmpfh, $tmpfile) = tempfile();
my $out = Net::LDAP::LDIF->new($tmpfh, 'w');

# get admin password from the config file to use as an scLocation password
my $adminpass = $config->{'POS_ADMIN_PASSWORD'};
if (!defined $adminpass) {
    print(
"[ERROR] 'POS_ADMIN_PASSWORD' not found in the /etc/SLEPOS/branchserver.conf or /etc/SLEPOS/adminserver.conf, aborting.. \n\n"
    );
    exit 255;
}
# hash to store scDhcpOptionsRemote from scPosImages objects, needed for proper scHardware setup
my %scDhcpOptionsRemoteInImages = ();
# hash of containers DNs whose branchservers to skip
my %branchServersToDelete = ();
# Net::LDAP::Entry of scHarware standars
my $scHardwareEntry;
print "Updating services and locations\n";
while (not $in->eof()) {
    my $entry         = $in->read_entry();
    my @objectClasses = $entry->get_value('objectClass');
    my $isLocation    = 0;
    my $isHAService   = 0;
    my $isHardware    = 0;
    my $i             = 0;
    for ($i = 0; $i < scalar @objectClasses; ++$i) {
        if ($objectClasses[$i] eq 'scHAService') {
            $isHAService = 1;
            last;
        }
        elsif ($objectClasses[$i] eq 'scLocation') {
            $isLocation = 1;
            last;
        }
        elsif ($objectClasses[$i] eq 'scHardware') {
            $isHardware = 1;
            last;
        }
        elsif ($objectClasses[$i] eq 'scPosImage') {
            $scDhcpOptionsRemoteInImages{ $entry->dn() } = $entry->get_value('scDhcpOptionsRemote');
            last;
        }
    }
    if ($isHAService) {
        $isHAService = 0;
        ### convert scHAService -> scService
        my ($branchDN) = $entry->dn() =~ /^[^,]+,(.+)$/;
        # initialize branch score
        if (not $branchServersToDelete{$branchDN}) {
            $branchServersToDelete{$branchDN} = 0;
        }
        my $primary = $entry->get_value('scPrimaryService') eq 'TRUE';
        if (not $primary) {
            $branchServersToDelete{$branchDN} += 1;
            # do not store non primary scHAService
            next;
        }
        else {
            $branchServersToDelete{$branchDN} -= 1;
            $objectClasses[$i] = 'scService';
            $entry->replace('objectClass' => \@objectClasses);
            $entry->delete('scPrimaryService');
            $entry->delete('structuralObjectClass');
            printf('Updating scHAService to scService at %s%s', $entry->dn(), "\n") if $verbose;
        }
    }
    elsif ($isLocation) {
        $isLocation = 0;
        ### check for missing scDhcpExtern and userPassword
        my @tmp = $entry->get_value('scDhcpExtern');
        if ($#tmp == -1) {
            $entry->add('scDhcpExtern' => 'FALSE');
            printf('Adding missing scDhcpExtern => FALSE to scLocation %s%s', $entry->dn(), "\n") if $verbose;
        }
        @tmp = $entry->get_value('userPassword');
        if ($#tmp == -1) {
            $entry->add('userPassword' => $adminpass);
            printf('Adding missing userPassword to scLocation %s%s', $entry->dn(), "\n") if $verbose;
        }
    }
    elsif ($isHardware) {
        $isHardware      = 0;
        $scHardwareEntry = $entry;
        # do not write entry yet, wait for all images to be parsed
        next;
    }
    $entry->delete(
        'structuralObjectClass' => [],
        'entryUUID'             => [],
        'creatorsName'          => [],
        'createTimestamp'       => [],
        'entryCSN'              => [],
        'contextCSN'            => [],
        'modifiersName'         => [],
        'modifyTimestamp'       => []
    );
    $out->write_entry($entry);
}
# close input file
$in->done();
# now finish writing scHardware
print "Updating scHardware\n";
if ($scHardwareEntry) {
    my @tmp = $scHardwareEntry->get_value('scDhcpOptionsRemote');
    if ($#tmp == -1) {
        my $pxe = $pxedefault;
        # there may be scPosImageDn attributes. If yes, look at that image DN and find its scDhcpOptionRemote
        @tmp = $scHardwareEntry->get_value('scPosImageDn');
        if ($#tmp == 0) {
            #we have exactly one, try to use it as scDhcpOptionsRemote source
            if ($scDhcpOptionsRemoteInImages{ $tmp[0] }) {
                # scDhcpOptionsRemote was set, use it
                $pxe = $scDhcpOptionsRemoteInImages{ $tmp[0] };
            }
        }
        # $pxe is either $pxedefault or remote options from refered image
        $scHardwareEntry->add('scDhcpOptionsRemote' => $pxe);
        printf('Adding scDhcpOptionsRemote => %s to scHardware %s\n', $pxe, $scHardwareEntry->dn()) if $verbose;
    }
    $out->write_entry($scHardwareEntry);
}
# close output file, close fh as it was created manually
$out->done();
close($tmpfh);
# if we have any scBranchServer to delete perform second pass and strip them out
print "Checking for HA scBranchServer subtrees...";
if (keys %branchServersToDelete) {
    print "found. Deleting redundant subtrees\n";
    # evaluate branchserver to delete -> delete all which does not have at least 3 converted services
    # assuming working input solution, so no typos in primary services, allowing typos in secondary services
    my @deleteMasks = ();
    foreach (keys %branchServersToDelete) {
        if ($branchServersToDelete{$_} > -3) {
            push @deleteMasks, $_;
        }
    }
    $in  = Net::LDAP::LDIF->new($tmpfile, 'r');
    $out = Net::LDAP::LDIF->new($output,  'w');
    while (not $in->eof()) {
        my $entry      = $in->read_entry();
        my $dn         = $entry->dn();
        my $allowWrite = 1;
        foreach (@deleteMasks) {
            if ($dn =~ $_) {
                printf('Deleting %s%s', $dn, "\n") if $verbose;
                $allowWrite = 0;
                last;
            }
        }
        if ($allowWrite) {
            $out->write_entry($entry);
        }
    }
    $in->done();
    $out->done();
    unlink($tmpfile);
}
else {
    print "not found.\n";
    rename($tmpfile, $output);
}
print "All done!\n";
exit 0;
