#!/bin/bash

#
#==========================================================================
# FILE        : slepos11_get_deploy_type.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Ondrej Holecek <oholecek@novell.com>, 2009
#
# DESCRIPTION : Determining installed patterns.
#               AS/BS/IS and combinations
#==========================================================================

appName=$(basename ${0})
appVersion="0.1"
appDir=$(dirname ${0})

#source common
MIGRATION_COMMON="${appDir}/migration_common.sh"
test -r "$MIGRATION_COMMON" || exit 255
. "$MIGRATION_COMMON"

function usage
{
    cat << EOF_USAGE
${appName} v${appVersion}

Usage:  ${appName} [OPTIONS] [-f|--file=<FILE>] -o|--old-release=<FILE>
        ${appName} [OPTIONS] [-d|--directory=<DIR>] -o|--old-release=<FILE>
        ${appName} -h|--help
EOF_USAGE
}

function displayHelp
{
    usage
    cat << EOF_HELP

Description:
    Detects installed components of NLPOS9/SLEPOS10 system.

Returns:
    0 if no component is installed
    1 for AS
    2 for BS
    4 for IS
    and combinations
    255 for error

Options:
 -f, --file=<FILE>         path to file with backed up configuration from
                           before migration
 -d, --directory=<DIR>     path to directory with backed up configuration
                           from before migration
 -o, --old-release=<FILE>  path to stored old SuSE-release file
 -v, --verbose             enables verbose logging
 -V, --version             print the version of script being used
 -h, --help                displays this help
 --usage                   display brief usage message
EOF_HELP
}

# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o f:d:o:l:vVh -l file:,directory:,old-release:,log:,help,usage,verbose,version \
      -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 254; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"

#load up parameters
while true; do
    case $1 in
        -f|--file)
            log "Loading backup file" $LOG_VERBOSE
            BACKUPFILE=$2
            shift 2
            if [ ! -f "$BACKUPFILE" ]; then
                err "Backup file not found. Error during -f parameter validation.(${appName})"
            fi
            ;;
        -d|--directory)
            log "Loading backup directory" $LOG_VERBOSE
            BACKUPDIR=$2
            shift 2
            if [ ! -d "$BACKUPDIR" ]; then
                err "Backup directory not found. Error during -d parameter validation.(${appName})"
            fi
            ;;
        -o|--old-release)
            log "Loading old SuSE-release" $LOG_VERBOSE
            OLDRELEASE=$2
            shift 2
            if [ ! -f "$OLDRELEASE" ]; then
                err "Old SuSE-release file not found.(${appName})"
            fi
            ;;
        -l|--log)
            setLogLevel $LOG_VERBOSE
            LOGFILENAME=$2
            SetFilOutput 1
            shift 2
            ;;
        -v|--verbose)
            setLogLevel $LOG_VERBOSE
            shift
            ;;
        -q|--quiet)
            setLogLevel $LOG_OFF
            shift
            ;;
        -h|--help)
            displayHelp; exit 0 ;;
        --usage) usage; exit 0 ;;
        -V|--version) echo "${appName} ${appVersion}"; exit 0 ;;
        --) shift; break ;;
        *)
           err "Internal error!"
           exit 254
           ;;
    esac
done

if [ $# -gt 0 ]; then
    log "[ERROR] odd addition arguments '$*'"
    usage
    exit 255
fi

tmpdir=`mktemp --tmpdir -d tmp.nlpos9-migration.XXXXXXXXXX`

RESULT=0
log "Checking previous system installation version."

#set value if is not set
if [ ! "$OLDRELEASE" ] && GetOldSuseRelease; then
    OLDRELEASE="$OLDSUSERELEASE"
fi

SUSEVERSION=$(GetSuseVersion "$OLDRELEASE")

if [ "$SUSEVERSION" = "9" ]; then
    GetOldConfFile
elif [ "$SUSEVERSION" = "10" ]; then
    # Migration from SLEPOS10 is done by OS upgrade
    CONFFILE="/etc/SLEPOS/branchserver.conf"
else
    logFailure "Checking SuSE-release."
    err "Unrecognized SuSE product version!"
fi
log "Checking for configuration file...."
if [ -f "$CONFFILE" ]; then
    logSuccess
    ADMIN=$(grep "POS_ADMIN_DEPLOYED" "$CONFFILE" | cut -d= -f2)
    BRANCH=$(grep "POS_BRANCH_DEPLOYED" "$CONFFILE" | cut -d= -f2)
    if [ "$ADMIN" = "yes" ]; then
        log "Administration Server installation detected."
        (( RESULT=$RESULT + 1 ))
    fi
    if [ "$BRANCH" = "yes" ]; then
        log "Branch Server installation detected."
        (( RESULT=$RESULT + 2 ))
    fi
fi

# POS_Image2 for NLPOS9 or POS_Image3 for SLEPOS10 
if ( [ "$SUSEVERSION" = "10" ] && rpm --quiet -q POS_Image3 ) ||
   ( [ "$SUSEVERSION" = "9" ] && CheckFileInBackup "${NLPOSMIGRATIONDIR}/RPM/POS_Image2" ); then
    log "Image Server installation detected."
    (( RESULT=$RESULT + 4 ))
fi

log "[DEBUG] exitcode = ${RESULT}" $LOG_VERBOSE
rm -rf "$tmpdir"
exit $RESULT
