#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <hd.h>		// for SMB table entry types (sm_biosinfo...)
#include <ctype.h>	// isalpha


#define POSBIOS_MAJOR		1
#define POSBIOS_MINOR		1
#define POSBIOS_REVISION	0

// 1.1.0 - 05/15/2006
//         Added code to remove punctuation from manufacturer name

// was 0xc0000
#define MEMORY_LOW_RANGE	0xf0000
#define MEMORY_SIZE			0x10000
#define SMB_PAR_SIZE			0x10
#define SMB_FILE_DEV			"/dev/mem"

#pragma pack(1)
typedef struct _st_SMB_Header
{
	char				ch_SM_[4];			// 0x00 (4) "_SM_" (5F 53 4D 5F)
	char				bEPSchecksum;		// 0x04 (1) checksum byte (added to other bytes in EPS (Entry Point Structure) == 00
	char				bEPSlen;				// 0x05 (1)	length of EPS ( >= 0x1F )
	char				bEPSmajor;			// 0x06 (1)	major version of tables spec
	char				bEPSminor;			// 0x07 (1)	minor version of tables spec
	short 			iMaxStructSize;	// 0x08 (2)	size of largest SMBIOS structure 
	char				bEPSrevision;		// 0x0A (1) EPS revision (00 = SMBIOS 2.1)
	char				chFormatted[5];	// 0x0B (5)	formatted based on EPS revision
	char				ch_DMI_[5];			// 0x10 (5) "_DMI_" - 5F 44 4D 49 5F
	char				bIEPSchecksum;		// 0x15 (1) checksum byte (added to bytes in IEPS(0x10 for 0x0F bytes) == 00
	short				iTableSize;			// 0x16 (2) SMBIOS Structure Table size in bytes
	unsigned 		iTableAddress;		// 0x18 (4) SMBIOS Structure Table address
	short				iTableEntries;		// 0x1C (2) SMBIOS Structure Table number of entries
	char				bBCDrevision;		// 0x1E (1) BCD revision - upper nibble is major version lower nibble is minor version

}	st_SMB_Header, *pst_SMB_Header;

typedef struct _st_SMB_Table_Entry
{
	unsigned char	bType;				// 0x00 (1) entry type (hd_smbios_type_t in hd.h)
	unsigned char	bEntryLen;			// 0x01 (1) entry size (not including string set?)
	unsigned short	uHandle;				// 0x02 (2) entry unique handle

} st_SMB_Table_Entry, *pst_SMB_Table_Entry;
#pragma pack()

typedef struct _st_funcparms
{
	int					iVerbose;	// display calculations
	hd_smbios_type_t	hd_type;		// smbios entry type
	int					iPrefix;		// prefix manufacturer

}	st_funcparms, *pst_funcparms;

typedef struct _st_membuff
{
	int				iBuffSize;
	unsigned 		iBaseAddress;
	unsigned char* cpBuff;

}	st_membuff, *pst_membuff;


void show_usage()
{
	fprintf( stderr, "\nposbios (%d.%d.%d)\n\n", POSBIOS_MAJOR, POSBIOS_MINOR, POSBIOS_REVISION );
	fprintf( stderr, "posbios [-m] [-b] [-s] [-p] [-v]\n\n" );
	fprintf( stderr, "   -m=  Show manufacturer prefix\n\n" );
	fprintf( stderr, "   -b=  Show BIOS\n" );
	fprintf( stderr, "   -s=  Show System (default)\n" );
	fprintf( stderr, "   -p=  Show Product (boardinfo)\n" );
	fprintf( stderr, "   -v=  Verbose Output\n" );
	fprintf( stderr, "\n\n" );
}

int parse_parms( int argc, char** argv, pst_funcparms p )
{
   char        *cp;
   int         argi;

	p->iVerbose= 0;
	p->hd_type=  sm_end;
	p->iPrefix=  0;

   argi= 1;
   while( argi < argc )
   {
		if( argv[argi][0] == '-' )
		{
			cp= &argv[argi][1];
			while( *cp && *cp == ' ' )
				cp++;
			while( *cp )
			{
				if( *cp == '?' || *cp == 'h' )
					show_usage();
				if( *cp == 'b' || *cp == 'B' )
					p->hd_type= sm_biosinfo;
				if( *cp == 's' || *cp == 'C' )
					p->hd_type= sm_sysinfo;
				if( *cp == 'p' || *cp == 'P' )
					p->hd_type= sm_boardinfo;
				if( *cp == 'm' || *cp == 'M' )
					p->iPrefix= 1;
				if( *cp == 'v' || *cp == 'V' )
					p->iVerbose= 1;
				cp++;
			}
		}
		argi++;
	}
	// default to type 1
	if( p->hd_type == sm_end )
		p->hd_type= sm_sysinfo;

	return 0;
}


int calloc_and_copy_prefix( char** cppTo, char* cpFrom )
{
	char* cpIndex;


	if( cppTo )
	{
		*cppTo= NULL;
		cpIndex= cpFrom;
		if( cpIndex )
		{
			while( *cpIndex && isalpha(*cpIndex) )
				cpIndex++;
			if( cpIndex > cpFrom )
			{
				*cppTo= (char*)calloc(cpIndex - cpFrom +1, sizeof(char) );
				if( *cppTo )
				{
					strncpy( *cppTo, cpFrom, cpIndex - cpFrom );
					(*cppTo)[cpIndex - cpFrom]= '\0';
					return cpIndex - cpFrom;
				}
			}
		}
	}

	return 0;
}



int read_buffer( pst_membuff pBuff, char* cpDevice )
{
	struct stat	sstat;
	off_t 		omapstart;
	void			*pvoid;
	unsigned 	umapsize;
	unsigned		upagesize;
	int			fDev;


	if( !pBuff || !pBuff->cpBuff || !pBuff->iBuffSize )
		return 0;

	if( !cpDevice || !strlen(cpDevice) )
		return 0;

	memset( pBuff->cpBuff, 0, pBuff->iBuffSize );

	upagesize= getpagesize();
	omapstart= pBuff->iBaseAddress & (-upagesize);
	umapsize=  (pBuff->iBuffSize + (pBuff->iBaseAddress - omapstart) + upagesize -1) & -upagesize;

	fDev= open( cpDevice, O_RDONLY );
	if( fDev > -1 )
	{
		if( !fstat( fDev, &sstat ) )
		{
			// if regular file then make requested block is in file range
			if( S_ISREG( sstat.st_mode ) && sstat.st_size < pBuff->iBaseAddress + pBuff->iBuffSize )
			{
				if( sstat.st_size > pBuff->iBaseAddress )
				{
					pBuff->iBuffSize= sstat.st_size - pBuff->iBaseAddress;
				}
				else
				{
					fprintf( stderr, "read_buffer() called with address outside of file %s.\n", cpDevice );
					pBuff->iBuffSize= 0;
				}
			}
		}
		else
		{
			fprintf( stderr, "read_buffer() failed to get stats for %s.\n", cpDevice );
			pBuff->iBuffSize= 0;
		}

		if( pBuff->iBuffSize )
		{
			pvoid= mmap( NULL, umapsize, PROT_READ, MAP_PRIVATE, fDev, omapstart );
			if( pvoid != MAP_FAILED )
			{
				memcpy( pBuff->cpBuff, pvoid + (pBuff->iBaseAddress - omapstart), pBuff->iBuffSize );
				munmap( pvoid, umapsize );
			}
			else
			{
				fprintf( stderr, "read_buffer() failed to map memory for %s.\n", cpDevice );
				pBuff->iBuffSize= 0;
			}
		}

		close( fDev );
	}
	else
	{
		fprintf( stderr, "read_buffer() failed to open %s. (are you running as root?)\n", cpDevice );
		pBuff->iBuffSize= 0;
	}

	return pBuff->iBuffSize;
}

// return pointer to smb header
pst_SMB_Header scan_buff_for_smb( pst_membuff pBuff, pst_funcparms fp )
{
	pst_SMB_Header	pSMB = { 0 };
	unsigned int	uMemOffset;
	unsigned char	chSum;
	int				iIndex;
	int				iHaveSMB;


	if( !pBuff || !pBuff->cpBuff || !pBuff->iBuffSize )
		return NULL;

	uMemOffset= 0;
	iHaveSMB=   0;
	while( !iHaveSMB && uMemOffset < pBuff->iBuffSize - SMB_PAR_SIZE )
	{
		if( *(unsigned*)(pBuff->cpBuff + uMemOffset) == (unsigned)0x5f4d535f )
		{
			pSMB= (pst_SMB_Header)(pBuff->cpBuff + uMemOffset);
			for( chSum= 0, iIndex= 0; iIndex < pSMB->bEPSlen; iIndex++ )
				chSum+= *(unsigned char*)(pBuff->cpBuff + uMemOffset + iIndex);
			if( chSum == 0x00 )
				iHaveSMB= 1;
		}
		uMemOffset+= SMB_PAR_SIZE;
	}

	if( fp->iVerbose )
	{
		if( iHaveSMB )
			fprintf( stderr, "scan_buff_for_smb() - Have SMB header at offset %04x\n", (unsigned)pSMB - (unsigned)pBuff->cpBuff );
		else
			fprintf( stderr, "scan_buff_for_smb() - SMB header not found\n" );
	}

	if( iHaveSMB )
		return pSMB;
	return NULL;
}


pst_SMB_Table_Entry smb_next_table_entry( pst_SMB_Table_Entry pEntry, pst_membuff pMem )
{
	unsigned char* cpIndex;
	unsigned char*	cpReturn;


	cpReturn= NULL;
	if( pEntry && pMem && pEntry->bType != sm_end )
	{
		cpIndex= (unsigned char*)((unsigned)pEntry + pEntry->bEntryLen);
		while( !cpReturn && (unsigned)pEntry <= (unsigned)(pMem->cpBuff + pMem->iBuffSize - sizeof(st_SMB_Table_Entry)) )
		{
			while( *cpIndex )
				cpIndex++;
			if( *cpIndex == 0x00 && *(cpIndex+1) == 0x00 )
				cpReturn= cpIndex + 2;
			else
				cpIndex++;
		}
	}

	if( !pEntry || pEntry->bType == sm_end )
		return NULL;
	return (pst_SMB_Table_Entry)cpReturn;
}


int main( int argc, char** argv )
{
   st_funcparms   		fp;
	st_membuff				st_smb;
	st_membuff				st_tbl;
	pst_SMB_Header			pSMB;
	pst_SMB_Table_Entry	pTableStart;
	pst_SMB_Table_Entry	pEntry;
	int						iFailed;
	char						*cp;
	char						*cpManufacturer;
	char						*cpProduct;


	iFailed= -1;		// assume failed until product is printed

	cpManufacturer= NULL;
	cpProduct=      NULL;

	pSMB=        	NULL;
	pTableStart= 	NULL;
	st_tbl.cpBuff= NULL;

	parse_parms( argc, argv, &fp );

	st_smb.iBaseAddress= MEMORY_LOW_RANGE;
	st_smb.iBuffSize= 	MEMORY_SIZE;
	st_smb.cpBuff= (unsigned char*)calloc(st_smb.iBuffSize, sizeof(unsigned char));
	if( st_smb.cpBuff )
	{
		if( read_buffer( &st_smb, SMB_FILE_DEV ) )
			pSMB= scan_buff_for_smb( &st_smb, &fp );

		if( pSMB && pSMB->iTableAddress && pSMB->iTableEntries && pSMB->iTableSize )
		{
			st_tbl.iBaseAddress= pSMB->iTableAddress;
			st_tbl.iBuffSize=    pSMB->iTableSize;
			st_tbl.cpBuff= (unsigned char*)calloc(st_tbl.iBuffSize, sizeof(unsigned char));
			if( st_tbl.cpBuff )
			{
				if( read_buffer( &st_tbl, SMB_FILE_DEV ) )
					pTableStart= (pst_SMB_Table_Entry)st_tbl.cpBuff;
			}
		}
	}

	if( pSMB && pTableStart )
	{
		pEntry= pTableStart;
		while( pEntry && pEntry->bType != fp.hd_type )
		{
			if( fp.iVerbose )
				fprintf( stderr, "main() - %3d:  #%6d   skipped\n", pEntry->bType, pEntry->uHandle );
			pEntry= smb_next_table_entry( pEntry, &st_tbl );
		}

		if( pEntry && pEntry->bType == fp.hd_type )
		{
			if( fp.iVerbose )
				fprintf( stderr, "main() - %3d:  #%6d   MATCH\n", pEntry->bType, pEntry->uHandle );

			// for types 0,1,2 first string is Manufacturer
			cp= (char*)((unsigned)pEntry + pEntry->bEntryLen);
			if( *cp )
			{
				if( fp.iVerbose )
					fprintf( stderr, "Manufacturer: %s\n", cp );
				calloc_and_copy_prefix( &cpManufacturer, cp );
				cp+= strlen(cp)+1;
			}

			if( *cp )
			{
				if( fp.iVerbose )
					fprintf( stderr, "Product: %s\n", cp );
				cpProduct= (char*)calloc(strlen(cp) +1, sizeof(char));
				if( cpProduct )
				{
					strcpy( cpProduct, cp );
					iFailed= 0;						// if product is printed then return success
				}
				cp+= strlen(cp)+1;
			}

			if( fp.iVerbose )
			{
				while( *cp )
				{
					fprintf( stderr, "string: %s\n", cp );
					cp+= strlen(cp)+1;
				}
			}
		}
		else
		{
			if( fp.iVerbose )
				fprintf( stderr, "main() - end of table looking for SMB type %d\n", fp.hd_type );
		}
	}
	else
	{
		if( fp.iVerbose )
			fprintf( stderr, "main() - failed to find SMB header information.\n" );
	}

	if( fp.iPrefix && cpManufacturer )
		fprintf( stdout, "%s", cpManufacturer );
	fprintf( stdout, "%s", cpProduct ? cpProduct : "UNKNOWN"  );
	fprintf( stdout, "\n" );

	if( cpProduct )
		free( cpProduct );
	if( cpManufacturer )
		free( cpManufacturer );

	if( st_tbl.cpBuff )
		free( st_tbl.cpBuff );
	if( st_smb.cpBuff )
		free( st_smb.cpBuff );

	return iFailed;
}
