#================
# FILE          : KIWILinuxRC-lite.sh
#----------------
# PROJECT       : SLEPOS
# COPYRIGHT     : (c) 2013 SUSE LINUX Products GmbH, Germany
#               :
# AUTHOR        : Marcus Schaefer <ms@suse.de>
#               : Vladimir Nadvornik <nadvornik@suse.cz>
#               :
# BELONGS TO    : Operating System images
#               :
# DESCRIPTION   : This module contains some functions
#               : extracted from KIWILinuxRC.sh
#               :
# STATUS        : Development
#----------------
#======================================
# Exports (General)
#--------------------------------------

export TMPDIR=$(mktemp -d /tmp/slepos.XXXXXXXXXX)

export TRANSFER_ERRORS_FILE=$TMPDIR/transfer.errors

#======================================
# Echo
#--------------------------------------
function Echo {
	# /.../
	# print a message to the controling terminal
	# ----
	echo "$@"
}
#======================================
# importFile
#--------------------------------------
function importFile {
	# /.../
	# import the config.<MAC> style format. the function
	# will export each entry of the file as variable into
	# the current shell environment
	# ----
	local prefix=$1
	cat - | grep -v ^# > $TMPDIR/srcme
	# remove start/stop quoting from values
	sed -i -e s"#\(^[a-zA-Z0-9_]\+\)=[\"']\(.*\)[\"']#\1=\2#" $TMPDIR/srcme
	# remove backslash quotes if any
	sed -i -e s"#\\\\\(.\)#\1#g" $TMPDIR/srcme
	# quote simple quotation marks
	sed -i -e s"#'#'\\\\''#g" $TMPDIR/srcme
	# add '...' quoting to values
	sed -i -e s"#\(^[a-zA-Z0-9_]\+\)=\(.*\)#$prefix\1='\2'#" $TMPDIR/srcme
	source $TMPDIR/srcme
	while read line;do
		key=$(echo "$line" | cut -d '=' -f1)
		eval "export $key"
	done < $TMPDIR/srcme
	if [ ! -z "$ERROR_INTERRUPT" ];then
		Echo -e "$ERROR_INTERRUPT"
		systemException "*** interrupted ****" "shell"
	fi
}
#======================================
# unsetFile
#--------------------------------------
function unsetFile {
	# /.../
	# unset variables specified within the given file.
	# the file must be in the config.<MAC> style format
	# ----
	IFS="
	"
	local prefix=$1 #change name of key with a prefix
	while read line;do
		echo $line | grep -qi "^#" && continue
		key=`echo "$line" | cut -d '=' -f1`
		if [ -z "$key" ];then
			continue
		fi
		Debug "unset $prefix$key"
		eval unset "$prefix$key"
	done
	IFS=$IFS_ORIG
}
#======================================
# systemException
#--------------------------------------
function systemException {
	Echo "$1"
	exit 1
}
#======================================
# runHook
#--------------------------------------
function runHook {
	HOOK="/kiwi-hooks/$1.sh"
	if [ ! -e $HOOK ];then
		HOOK="/lib/kiwi/hooks/$1.sh"
	fi
	if [ -e $HOOK ]; then
		. $HOOK "$@"
	fi
}
#======================================
# validateBlockSize
#--------------------------------------
function validateBlockSize {
	# /.../
	# check the block size value. atftp limits to a maximum of
	# 65535 blocks, so the block size must be checked according
	# to the size of the image. The block size itself is also
	# limited to 65464 bytes
	# ----
	local blkTest
	local nBlk
	if [ -z "$zblocks" ] && [ -z "$blocks" ];then
		# md5 file not yet read in... skip
		return
	fi
	if [ ! -z "$zblocks" ];then
		isize=`expr $zblocks \* $zblocksize`
	else
		isize=`expr $blocks \* $blocksize`
	fi
	local IFS=' '
	testBlkSizes="32768 61440 65464"
	if [ "$imageBlkSize" -gt 0 ]; then
		testBlkSizes="$imageBlkSize $testBlkSizes"
	fi
	for blkTest in $testBlkSizes ; do
		nBlk=`expr $isize / $blkTest`
		if [ $nBlk -lt 65535 ] ; then
			imageBlkSize=$blkTest
			return
		fi
	done
	systemException \
		"Maximum blocksize for atftp protocol exceeded" \
	"reboot"
}

#======================================
# atftpProgress
#--------------------------------------
function atftpProgress {
	# /.../
	# atftp doesn't use a stream based download and sometimes
	# seek back and forth which makes it hard to use pipes for
	# progress indication. Therefore we watch the trace output
	# ----
	local imgsize=$1    # image size in MB
	local prefix=$2     # line prefix text
	local file=$3       # file with progress data
	local blocksize=$4  # blocksize use for download
	local bytes=0       # log lines multiplied by blocksize
	local lines=0       # log lines
	local percent=0     # in percent of all
	local all=$((imgsize * 1024 * 1024))
	local line
	local step=0
	# number of cycles for approx. 2% steps
	local max_step=$(($all / $blocksize / 25))
	cat < dev/null > $file.tmp
	#======================================
	# print progress information
	#--------------------------------------
	while read line ;do
		echo "$line" >> $file.tmp
		let step=step+1
		if [ $step -lt $max_step ]; then
			continue
		fi
		step=0
		# /.../
		# the trace logs two lines indicating one download block of
		# blocksize bytes. We assume only full blocks. At the end
		# it might happen that only a part of blocksize bytes is
		# required. The function does not precisely calculate them
		# and assumes blocksize bytes. imho that's ok for the progress
		# bar. In order to be exact the function would have to sum
		# up all bytes from the trace log for each iteration which
		# would cause the download to pause because it has to wait
		# for the progress bar to get ready
		# ----
		# the same block can be transferred multiple times
		lines=$(grep "^sent ACK" $file.tmp | sort -u | wc -l)
		bytes=$((lines * $blocksize))
		percent=$(echo "scale=2; $bytes * 100"  | bc)
		percent=$(echo "scale=0; $percent / $all" | bc)
		echo -en "$prefix ( $percent%)\r"
	done
	grep -v "^\(received \)\|\(sent \)" $file.tmp > $file
	rm $file.tmp
	echo
}

#======================================
# encodeURL
#--------------------------------------
function encodeURL {
	# /.../
	# encode special characters in URL's to correctly
	# serve as input for fetchFile and putFile
	# ----
	local STR
	local CH
	STR="$@"
	echo -n "$STR" | while read -n1 CH; do
		[[ $CH =~ [-_A-Za-z0-9./] ]] && printf "$CH" || printf "%%%x" \'"$CH"
	done
}

#======================================
# fetchFile
#--------------------------------------
function fetchFile {
	# /.../
	# the generic fetcher which is able to use different protocols
	# tftp,ftp, http, https. fetchFile is used in the netboot linuxrc
	# and uses curl and atftp to download files from the network
	# ----
	local path=$1
	local dest=$2
	local izip=$3
	local host=$4
	local type=$5
	local chunk=$6
	local encoded_path
	local dump
	local call
	local call_pid
	if test -z "$chunk";then
		chunk=4k
	fi
	if test -z "$path"; then
		systemException "No path specified" "reboot"
	fi
	if test -z "$host"; then
		if test -z "$SERVER"; then
			systemException "No server specified" "reboot"
		fi
		host=$SERVER
	fi
	if test -z "$type"; then
		if test -z "$SERVERTYPE"; then
			type="tftp"
		else
			type="$SERVERTYPE"
		fi
	fi
	if test "$izip" = "compressed"; then
		path=$(echo "$path" | sed -e s@\\.gz@@)
		path="$path.gz"
	fi
	encoded_path=$(encodeURL "$path")
	#======================================
	# setup progress meta information
	#--------------------------------------
	dump="dd bs=$chunk of=\"$dest\""
	showProgress=0
	if [ -x /usr/bin/dcounter ] && [ -f /etc/image.md5 ] && [ -b "$dest" ];then
		showProgress=1
		read sum1 blocks blocksize zblocks zblocksize < /etc/image.md5
		needBytes=$((blocks * blocksize))
		needMByte=$((needBytes / 1048576))
		if [ ! -z "$zblocks" ];then
			needZBytes=$((zblocks * zblocksize))
			needZMByte=$((needZBytes / 1048576))
		fi
		progressBaseName=$(basename "$path")
		TEXT_LOAD=$(getText "Loading %1" "$progressBaseName")
		TEXT_GZIP=$(getText "Uncompressing %1" "$progressBaseName")
		dump="dcounter -s $needMByte -l \"$TEXT_LOAD \" 2>/progress | $dump"
	fi
	#======================================
	# build download command
	#--------------------------------------
	case "$type" in
		"local")
			if test "$izip" = "compressed"; then
				call="gzip -d < $host/$path \
					2>$TRANSFER_ERRORS_FILE | $dump"
			else
				call="dd if=$host/$path bs=$imageBlkSize |\
					$dump"
			fi
			;;
		"http")
			if test "$izip" = "compressed"; then
				call="curl -f http://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					gzip -d 2>>$TRANSFER_ERRORS_FILE | $dump"
			else
				call="curl -f http://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					$dump"
			fi
			;;
		"https")
			if test "$izip" = "compressed"; then
				call="curl -f -k https://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					gzip -d 2>>$TRANSFER_ERRORS_FILE | $dump"
			else
				call="curl -f -k https://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					$dump"
			fi
			;;
		"ftp")
			if test "$izip" = "compressed"; then
				call="curl ftp://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					gzip -d 2>>$TRANSFER_ERRORS_FILE | $dump"
			else
				call="curl ftp://$host/$encoded_path \
					2>$TRANSFER_ERRORS_FILE |\
					$dump"
			fi
			;;
		"tftp")
			validateBlockSize
			# /.../
			# atftp activates multicast by '--option "multicast"'
			# and deactivates it again  by '--option "disable multicast"'
			# ----
			if [ -f /etc/image.md5 ] && [ -b "$dest" ];then
				# enable multicast for system image and transfer to block device
				multicast_atftp="multicast"
			else
				# disable multicast for any other transfer
				multicast_atftp="disable multicast"
			fi
			havetemp_dir=1
			if [ -z "$FETCH_FILE_TEMP_DIR" ];then
				# we don't have a tmp dir available for downloading
				havetemp_dir=0
			elif [ -e "$FETCH_FILE_TEMP_DIR/${path##*/}" ];then
				# temporary download data already exists
				havetemp_dir=0
			else
				# prepare use of temp files for compressed download
				FETCH_FILE_TEMP_FILE="$FETCH_FILE_TEMP_DIR/${path##*/}"
				export FETCH_FILE_TEMP_FILE
			fi
			if test "$izip" = "compressed"; then
				if [ $havetemp_dir -eq 0 ];then
					# /.../
					# operate without temp files, standard case
					# ----
					call="busybox tftp \
						-b $imageBlkSize -g -r \"$path\" \
						-l >(gzip -d 2>>$TRANSFER_ERRORS_FILE | $dump) \
						$host 2>>$TRANSFER_ERRORS_FILE"
				else
					# /.../
					# operate using temp files
					# export the path to allow temp file management in a hook
					# ----
					if [ $showProgress -eq 1 ];then
						call="(atftp \
							--trace \
							--option \"$multicast_atftp\"  \
							--option \"blksize $imageBlkSize\" \
							-g -r \"$path\" -l \"$FETCH_FILE_TEMP_FILE\" \
							$host 2>&1 | \
							atftpProgress \
								$needZMByte \"$TEXT_LOAD\" \
								$TRANSFER_ERRORS_FILE $imageBlkSize \
							>&2 ; \
							gzip -d < \"$FETCH_FILE_TEMP_FILE\" | \
								dcounter -s $needMByte -l \"$TEXT_GZIP \" | \
								dd bs=$chunk of=\"$dest\" ) 2>/progress "
					else
						call="atftp \
							--option \"$multicast_atftp\"  \
							--option \"blksize $imageBlkSize\" \
							-g -r \"$path\" -l \"$FETCH_FILE_TEMP_FILE\" $host \
							&> $TRANSFER_ERRORS_FILE ; \
							gzip -d < \"$FETCH_FILE_TEMP_FILE\" | \
							dd bs=$chunk of=\"$dest\" "
					fi
				fi
			else
				if [ $showProgress -eq 1 ];then
					call="atftp \
						--trace \
						--option \"$multicast_atftp\"  \
						--option \"blksize $imageBlkSize\" \
						-g -r \"$path\" -l \"$dest\" $host 2>&1 | \
						atftpProgress \
							$needMByte \"$TEXT_LOAD\" \
							$TRANSFER_ERRORS_FILE $imageBlkSize \
						> /progress"
				else
					call="atftp \
						--option \"$multicast_atftp\"  \
						--option \"blksize $imageBlkSize\" \
						-g -r \"$path\" -l \"$dest\" $host \
						&> $TRANSFER_ERRORS_FILE"
				fi
			fi
			;;
		*)
			systemException "Unknown download type: $type" "reboot"
			;;
	esac
	#======================================
	# run the download
	#--------------------------------------
	if [ $showProgress -eq 1 ];then
		test -e /progress || mkfifo /progress
		test -e $TMPDIR/load_code && rm -f $TMPDIR/load_code
		errorLogStop
		(
			eval $call \; 'echo ${PIPESTATUS[0]} > $TMPDIR/load_code' &>/dev/null
		)&
		call_pid=$!
		echo "cat /progress | dialog \
			--backtitle \"$TEXT_INSTALLTITLE\" \
			--progressbox 3 65
		" > $TMPDIR/progress.sh
		if [ -e /dev/fb0 ];then
			fbiterm -m $UFONT -- bash -e $TMPDIR/progress.sh || \
			bash -e $TMPDIR/progress.sh
		else
			bash -e $TMPDIR/progress.sh
		fi
		clear
		wait $call_pid
		loadCode=`cat $TMPDIR/load_code`
		if [ -z "$loadCode" ]; then
			systemException \
				"Failed to get the download process return value" \
			"reboot"
		fi
	else
		eval $call \; 'loadCode=${PIPESTATUS[0]}'
	fi
	if [ $showProgress -eq 1 ];then
		errorLogContinue
	fi
	loadStatus=`cat $TRANSFER_ERRORS_FILE`
	return $loadCode
}

#======================================
# fetchFileLocal
#--------------------------------------
function fetchFileLocal {
	# /.../
	# ----
	fetchFile "$1" "$2" "$3" "$KIWI_LOCAL_CACHE_DIR" "local" 
}

#======================================
# putFile
#--------------------------------------
function putFile {
	# /.../
	# the generic putFile function is used to upload boot data on
	# a server. Supported protocols are tftp, ftp, http, https
	# ----
	local path=$1
	local dest=$2
	local host=$3
	local type=$4
	local encoded_dest
	if test -z "$path"; then
		systemException "No path specified" "reboot"
	fi
	if test -z "$host"; then
		if test -z "$SERVER"; then
			systemException "No server specified" "reboot"
		fi
		host=$SERVER
	fi
	if test -z "$type"; then
		if test -z "$SERVERTYPE"; then
			type="tftp"
		else
			type="$SERVERTYPE"
		fi
	fi
	encoded_dest=$(encodeURL "$dest")
	case "$type" in
		"local")
			cp -f "$path" "$host/$dest" > $TRANSFER_ERRORS_FILE 2>&1
			return $?
			;;
		"http")
			curl -f -T "$path" http://$host/$encoded_dest \
				> $TRANSFER_ERRORS_FILE 2>&1
			return $?
			;;
		"https")
			curl -f -T "$path" https://$host/$encoded_dest \
				> $TRANSFER_ERRORS_FILE 2>&1
			return $?
			;;
		"ftp")
			curl -T "$path" ftp://$host/$encoded_dest \
				> $TRANSFER_ERRORS_FILE 2>&1
			return $?
			;;
		"tftp")
			atftp -p -l "$path" -r "$dest" $host >/dev/null 2>&1
			return $?
			;;
		*)
			systemException "Unknown download type: $type" "reboot"
			;;
	esac
}


# vim: set noexpandtab:
