#!/bin/sh
#
#==========================================================================
# FILE        : posInitCommon.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2008 Novell inc, USA. All rights reserved
#
# Author      :
#             : Ondrej Holecek <oholecek@novell.com>, 2008
#
# DESCRIPTION : The script is not standalone. It provides support for
#               posInitAdminserver.sh, posInitBranchserver.sh, etc..
#==========================================================================


# CONFTEMPLATE variable must be specified BEFORE this file is . included. Because Admin and Branch servers have different config files
if [ -z "$CONFTEMPLATE" ]; then
    echo -e "CONFTEMPLATE variable must be specified BEFORE this file is . included!\nPlease note, that \"posInitCommon.sh\" should not be executed directly."
    exit 255
fi

# Checking for mandatory files
if [ ! -f "$CONFTEMPLATE" ]; then
    echo "${CONFTEMPLATE} file is missing. Try reinitializing product."
    exit 255
fi

# when changed this POS::Config.pm has to be changed as well
POSCONFPATH=/etc/SLEPOS
KEYFILE=$POSCONFPATH/salt.key

# ERROR CODES
ERR_LDAPLOGIN=1
ERR_ADMINLDAP_NOT_RESOLVABLE=2
ERR_CERTNOTVERIFIED=3
ERR_BRANCHLDAP_NOT_RESOLVABLE=4
ERR_BRANCHLDAP_NOT_CONNECTABLE=5
ERR_BRANCH_LDIF=6
ERR_GLOBAL_LDIF=7
ERR_SERVICE=8
ERR_BRANCH_NETWORK=9
ERR_DN_VALIDATION=10
ERR_GENERAL=255


# Suffix of saved/backuped configuration files
POSSAVE=".POSsave"

#
# backups configuration file before modification (appends ${POSSAVE} suffix)
# $1 - operation to perform (e.g. cp, mv ...)
# $2 - file to backup
#
function backupConf(){
    cmd="$1"
    file="$2"
    [ -f "${file}" ] && [ ! -f "${file}${POSSAVE}" ] && "${cmd}" "${file}" "${file}${POSSAVE}"
}

#
# reads and parses key=value pairs from $CONFTEMPLATE file, chop '"' in the begin
# and end of the string and allow '\"' inside value. If '"' misses in the value
# then  returns whole string after 'key='.
# $1 - key
# $2 - (optional) config file (${CONFTEMPLATE} by default)
# return - value of the key
#
function readValue(){
    if grep "^$1=" /usr/lib/SLEPOS/defaults ; then
        echo "[ERROR] $1 has been moved to /usr/lib/SLEPOS/defaults and should not be read by readValue()"
        exit 1
    fi

    if [ -n "$2" -a -f "$2" ]; then
        eval "source $2;echo \$$1"
    else
        eval "source ${CONFTEMPLATE};echo \$$1"
    fi
}

#
# reads user value Yes No value, make sure User will answer only right values
# provide legend
# $1 - legend
# $2 - default value
# return 0 in case of positive answere (y,Yes)
#        1 in case of negate one
#
function readYesNo(){
    local legend=$1
    local default=$2

    while true; do
        echo -ne "$legend "
        if [ "$default" = "y" ]; then
            echo -n "(yes/no) [yes] "
        elif [ "$default" = "n" ]; then
            echo -n "(yes/no) [no] "
        else
            echo -n "(yes/no) "
        fi
        read -r reply
        case "$reply" in
            [yY]|[yY][eE][sS])
              return 0
              ;;
            [nN]|[nO][oO])
              return 1
              ;;
            "" )
              if [ "$default" = "y" ]; then
                  return 0
              elif [ "$default" = "n" ]; then
                  return 1
              elif [ -n "$default" ]; then
                echo "wrong answer, please try again Valid answers are y yes n no or enter for default"
              fi
              ;;
            *)
              echo "wrong answer, please try again Valid answers are y yes n no"
              ;;
        esac
 done
}

#
# asks for input, disallow empty answers
# $1 - variable name to store result
# $2 - question string
#
function askValue(){
    local ALLOWED_CHARS="[:alnum:]-_."
    local TITLE="$2"
    local CHECK="$3"
    while true; do
        echo -ne "$TITLE"
        read -r TMP
        if [ -n "$TMP" ]; then
            eval "$1=\"$TMP\""
        else
            eval "TMP=\"\$$1\""
        fi
        if eval "[ -n \"\$$1\" ]"; then
            if [ -z $CHECK ]; then
                TMP=`echo "$TMP" | tr -cd $ALLOWED_CHARS`
                if eval "[ \"\$$1\" != $TMP ]"; then
                    eval "$1=$TMP"
                    TITLE=`echo "$TITLE" | sed "s;\[[^\]]*\];\[$TMP\];"`
                    echo "Not allowed characters detected in entered value. Please try again, or press ENTER to accept proposed modification."
                    continue
                fi
            fi
               break
        else
            echo "Value can not be empty. Please try again."
        fi
    done
}

function askPassword(){
    local MSG="$2"
    echo -ne "$MSG: "
    local PASS=`/usr/lib/SLEPOS/posReadPassword.pl`
    echo
    eval "$1=\"$PASS\""
}

# decode password $2 and stores back to $1
function decodePassword(){
    local PASS=`echo "$2" | openssl enc -aes128 -d -kfile  "/etc/SLEPOS/salt.key" -a 2>/dev/null`
    if [ -z "$PASS" ]; then
        #decoding failed, load raw value
        PASS="$2"
    fi
    eval "$1=\"$PASS\""
}

#
# changes option values in sysconfig files
# $1 - variable/option
# $2 - new value
# $3 - sysconfig file
# $4 - "single" for replacing old value by new one otherwise append new value
#      to variable/option (usable for multivalue options)
#
function updateSysconfigFile(){
    local VAR="$1"
    local VALUE="$2"
    local FILE="$3"
    local SINGLE="$4"
    local OLDVALUE
    local tmpline
    local tmpbar=""

    if ! tmpline=`grep "^[[:space:]]*$VAR=" "$FILE"`; then
        echo -e "\n$VAR=\"$VALUE\"" >> "$FILE"
        return;
    fi

    OLDVALUE=`eval "eval '$tmpline' ; echo \"\\\$$VAR\"" 2>/dev/null`

    if [ "$SINGLE" = "single" ]; then
        if [ "$OLDVALUE" != "$VALUE" ]; then
            #only modify variable/option when our entry/value not yet there
            sed -i -e 's@^'"$tmpline"'$@'"$VAR=\"$VALUE\""'@' "$FILE"
        fi
    elif [ "$SINGLE" = "first" ]; then
        # add $VALUE as first
        if [ -n "$OLDVALUE" ]; then
            tmpbar=" "
        fi
        if ! echo "$OLDVALUE" | egrep -q "(^|[ \t])$VALUE([ \t]|$)"; then
            sed -i -e 's@^'"$tmpline"'$@'"$VAR=\"${VALUE}${tmpbar}${OLDVALUE}\""'@' "$FILE"
        fi
    else
        if [ -n "$OLDVALUE" ]; then
            tmpbar=" "
        fi
        if ! echo "$OLDVALUE" | egrep -q "(^|[ \t])$VALUE([ \t]|$)"; then
            sed -i -e 's@^'"$tmpline"'$@'"$VAR=\"${OLDVALUE}${tmpbar}${VALUE}\""'@' "$FILE"
        fi
    fi
}

function err(){
    local ERROR=$1
    local MSG=$2
    test -z "$ERROR" && ERROR=$ERR_GENERAL
    case $ERROR in
        $ERR_LDAPLOGIN)
            test -z "$MSG" && MSG="[ERROR: LDAP] Bind DN or ldap password or both are incorrect. LDAP login FAILED"
            ;;
        $ERR_ADMINLDAP_NOT_RESOLVABLE)
            test -z "$MSG" && MSG="[ERROR: LDAP] Admin LDAP is not resolvable. Please check entered Admin server hostname, or your dns settings."
            ;;
        $ERR_CERTNOTVERIFIED)
            test -z "$MSG" && MSG="[ERROR: CERT] Certificate rejected by user. Please install valid certificate and try again."
            ;;
        $ERR_BRANCHLDAP_NOT_RESOLVABLE)
            test -z "$MSG" && MSG="[ERROR: LDAP] Branch-local LDAP server hostname is not resolvable. Please check entered hostname, or your dns settings."
            ;;
        $ERR_BRANCHLDAP_NOT_CONNECTABLE)
            test -z "$MSG" && MSG="[ERROR: LDAP] Branch-local LDAP server is not connectable. Please check whether LDAP server is running, correct address is entered and is connected to network (if not localhost)."
            ;;
        $ERR_BRANCH_LDIF)
            test -z "$MSG" && MSG="[ERROR: LDAP] Branch LDAP part is corrupted. Unable to import. New offline installation source will probably be needed."
            ;;
        $ERR_BRANCH_NETWORK)
            test -z "$MSG" && MSG="[ERROR: NET] BranchServer network error."
            ;;
        $ERR_SERVICE)
            test -z "$MSG" && MSG="[ERROR: SERVICE] Service failed to start. See syslog for details."
            ;;
        *)
            test -z "$MSG" && MSG="$ERROR" && ERROR=$ERR_GENERAL
            ;;
    esac
    echo "$MSG"
    test -f "$CONFPATH" && echo "Storing temporary configuration as branchserver.conf.error and exiting..." && mv "$CONFPATH" "${CONFPATH}.error"
    exit $ERROR
}

function encryptPassword(){
    ## check salt
    # always generate new if keyfile is missing. Do not touch existing key when on combo and/or "no-key-reset" option is used
    if [ ! -f "$KEYFILE" -o -z "$COMBO" -a "x$1" != "xno-key-reset" ]; then
        echo -n $RANDOM > "$KEYFILE"
    fi
    # obfuscate password with salt
    export ENCRYPTEDPASSWORD=`echo -n $PASSWORD | openssl enc -aes128 -kfile $KEYFILE | base64 -w0`
}

function processTemplate(){
    local I=0
    local ARG
    local KEY
    local VAL
    local INPLACE
    local IN
    local OUT

    if test "x$1" == "x-i" ; then
        INPLACE=yes
        IN="$2"
    else
        IN="$1"
        OUT="$2"
    fi
    shift 2

    while test "x$1" != "x" ; do
        KEY="$1"
        VAL="$2"
        shift 2

        VAL="${VAL//\\/\\\\}"
        VAL="${VAL//|/\\|}"

        ARG[$I    ]="-e"
        ARG[$I + 1]="s|$KEY|$VAL|g"
        I=$[$I + 2]
    done

    if test -n "$INPLACE" ; then
        sed -i "${ARG[@]}" "$IN"
    else
        sed "${ARG[@]}" "$IN" > "$OUT"
    fi
}

function warnNotReplaced() {
    if grep "@[A-Z_]*@" "$@" || grep "##[A-Z_]\+##" "$@"; then
        echo WARNING: unreplaced template strings found
    fi
    true
}

# 
# process tftp template (... pxelinux.cfg/default)
# $1 - config path
# $2 - config file
#
function processTFTPTemplate() {
    local POS_MACSEP_PARAM=""
    local CONF_PATH="$1"
    local CONF_FILE="$2"

    install -d -m 0755 -o root -g nogroup "${CONF_PATH}"

    if [ ! "${POS_MACSEP}" = "" ] && [ ! "${POS_MACSEP}" = ":" ]; then
        POS_MACSEP_PARAM="${POS_MACSEP_PREFIX}=\"${POS_MACSEP}\""
    fi

    backupConf mv "${CONF_FILE}"
    processTemplate "${TFTP_BOOT_PXELINUXCFG_CONFIG_TEMPLATE}" "${CONF_FILE}" \
                @LINUXKERNEL@       "$TFTP_DEFAULT_KERNEL"            \
                @LINUXINITRDGZ@     "$TFTP_DEFAULT_INITRD"            \
                @POS_MACSEP@        "${POS_MACSEP_PARAM}"             \
                @USERKERNELPARAMS@  "POS_KERNEL_PARAMS_HASH=${TFTP_DEFAULT_HASH}"
}

# validateDN($dn)
# check given DN if is complete and without illegal characters
# returns 1 on error, else 0
function validateDN() {
    local DN=`echo $1 | tr ',' ' '`
    local RESULT=0
    for x in $DN; do
        #$x must begin with cn=,ou=,o=,c= or dc= and then unempty string must follow
        case "$x" in
            cn=*|ou=*|o=*|c=*|dc=*)
            tmp=$(echo $x | sed -e 's/^[^=]*=\(.*\)/\1/')
            if [ -z "$tmp" ]; then
                echo "[ERROR: DN] Empty entry name in $x"
                RESULT=1
            elif [ $tmp !=  $(echo $tmp | tr -d '=') ]; then
                echo "[ERROR: DN] Illegal characters in entry $x"
                RESULT=1
            fi
            ;;
            *)
            #nok
            echo "[ERROR: DN] Unsupported LDAP objectClass or malformed entry $x"
            RESULT=1
            ;;
        esac
    done
    return $RESULT
}

# start service
# $1 specifies service name
# function checks init script return value and in case it failed, waits a little and check service status again
# if second check failed, it err
function startService() {
    INITDIR="/etc/init.d"
    SERVICE="$1"
    if ${INITDIR}/${SERVICE} status >/dev/null 2>&1; then
        #service is running, stop it at first
        stopService $SERVICE
    fi
    ${INITDIR}/${SERVICE} start && return 0
    # init script reports it failed to start, lets wait a little and check status again
    sleep 5
    if ! ${INITDIR}/${SERVICE} status; then
        # service is still not running, something else is going on
        err $ERR_SERVICE "[ERROR] ${SERVICE} service failed to start!"
    fi
}

# stop service
# $1 specifies service name
# function checks init script return value and in case it failed, waits a little and check service status again
# if second check failed, it err
function stopService() {
    INITDIR="/etc/init.d"
    SERVICE="$1"
    ${INITDIR}/${SERVICE} stop && return 0
    # init script reports it failed to start, lets wait a little and check status again
    sleep 5
    if ${INITDIR}/${SERVICE} status; then
        # service is still running, something else is going on
        err $ERR_SERVICE "[ERROR] ${SERVICE} service failed to stop!"
    fi
}