#!/bin/bash
#
#==========================================================================
# FILE        : save_poslogs.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2013 SUSE LINUX Products GmbH, Nuernberg, All rights reserved
#               (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Ondrej Holecek <oholecek@suse.com>
#               Petr Ostadal <postadal@suse.com>
#
# DESCRIPTION : Collects logs, configuration, messages and SLEPOS LDAP database
#               for problem and error reporting
#==========================================================================

appName=`basename ${0}`
appVersion="0.11.3"

################################################################
##############          Local functions          ###############
################################################################

function usage(){
    cat << EOF_USAGE
${appName} v${appVersion}

Usage:  ${appName} [--noldap] [--nolog] [--noconfig] [--force] <tgz-file-name>|<bzip2-file-name>
        ${appName} -h|--help
EOF_USAGE
}

function help(){
    usage
    cat << EOF_HELP

Description:
    Collects logs, messages and configurations for problem and error reporting.
    Result is stored in specified compressed tar file.

Options:
 --noldap                    do not include ldap database dump
 --nolog                     do not include logging files
 --noconfig                  do not include configuration files
 -V, --version               print the version of script being used
 -h, --help                  displays this help
 --usage                     display brief usage message
EOF_HELP
}

################################################################
# config files enumeration
SLEPOSCONFDIR="/etc/SLEPOS"
LDAPCONFDIR="/etc/openldap"
SRVCONFDIR="/srv/SLEPOS"
SYSCONFDIR="/etc/sysconfig"
KIWIDIR="/srv/tftpboot/KIWI"
ROLLBACKBOOT="$KIWIDIR/rollback/boot"
UPLOADDIR="/srv/tftpboot/upload"

BRANCHSERVERCONF="$SLEPOSCONFDIR/branchserver.conf"
ADMINSERVERCONF="$SLEPOSCONFDIR/adminserver.conf"

######## CONFIGURATION FILELIST ########
# use '!' as first character to exclude such path(name)
# from all above path(name)s
function configFiles(){
    if [ -n "$BSINSTALLED" ]; then
        # BRANCH SERVER filelist         ########
        cat << EOBSCF
$BRANCHSERVERCONF
$BRANCHSERVERCONF.*
$SLEPOSCONFDIR/rsyncd
$SLEPOSCONFDIR/dhcpd
$SLEPOSCONFDIR/named.d
$SLEPOSCONFDIR/named.conf
$SLEPOSCONFDIR/named.conf.include
$LDAPCONFDIR/slapd.conf
$SYSCONFDIR/atftpd
$SYSCONFDIR/dhcpd
$SYSCONFDIR/named
$SYSCONFDIR/openldap
$SYSCONFDIR/network/config
$KIWIDIR
!$ROLLBACKBOOT
$ROLLBACKBOOT/pxe.*
$UPLOADDIR
/srv/tftpboot/boot/pxelinux.cfg
/var/lib/dhcp/db
/var/lib/named/etc
/etc/pure-ftpd/pure-ftpd.conf
EOBSCF
        # END OF BRANCH SERVER filelist  ########
    fi
    if [ -n "$ASINSTALLED" ]; then
        # ADMIN SERVER filelist        ########
        cat << EOASCF
$ADMINSERVERCONF
$ADMINSERVERCONF.*
$SLEPOSCONFDIR/rsyncd.conf
$SYSCONFIG/openldap
EOASCF
        # END OF ADMIN SERVER filelist ########
    fi
}
######## END OF CONFIGURATION FILELIST ########

######## LOG FILELIST (excluding yast2 logs) ########
# use '!' as first character to exclude such path(name)
# from all above path(name)s
function logFiles(){
    cat << EOLF
/var/log/messages
/var/log/messages-*.bz2
/var/log/rsyncd.log
/var/log/rc.dhcp.log
/var/log/pos
EOLF
}
######## END OF LOG FILELIST ########

# For every file check if exist, expand directories. 
# Expand wildcards. If negative selection is encountered,
# it is applied to all previous (expanded) items in the list
function existing_Files(){
    sep=''
    while read f_glob; do
        for f in $f_glob ; do
            ff=`echo $f | sed -e 's/^!//'`
            if [ "$f" != "$ff" ];then
                #remove negative selections from previously processed files (but not from those that come later)
                flist=`echo "$flist" | grep -v "$ff"`
            else
                ffind=`find  "$f" -type f -o -type l 2>/dev/null`
                if [ -n "$ffind" ]; then
                    flist="$flist$sep$ffind"
                    sep="$IFS"
                fi;
            fi;
        done
    done
    echo "$flist"
}

# content generating functions
function saveLdap(){
    local db_suffix
    echo "########################################"
    echo "# LDAP DB export"
    if test -n "$ALL_LDAPBASE"; then
        db_suffix="-b $ALL_LDAPBASE"
    fi
    if slapcat -d 0 -o ldif-wrap=no $db_suffix | grep -v "userPassword" > "$TARDIR/ldap.ldif"; then
        echo "LDAP DB exported to ldap.ldif"
    else
        echo "[ERROR] Failed to export LDAP DB!"
    fi
    echo "# END OF LDAP DB export"
    echo "########################################"
}

function saveLogs(){
    echo "########################################"
    echo "# LOG FILE LIST"
    mkdir "$TARDIR/logs"
    logFiles |existing_Files | while read f; do
        cp -d --parents "$f" "$TARDIR/logs"
        echo "Collecting log file: $f"
    done
    save_y2logs "$TARDIR/logs/y2logs.tgz"
    echo "# END OF LOG FILE LIST"
    echo "########################################"
}

function saveConf(){
    echo "########################################"
    echo "# CONFIGURATION FILE LIST"
    mkdir "$TARDIR/configs"
    configFiles | existing_Files | sort -u | while read f; do
        cp -d --parents "$f" "$TARDIR/configs"
        echo "Collecting config file: $f"
    done
    ls -lR $SRVCONFDIR /srv/tftpboot /var/lib/SLEPOS > "$TARDIR/ls-lR" 2>/dev/null
    echo "# END OF CONFIGURATION FILE LIST"
    echo "########################################"
}

function saveLdapConf(){
    echo "########################################"
    echo "# LDAP CONFIGURATION"
    [ -f "/etc/sysconfig/openldap" ] && . "/etc/sysconfig/openldap"
    if [ "$OPENLDAP_CONFIG_BACKEND" == "ldap" ]; then
        echo "# parsing ${LDAPCONFDIR}/slapd.d/"
        echo "${LDAPCONFDIR}/slapd.d/" | existing_Files | while read f; do
            cp -d --parents "$f" "$TARDIR/configs"
            # replace password by generic adminpass so configurations are testable
            sed -i -e 's/^olcRootPW:.*$/olcRootPW: adminpass/' "${TARDIR}/configs/$f"
            # remove DB configs with different suffix
            # first grep is to filter out non-db configs, second to match only our suffix
            if grep "olcSuffix: " "${TARDIR}/configs/$f" |
               grep -v "olcSuffix: $ALL_LDAPBASE" >/dev/null; then
                rm -f "${TARDIR}/configs/$f"
            else
                echo "Collecting LDAP config file: $f"
            fi
        done
    else
        echo "# parsing ${LDAPCONFDIR}/slapd.conf"
        echo "${LDAPCONFDIR}/slapd.conf \
${LDAPCONFDIR}/acl.ldif" | existing_Files | while read f; do
            cp -d --parents "$f" "$TARDIR/configs"
            echo "Collecting LDAP config file: $f"
        done
    fi
    echo "# END OF LDAP CONFIGURATION"
    echo "########################################"
}

################################################################
##############              Main                 ###############
################################################################

#parsing commandline params
# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o Vh -l noldap,nolog,noconfig,help,usage,version \
      -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 254; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"

while [ $# -gt  0 ]; do
    case $1 in
        --noldap)
            noLdap="yes"
            ;;
        --nolog)
            noLog="yes"
            ;;
        --noconfig)
            noConfig="yes"
            ;;
        -h|--help)
            #just show help and exit
            help
            exit 0
            ;;
        -V|--version)
            echo "${appName} ${appVersion}"
            exit 0
            ;;
        --usage)
            usage
            exit 0
            ;;
        --)
            ;;
        *)
            outFile=$1
            # check for output file and its encoding format
            case "$outFile" in
                *.tgz|*.tar.gz)
                    tar_compression="--gzip" ;;
                *.tar.bz2)
                    tar_compression="--bzip2" ;;
                *.tar)
                    tar_compression="" ;;
                *)
                    echo "[ERROR] Archive file '${outFile}' is not supported" >&2
                    echo "        Use one of: .tar .tgz .tar.gz .tar.bz2"
                    usage
                    exit 2
                    ;;
            esac
            ;;
    esac
    shift
done

if [ -z "$outFile" ]; then
    echo "[ERROR] Output file not specified!"
    usage
    exit 2
fi

#detect installation type
if [ -f "$BRANCHSERVERCONF" ]; then
    BSINSTALLED="yes"
    echo "BranchServer installation detected"
fi
if [ -f "$ADMINSERVERCONF" ]; then
    ASINSTALLED="yes"
    # detect database DN suffix
    eval `grep ALL_LDAPBASE "$ADMINSERVERCONF"`
    echo "AdminServer installation detected, DB suffix $ALL_LDAPBASE"
fi
if [ -z "$ASINSTALLED" ] && [ -z "$BSINSTALLED" ]; then
    echo "[WARN] Unable to identify installed SLEPOS type. Trying both AS and BS." >&2
    ASINSTALLED="yes"
    BSINSTALLED="yes"
fi

TMPDIR=`mktemp --tmpdir -d "tmp.SLEPOS.log.XXXXXXXXXX"`
LOGDIR="SLEPOS11-data"
TARDIR="$TMPDIR/$LOGDIR"
LOGFILE="$TARDIR/diag.log"

mkdir "$TARDIR"

#save stdout
exec 7>&1
#redirect to logfile
exec 1>"$LOGFILE"

cat << EOH
########################################
# SLEPOS11 diagnosis log
########################################
########################################
EOH

if [ -z "$noConfig" ]; then
    saveConf
    saveLdapConf
fi
if [ -z "$noLog" ]; then
    saveLogs
fi
if [ -z "$noLdap" ]; then
    saveLdap
fi

cat << EOF
########################################
########################################
# END OF SLEPOS11 diagnosis log
########################################
EOF

#now tar tmpdir to output file
tar $tar_compression -cS -f "$outFile" -C "$TMPDIR" "$LOGDIR"

#close log file
exec 1>&-
#restore stdout
exec 1>&7
#close FD7
exec 7>&-

rm -rf "$TMPDIR"

echo "Diagnosis log saved to $outFile file"
